'use strict';

/**
 * @ngdoc overview
 * @name exampleWebappClientApp
 * @description
 * # exampleWebappClientApp
 *
 * Main module of the application.
 */
angular
  .module('smartexploreApp', [
    'ngAnimate',
    'ngAria',
    'ngCookies',
    'ngMessages',
    'ngResource',
    'ngRoute',
    'ngSanitize',
    // 'ngTouch',
    'ngMaterial',
    'hc.downloader',
    'material.components.expansionPanels',
    'angular-loading-bar',
    'blockUI'
  ])
  .config(function ($routeProvider) {

    $routeProvider
      .when('/', {
        templateUrl: 'views/main.html',
        controller: 'MainCtrl',
        controllerAs: 'main'
      })
      .when('/dataset', {
        templateUrl: 'views/dataset.html',
        controller: 'DatasetCtrl',
        controllerAs: 'dataset',
        reloadOnSearch: false       // do not reload the page, once search parameters have changed.
      })
      .when('/visualization', {
        templateUrl: 'views/visualization.html',
        controller: 'VisualizationCtrl',
        controllerAs: 'vis',
        reloadOnSearch: false       // do not reload the page, once search parameters have changed.
      })
      .when('/about', {
        templateUrl: 'views/about.html',
        controller: 'AboutCtrl',
        controllerAs: 'about'
      })
      .otherwise({
        redirectTo: '/'
      });

  })
  .config(function($mdThemingProvider) {
    $mdThemingProvider.theme('default')
      .primaryPalette('indigo')
      .accentPalette('indigo');
  })
  .config(function($mdAriaProvider) {
    // Globally disables all ARIA warnings. Only warnings for angular. Taken from: https://github.com/textAngular/textAngular/issues/1446
    $mdAriaProvider.disableWarnings();
  });

'use strict';

/**
 * @ngdoc function
 * @name exampleWebappClientApp.controller:MainCtrl
 * @description
 * # MainCtrl
 * Controller of the exampleWebappClientApp
 */
angular.module('smartexploreApp')
  .controller('MainCtrl', function () {
    this.awesomeThings = [
      'HTML5 Boilerplate',
      'AngularJS',
      'Karma'
    ];
  });

'use strict';

/**
 * @ngdoc function
 * @name exampleWebappClientApp.controller:AboutCtrl
 * @description
 * # AboutCtrl
 * Controller of the exampleWebappClientApp
 */
angular.module('smartexploreApp')
  .controller('AboutCtrl', function ($scope, dataService) {

    // $scope.d = {};
    // $scope.d.appName = "SMARTexplore";
    // $scope.d.appVersion = "1.0v";

    $scope.d = dataService.getAppInfo();

  });

'use strict';

/**
 * @ngdoc function
 * @name langchangeClientApp.controller:HeaderCtrl
 * @description
 * # HeaderCtrl
 * Controller of the langchangeClientApp
 */
angular.module('smartexploreApp')
  .controller('HeaderCtrl', function ($scope, $location) {
    $scope.isActive = function (viewLocation) {
      return viewLocation === $location.path();
    };

    /**
     * Navigate to the given location. Used instead of <a href="location"> as navigation using $location.path preserves all parameters.
     * @param location the location to navigate to, e.g. "/about"
     */
    $scope.navigateTo = function(location){
      $location.path(location);
    }

  });

'use strict';

/**
 * @ngdoc function
 * @name smartexploreApp.controller:VisualizationCtrl
 * @description
 * # VisualizationCtrl
 * Controller of the smartexploreApp
 */
angular.module('smartexploreApp')
  .controller('VisualizationCtrl', function ($scope, dataService, tooltipService, urlParameterService,
                                             $location, $q, aggregationPreviewService, $mdSidenav,
                                             $mdExpansionPanel, $mdExpansionPanelGroup, statisticsService,
                                             $mdDialog, selectionService, updateVisService, $timeout, featureTableVisualization) {

    /** Stores all dimensions of the dataset (and not only the currently filtered ones). */
    $scope.allDimensionsOfDataset = null;

    /** List with categorical dimensions; used for the aggregation. */
    $scope.categoricalDimensions = [];

    /** Feature table (without layout information). */
    $scope.featureTable = null;

    /** Layout information of the feature table. */
    $scope.featureTableLayout = null;

    /** This boolean will be flipped in case the visualization needs to be updated. */
    $scope.notifyUpdateFeatureTable = {'update': true};

    /** The id of the currently analyzed dataset. */
    $scope.idOfDataset = null;

    /** Holds the information about the currently used dataset. */
    $scope.selectedDataset = null;

    /** The current aggregation of the feature table. */
    $scope.aggregation = null;
    /** The current aggregation of the second aggregation level. */
    $scope.aggregation2ndLevel = null;

    /** The layout of the glyphs. */
    $scope.glyphLayout = undefined;

    /** The sort order of all dimension groups.
     *  The value is 'undefined' if there are different orderings for different groups.
     */
    $scope.sortingOrderAllDimensionGroups = undefined;
    $scope.distanceFunctionsAllDimensionGroups = undefined;
    $scope.normalizationAllDimensionGroups = undefined;

    /** The layout of the data record groups (rows of the visualization). */
    $scope.layoutDataRecordGroups = undefined;

    /** The layout of the dimension groups (groups of columns). */
    $scope.layoutDimensionGroups = undefined;

    /** The dimension which is selected for the major aggregation (1. level). */
    $scope.selectedDimensionForAggregation = false;
    /** The dimension which is selected for the 2nd level aggregation. */
    $scope.selectedDimensionForAggregation2ndLevel = undefined;

    /** Data for the tooltip of a categorical glyph. See tooltipService.js.  */
    $scope.dataTooltipCategorical = undefined;

    /** Data for the tooltip of a numerical glyph. See tooltipService.js.  */
    $scope.dataTooltipNumerical = undefined;

    /** All available glyph designs for distance to mean. */
    $scope.availableGlyphDesignsForDistanceToMean = NumericalGlyph.GLYPH_DESIGN;
    $scope.availableColormapsForMean = NumericalGlyph.COLORMAPS;

    /** All available glyph designs for distance to average distribution. */
    $scope.availableGlyphDesignsForDistanceToAverageDistribution = CategoricalGlyph.GLYPH_DESIGN_CATEGORICAL;

    /** All available glyph overlays/designs for significance analysis. */
    $scope.availableGlyphDesignsForSignificance = GlyphLayout.GLYPH_DESIGN_SIGNIFICANCE;

    /** Lists all available sortings; necessary for the GUI. */
    $scope.availableSortings = Sorting.AVAILABLE_SORTINGS;

    /** List all available distance functions; necessary for the GUI. */
    $scope.availableDistanceFunctions = DimensionDistance.AVAILABLE_DISTANCES_DIMENSION;

    /** List all available normalizing strategies; necessary for the GUI. */
    $scope.availableNormalizingStrategies = AVAILABLE_NORMALIZING_STRATEGIES;

    // so that we can bind it to the gui.
    $scope.statisticsService = statisticsService;
    $scope.selectionService = selectionService;

    /** Object to highlight dimensions (similarity to another dimension). */
    $scope.highlight = undefined;

    // store the previous selected aggregation to go one step back
    let previousSelectedDimensionForAggregation = undefined;
    let previousSelectedDimensionFor2ndAggregation = undefined;

    $scope.initDimensionGroups = function(){
      $scope.layoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();
    };

    /**
     * Function to toggle side nav with component id of the side nav.
     */
    $scope.toggleSideNav = function(componentId) {
      return $mdSidenav(componentId).toggle();
    };

    /**
     * Called when an expansion panel is clicked. This updates the panelClickedName var and $watcher is fired.
     * @param panelClicked the name of the clicked panel
     */
    $scope.panelClicked = function(panelClicked) {
      $scope.panelClickedName = panelClicked;
    };


    /** Init the menu on the left. */
    function initMenu(){

      // parse the distribution from the gui.
      $mdExpansionPanel().waitFor('glyphMappingLayout').then(function (instance) {
        instance.expand();
      });
    }
    initMenu();

    /**
     * Checks the sorting in all dimension groups. If it is identical, set value in the GUI (menu), otherwise set to undefined (-> different for every group)
     */
    let checkSortingInGroups = function(){

      let groups = $scope.layoutDimensionGroups._dimensionGroups;
      let initialValue = groups[0]._sortingStrategy._sortOrder;

      for(let i=1; i<groups.length; i++){
        if(groups[i]._sortingStrategy._sortOrder !== initialValue){
            $scope.sortingOrderAllDimensionGroups = undefined;
          return;
        }
      }

      $scope.sortingOrderAllDimensionGroups = initialValue;
    };

    /**
     * Checks the normalization strategy in all dimension groups. If it is identical, set value in the GUI (menu), otherwise set to undefined (-> different for every group)
     */
    let checkNormalizationInGroups = function(){

      let groups = $scope.layoutDimensionGroups._dimensionGroups;
      let initialValue = groups[0]._normalizingStrategy._normalizing;

      for(let i=1; i<groups.length; i++){
        if(groups[i]._normalizingStrategy._normalizing !== initialValue){
          $scope.normalizationAllDimensionGroups = undefined;
          return;
        }
      }

      $scope.normalizationAllDimensionGroups = initialValue;

    };

    /**
     * Checks the distance functions in all dimension groups. If it is identical, set value in the GUI (menu), otherwise set to undefined (-> different for every group)
     */
    let checkDistanceMeasuresInGroups = function(){

      let groups = $scope.layoutDimensionGroups._dimensionGroups;
      let initialValue = groups[0]._dimensionDistance._distanceFunction;

      for(let i=1; i<groups.length; i++){
        if(groups[i]._dimensionDistance._distanceFunction !== initialValue){
          $scope.distanceFunctionsAllDimensionGroups = undefined;
          return;
        }
      }

      $scope.distanceFunctionsAllDimensionGroups = initialValue;
    };

    // =======================================================================
    // Load all data that is required to show this page and check if a dataset has already been selected.
    let checkParametersRetrieveDataAndVisualize = function(){

      // ---
      // check the layout of the individual glyphs - see GlyphLayout.js
      // if not available, load default and set to the URL parameters.
      $scope.glyphLayout = urlParameterService.getGlyphLayout();
      if($scope.glyphLayout === undefined){
        $scope.glyphLayout = GlyphLayout.getDefaultGlyphLayout();
        urlParameterService.setGlyphLayout($scope.glyphLayout);
      }


      // ---
      // check dataset
      // check if the URL parameters contain and information about the currently used dataset
      $scope.idOfDataset = urlParameterService.getDatabaseId();//  $location.search()[UrlParameters.DB_NAME];
      if($scope.idOfDataset === undefined || $scope.idOfDataset === null){
       alert("Please select a dataset, then continue to the visualization. \n You will now be forwarded to the data selection view." );
       $location.path("/dataset");
       return;
      }

      // the information of the dataset (name etc.) from the server - we don't have to wait for the call back
      $scope.selectedDataset = dataService.getDataset();


      // ---
      // get all dimensions of the dataset; init the dialog to select the dimension of the aggregation.
      let getAllDimensionsTmp = dataService.getAllDimensionsOfDataset().$promise;
      getAllDimensionsTmp.then(function(d){
        $scope.allDimensionsOfDataset = d;

        // extract the categorical dimensions for the aggregation
        $scope.categoricalDimensions = [];
        for(let i=0; i<$scope.allDimensionsOfDataset.length; i++){
          if($scope.allDimensionsOfDataset[i]['value'] === 'CATEGORICAL'){
            $scope.categoricalDimensions.push({'key':$scope.allDimensionsOfDataset[i]['key'], 'value':'CATEGORICAL'});
          }
        }
      });


      $scope.layoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();
      $scope.layoutDataRecordGroups = urlParameterService.getLayoutDataRecordGroups();

      // Visualization either by the (default) aggregation or by a customized layout for the dimension and record groups.
      if($scope.layoutDimensionGroups && $scope.layoutDataRecordGroups){
        retrieveByLayoutAndVisualize();
      }else{
        retrieveByAggregationAndVisualize();
      }
    };

    /**
     * Retrieve the feature table by the layout of the data record groups and dimension groups.
     * Update the visualization afterwards.
     */
    let retrieveByLayoutAndVisualize = function(){

      // set the sorting etc. of all gui elements, if they are the same.
      checkSortingInGroups();
      checkDistanceMeasuresInGroups();
      checkNormalizationInGroups();

      // Retrieve info about the applied aggregation (only for the GUI)
      $scope.aggregation = urlParameterService.getAggregation();
      $scope.selectedDimensionForAggregation = $scope.aggregation ? $scope.aggregation._dimensionName : undefined;
      $scope.aggregation2ndLevel = urlParameterService.getAggregation2ndLevel();
      $scope.selectedDimensionForAggregation2ndLevel = $scope.aggregation2ndLevel ? $scope.aggregation2ndLevel._dimensionName :  undefined;
      // store the previous selected aggregation to go one step back
      previousSelectedDimensionForAggregation = $scope.selectedDimensionForAggregation;
      previousSelectedDimensionFor2ndAggregation = $scope.selectedDimensionForAggregation2ndLevel;

      // Retrieve the feature table with the correct properties.
      let featureTableTmp = dataService.getFeatureTableByLayout().$promise;

      featureTableTmp.then(function(featureTableData) {

        $scope.featureTable = new FeatureTable(featureTableData);

        // ---
        console.log("$scope.featureTable", $scope.featureTable);
        console.log("$scope.layoutDataRecordGroups", $scope.layoutDataRecordGroups);
        console.log("$scope.layoutDimensionGroups", $scope.layoutDimensionGroups);
        console.log("You can now create the visualization.");

        // notify the directive to update the visualization.
        $scope.updateVisualization();

      }).catch(console.error);
    };

    /**
     * Retrieve the feature table the given aggregators.
     */
    let retrieveByAggregationAndVisualize = function(){

      // ---
      // check aggregation
      $scope.aggregation = urlParameterService.getAggregation();
      if($scope.aggregation === undefined || $scope.aggregation === null){
        $scope.selectedDimensionForAggregation = undefined;
        return;
      }else{
        // load the aggregation from the url.
        $scope.selectedDimensionForAggregation = $scope.aggregation._dimensionName;
      }

      // check aggregation 2nd level
      $scope.aggregation2ndLevel = urlParameterService.getAggregation2ndLevel();
      if($scope.aggregation2ndLevel === undefined || $scope.aggregation2ndLevel === null){
        $scope.selectedDimensionForAggregation2ndLevel = undefined;
      }else{
        $scope.selectedDimensionForAggregation2ndLevel = $scope.aggregation2ndLevel._dimensionName;
      }

      // store the previous selected aggregation to go one step back
      previousSelectedDimensionForAggregation = $scope.selectedDimensionForAggregation;
      previousSelectedDimensionFor2ndAggregation = $scope.selectedDimensionForAggregation2ndLevel;

      // ---
      // Get the feature table data. IMPORTANT: Wait until the data is there which also means the feature table is cached on the server.
      // Afterwards, the layout etc. can be loaded much faster.
      let featureTableTmp = dataService.getDefaultFeatureTable().$promise;
      // requestsToWaitForVis.push(featureTableTmp);
      featureTableTmp.then(function(featureTableData){

        // Do not store as $scope.featureTable here is only default in order to ge the layout for the data record groups and dimension groups.
        // $scope.featureTable = new FeatureTable(featureTableData);

        let requestsToWaitForVis = [];

        // ---
        // check layout of the data record groups
        $scope.layoutDataRecordGroups = urlParameterService.getLayoutDataRecordGroups();
        if($scope.layoutDataRecordGroups === undefined || $scope.layoutDataRecordGroups === null){
          let layout = dataService.getDefaultLayoutDataRecordGroups().$promise;
          requestsToWaitForVis.push(layout);
          layout.then(function(layoutData){
            $scope.layoutDataRecordGroups = new LayoutDataRecordGroups(layoutData);
            urlParameterService.setLayoutDataRecordGroups($scope.layoutDataRecordGroups);
          });
        }

        // ---
        // check layout of the dimension groups
        $scope.layoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

        if($scope.layoutDimensionGroups === undefined || $scope.layoutDimensionGroups === null){
          let layout = dataService.getDefaultLayoutDimensionGroups().$promise;
          requestsToWaitForVis.push(layout);
          layout.then(function(layoutData){
            $scope.layoutDimensionGroups = new LayoutDimensionGroups(layoutData);
            urlParameterService.setLayoutDimensionGroups($scope.layoutDimensionGroups);
          });
        }

        // wait until all requests are done. Then retrieve the correct layout.
        $q.all(requestsToWaitForVis).then(data => {
          retrieveByLayoutAndVisualize();
        });


      });

    };

    /**
     * This method is always called in the beginning. It checks for the correct parameters and visualizes the feature table.
     */
    checkParametersRetrieveDataAndVisualize();

    /**
     * Update the visualization with the currently set measures.
     */
    $scope.updateVisualization = function(){

      $scope.layoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();
      $scope.layoutDataRecordGroups = urlParameterService.getLayoutDataRecordGroups();

      // remove the current visualization.
      let d3Selector = d3.select("#visMyVis");
      d3Selector.selectAll("*").remove();

      featureTableVisualization.createFeatureTable(d3Selector, $scope.featureTable, $scope.glyphLayout, $scope.layoutDataRecordGroups, $scope.layoutDimensionGroups);
    };


    /**
     * This function is called, if the aggregation value has been changed.
     * The change is reflected in the variable: '$scope.selectedDimensionForAggregation'
     *
     * Action:
     *  - Remove the 2nd level aggregation.
     *  - Retrieve the new layout for the data record groups.
     */
    $scope.aggregationChanged = function(){

      let dimensionName = $scope.selectedDimensionForAggregation;

      let dataType = undefined;
      let allValuesOfDimension = undefined;

      let dataTypeTmp = dataService.getDataTypeOfDimension(dimensionName).$promise;
      dataTypeTmp.then(function(d){
        dataType = d.dimType;
        let allValuesOfDimensionTmp = dataService.getValuesOfDimension(dimensionName, dataType).$promise;
        allValuesOfDimensionTmp.then(function(d){
            allValuesOfDimension = d;

            // function which is executed after the user has confirmed the aggregation
            let confirm = function(){

              // revert the second level aggregation
              $scope.selectedDimensionForAggregation2ndLevel = undefined;
              previousSelectedDimensionFor2ndAggregation = $scope.selectedDimensionForAggregation2ndLevel;
              urlParameterService.setAggregation2ndLevel();

              previousSelectedDimensionForAggregation = $scope.selectedDimensionForAggregation;
              urlParameterService.setAggregation($scope.aggregation);

              // Get the layout for the data record groups based on the new aggregation
              let tmp = dataService.getLayoutDataRecordGroupsByAggregation().$promise;
              tmp.then(function(layoutDataRecordGroups){

                $scope.layoutDataRecordGroups = new LayoutDataRecordGroups(layoutDataRecordGroups);
                urlParameterService.setLayoutDataRecordGroups($scope.layoutDataRecordGroups);

                console.log("Layout Data Record Groups after Aggregation changed: ", $scope.layoutDataRecordGroups);

                // Retrieve the new feature table and visualize it. In the beginning, the dimension groups may not be available.
                checkParametersRetrieveDataAndVisualize();

              });

            };

            // distinguish between the different data types for the aggregation
            if(dataType === 'CATEGORICAL'){

              aggregationPreviewService.previewAggregationCategoricalDimension(dimensionName, allValuesOfDimension, function(){

                // success, apply the new aggregation
                $scope.aggregation = new AggregatorByUniqueValue({"_dimensionName": $scope.selectedDimensionForAggregation, "_dimensionType": "CATEGORICAL"});
                confirm();

              }, function(){
                // cancel, set the previous value of the aggregation.
                $scope.selectedDimensionForAggregation = previousSelectedDimensionForAggregation;
              })

            }else if(dataType === 'NUMERICAL'){

              aggregationPreviewService.previewAggregationNumericalDimension(dimensionName, allValuesOfDimension, function(histogram){

                // success, apply the new aggregation
                // histogram of type d3.histogram.
                // remove the histogram information which is not needed (=actual data values per bin, just keep x0, x1)
                let bins = [];
                for(let i=0; i<histogram.length; i++){
                  bins.push({'x0': histogram[i].x0, 'x1': histogram[i].x1})
                }
                $scope.aggregation = new AggregatorByEqualWidthBinning({"_dimensionName": $scope.selectedDimensionForAggregation, "_binning": bins});
                confirm();

              }, function(){
                // cancel, set the previous value of the aggregation.
                $scope.selectedDimensionForAggregation = previousSelectedDimensionForAggregation;
              })

            }
          });
      });
    };

    /**
     * This function is called, if the dimension of the second aggregation level is changed.
     */
    $scope.aggregation2ndLevelChanged = function(){

      let dimensionName = $scope.selectedDimensionForAggregation2ndLevel;

      // Function which is called once new data record groups need to be loaded.
      let updateAgg = function(){

        // Get the layout for the data record groups based on the new aggregation
        let tmp = dataService.getLayoutDataRecordGroupsByAggregation2Level().$promise;
        tmp.then(function(layoutDataRecordGroups){

          $scope.layoutDataRecordGroups = new LayoutDataRecordGroups(layoutDataRecordGroups);
          urlParameterService.setLayoutDataRecordGroups($scope.layoutDataRecordGroups);

          // Retrieve the new feature table and visualize it.
          retrieveByLayoutAndVisualize();

        });

      };

      if(dimensionName === undefined){
        // the second aggregation level has been removed by the user.

        previousSelectedDimensionFor2ndAggregation = undefined;
        urlParameterService.setAggregation2ndLevel();

        updateAgg();

      }else{
        // the second aggregation level has ben activated by the user.

        let dataType = undefined;

        let dataTypeTmp = dataService.getDataTypeOfDimension(dimensionName).$promise;
        dataTypeTmp.then(function(d){
          dataType = d.dimType;

          let allValuesOfDimensionTmp = dataService.getValuesOfDimension(dimensionName, dataType).$promise;
          allValuesOfDimensionTmp.then(function(allValuesOfDimension){

            // distinguish between the different data types for the aggregation
            if(dataType === 'CATEGORICAL'){

              aggregationPreviewService.previewAggregationCategoricalDimension(dimensionName, allValuesOfDimension, function(){

                // success, apply the new aggregation
                previousSelectedDimensionFor2ndAggregation = $scope.selectedDimensionForAggregation2ndLevel;

                $scope.aggregation2ndLevel = new AggregatorByUniqueValue({"_dimensionName": $scope.selectedDimensionForAggregation2ndLevel, "_dimensionType": "CATEGORICAL"});
                urlParameterService.setAggregation2ndLevel($scope.aggregation2ndLevel);
                updateAgg();

              }, function(){
                // cancel, set the previous value of the aggregation.
                $scope.selectedDimensionForAggregation2ndLevel = previousSelectedDimensionFor2ndAggregation;
              })

            }else if(dataType === 'NUMERICAL'){
              // cancel, set the previous value of the aggregation. Numerical dimensions are not available yet.
              $scope.selectedDimensionForAggregation2ndLevel = previousSelectedDimensionFor2ndAggregation;
            }

          });
        });
      }
    };

    /**
     * This function is called, once the layout of the glyph has been changed.
     * This change is reflected in the variable: '$scope.featureTableLayout.glyphLayout'
     * We need a new
     *  - FeatureTable
     *  - Layout of dimension groups (different sorting etc.)
     * from the server.
     */
    $scope.glyphLayoutChanged = function(){

      urlParameterService.setGlyphLayout($scope.glyphLayout);

      let featureTableTmp = dataService.getFeatureTableByLayout().$promise;
      featureTableTmp.then(function(featureTableData){
        $scope.featureTable = new FeatureTable(featureTableData);

        let layouttmp = dataService.getLayoutDimensionGroupsForGlyphLayout().$promise;
        layouttmp.then(function(dimensionLayoutData){

          $scope.layoutDimensionGroups = new LayoutDimensionGroups(dimensionLayoutData);
          urlParameterService.setLayoutDimensionGroups($scope.layoutDimensionGroups);

          // all data should be up to date.
          $scope.updateVisualization();

        })

      });

    };

    /**
     * This method is used once the colormap is changed or overlays are activated etc.
     * Whenever no re-computation of the feature table is necessary. Otherwise use $scope.glyphLayoutChanged.
     */
    $scope.overlayChanged = function(){

      urlParameterService.setGlyphLayout($scope.glyphLayout);
      $scope.updateVisualization();

    };

    /**
     * This method is called once the colormap is changed in the GUI.
     * @param newColormap the name of the new colormap
     * @param oldColormap the name of the old colormap.
     */
    $scope.colormapChanged = function(newColormap, oldColormap){

      if(newColormap.startsWith("Unified") && oldColormap.startsWith("Unified") || newColormap.startsWith("Bi") && oldColormap.startsWith("Bi")){
        $scope.overlayChanged();
      }else{
        $scope.glyphLayoutChanged();
      }
    };

    /**
     * Method is called once the statistics of all dimension(groups) is changed in the menu.
     * TODO Implement once computation by Steffi is available.
     */
    $scope.statisticsAllGroupsChanged = function(){
      console.error("$scope.statisticsAllGroupsChanged: NOT IMPLEMENTED YET.");
    };


    /**
     * The sorting of all dimension groups has changed.
     *  - Remove the customized sorting strategy.
     *  - Get new order from the server.
     */
    $scope.sortingAllGroupsChanged = function (orderId) {

      // Iterate over all dimension groups and set the sorting strategy.
      $scope.layoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();
      $scope.layoutDimensionGroups._dimensionGroups.forEach(function(d){
        d._sortingStrategy = new Sorting({'_sortOrder': orderId});
      });

      urlParameterService.setLayoutDimensionGroups($scope.layoutDimensionGroups);
      $scope.glyphLayoutChanged();
    };

    /**
     * The distance function of all dimension groups has changed.
     *  - Remove the customized sorting strategy.
     *  - Get new order from the server.
     */
    $scope.distanceFunctionAllGroupsChanged = function(distanceId){

      // Iterate over all dimension groups and set the sorting strategy.
      $scope.layoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();
      $scope.layoutDimensionGroups._dimensionGroups.forEach(function(d){
        d._dimensionDistance = new DimensionDistance({'_distanceFunction': distanceId})
      });

      urlParameterService.setLayoutDimensionGroups($scope.layoutDimensionGroups);
      $scope.glyphLayoutChanged();
    };

    /**
     * The normalization strategy of all dimension groups has changed.
     *  - Remove the customized sorting strategy.
     *  - Get new order from the server.
     */
    $scope.normalizationAllGroupsChanged = function(normalizationId){

      // Iterate over all dimension groups and set the sorting strategy.
      $scope.layoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();
      $scope.layoutDimensionGroups._dimensionGroups.forEach(function(d){
        d._normalizingStrategy = new Normalizing({'_normalizing': normalizationId})
      });

      urlParameterService.setLayoutDimensionGroups($scope.layoutDimensionGroups);
      $scope.glyphLayoutChanged();
    };

    /**
     * Register important listeners for the visualization.
     */
    let registerListeners = function(){

      $scope.tooltipService = tooltipService;

      // ----
      // listeners for tooltips
      $scope.$watch('tooltipService.dataTooltipCategoricalGlyph', function(){
        $scope.dataTooltipCategorical = tooltipService.dataTooltipCategoricalGlyph;
      }, true);

      $scope.$watch('tooltipService.dataTooltipNumericalGlyph', function(){
        $scope.dataTooltipNumerical = tooltipService.dataTooltipNumericalGlyph;
      }, true);

    };
    registerListeners();

    /**
     * @return {boolean} true, if the second aggregation level is enabled.
     */
    let isSecondLevelAggregationEnabled = function(){
      let currentLayout = urlParameterService.getLayoutDataRecordGroups();
      return currentLayout.maxLevelOfAggregation == 2;
    };

    /**
     * @return {Number} the currehtly used max aggregation level.
     */
    let getMaxAggregationLevel = function(){
      let currentLayout = urlParameterService.getLayoutDataRecordGroups();
      return currentLayout.maxLevelOfAggregation;
    };

    /**
     * @return {number} the number of groups in the second aggregation level. Return 1, if no second level is enabled.
     */
    let numberOfGroupsSecondLevel = function(){
      let currentLayout = urlParameterService.getLayoutDataRecordGroups();
      return currentLayout.getNumberOfRecordGroups2ndLevel();
    };


    /**
     * Remove the dimension group with the given name. Update the URL and repaint the visualization.
     * @param dimensionGroupName the name of the dimension group.
     */
    $scope.removeDimensionGroupVis = function (dimensionGroupName) {

      // delete GUI element from the screen
      d3.select("#dimensionGroup-" + dimensionGroupName).remove();

      let currentLayoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

      // find the index of the dimension group
      let pos = _.findIndex(currentLayoutDimensionGroups._dimensionGroups, function(d){
        return d._name === dimensionGroupName;
      });

      // remove the element from the array
      currentLayoutDimensionGroups._dimensionGroups.splice(pos, 1);

      // update the existing locations on the screen
      DimensionGroupVis.updatePositionOfGroups(currentLayoutDimensionGroups._dimensionGroups, getMaxAggregationLevel());

      // Update the url parameters.
      $scope.layoutDimensionGroups = currentLayoutDimensionGroups;
      urlParameterService.setLayoutDimensionGroups($scope.layoutDimensionGroups);
    };


    /**
     * Open windows to select the name of the group. Name needs to be unique.
     * This function is called from the GUI.
     * If name is unique, create the new group and movie it to the first position.
     */
    $scope.menuAddEmptyDimensionGroup = function(){

      let confirm = $mdDialog.prompt()
        .title('Please enter the name of the new group.')
        .placeholder('Name of Group')
        // .ariaLabel('Dog name')
        .initialValue(undefined)
        // .targetEvent(ev)
        .required(true)
        .ok('save')
        .cancel('cancel');

      $mdDialog.show(confirm).then(function(result) {

        // check if the name is unique, otherwise show error message.
        let allGroups = urlParameterService.getLayoutDimensionGroups()._dimensionGroups;
        for(let i=0; i<allGroups.length; i++){
          if(allGroups[i]._name === result.trim()){

            // show error message
            $mdDialog.show(
              $mdDialog.alert()
                .parent(angular.element(document.querySelector('#popupContainer')))
                .clickOutsideToClose(true)
                .title('Error')
                .textContent("The group name '" + result + "' already exists.")
                .ok('Got it!')
            );

            return;
          }
        }
        $scope.addEmptyDimensionGroup(result);
      }, function() {
        // cancel, no action required.
      });

    };

    /**
     * Remove the selection of any dimension.
     * This function is called from the GUI.
     */
    $scope.menuRemoveSelection = function () {
      selectionService.removeAllSelections();
      selectionService.updateSelectionDimensions();
      selectionService.updateSelectedDataRecordGroups();

      // remove slider for similarity based highlighting.
      $scope.highlight = undefined;
    };

    /**
     * Add an empty dimension group.
     * @param groupName the name of the new group. Assumes the group name is unique.
     */
    $scope.addEmptyDimensionGroup = function(groupName){

      let currentLayoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();
      let currentGlyphLayout = urlParameterService.getGlyphLayout();
      let currentRecordGroupLayout = urlParameterService.getLayoutDataRecordGroups();

      // Check the normalization, distance function etc.
      let _normalizingStrategy = $scope.normalizationAllDimensionGroups ? $scope.normalizationAllDimensionGroups :  $scope.availableNormalizingStrategies[0].id;
      let _sortingStrategy = $scope.sortingOrderAllDimensionGroups ? $scope.sortingOrderAllDimensionGroups : $scope.availableSortings[0].id;
      let _dimensionDistance = $scope.distanceFunctionsAllDimensionGroups ? $scope.distanceFunctionsAllDimensionGroups : $scope.availableDistanceFunctions[0].id;

      let newGroup = new DimensionGroup({'_name': groupName,
        '_dimensions':[],
        '_normalizingStrategy' : _normalizingStrategy,
        '_sortingStrategy' : _sortingStrategy,
        '_dimensionDistance' : _dimensionDistance});
      currentLayoutDimensionGroups._dimensionGroups.unshift(newGroup);
      $scope.layoutDimensionGroups = currentLayoutDimensionGroups;
      urlParameterService.setLayoutDimensionGroups($scope.layoutDimensionGroups);

      // add the group to the svg element.
      let parentSvgGroup = d3.select("#svgGroupAllDimensionGroups");
      let g = new DimensionGroupVis(parentSvgGroup, groupName, tooltipService, $scope.featureTable, currentGlyphLayout, currentRecordGroupLayout, currentLayoutDimensionGroups, statisticsService).group;
      g.attr("groupName", groupName);
      g.attr("transform","translate(" + 0 + ", " + 0 + ")");

      // Drag and Drop not enabled yet. After adding dimensins, the visualization will be repainted anyway.
      // addDragAndDropToAllDimensionGroups(urlParameterService);

      // update the existing locations on the screen
      DimensionGroupVis.updatePositionOfGroups(currentLayoutDimensionGroups._dimensionGroups, getMaxAggregationLevel());

    };

    /**
     * Move the selected dimensions to the given dimension group.
     */
    $scope.menuMoveSelectedDimensionsTo = function(dimensionGroup){

      let currentLayoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

      // Get the dimension group in which the method was executed.
      let groupName = d3.select("#dimensionGroupMenu")
        .attr("dimensionGroup");
      let dimGroupToDeleteFrom = _.findWhere(currentLayoutDimensionGroups._dimensionGroups, {_name:groupName});

      // The dimensions which has been selected.
      let dimensions = selectionService.getSelectedDimensions();

      // Remove the selected dimensions from the current group and add to the new group.
      for(let i=0; i<dimensions.length; i++){
        dimGroupToDeleteFrom._dimensions = _.without(dimGroupToDeleteFrom._dimensions, dimensions[i]);
        dimensionGroup._dimensions.unshift(dimensions[i]);
      }

      // Store the dimension group in the layout.
      currentLayoutDimensionGroups.replaceGroupWithName(groupName, dimGroupToDeleteFrom);
      currentLayoutDimensionGroups.replaceGroupWithName(dimensionGroup._name, dimensionGroup);

      // update the selection.
      selectionService.removeAllSelectedDimensions();

      // update the parameters in the URL
      urlParameterService.setLayoutDimensionGroups(currentLayoutDimensionGroups);
      $scope.layoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

      // recompute the sorting and layout.
      $scope.glyphLayoutChanged();
    };

    /**
     * Move the selected dimensions to a different dimension group.
     */
    $scope.menuCopySelectedDimensionsTo = function(dimensionGroup){

      let currentLayoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

      // The dimensions which has been selected.
      let dimensions = selectionService.getSelectedDimensions();

      // Add the selected dimensions to the new group.
      for(let i=0; i<dimensions.length; i++){
        dimensionGroup._dimensions.unshift(dimensions[i]);
      }

      currentLayoutDimensionGroups.replaceGroupWithName(dimensionGroup._name, dimensionGroup);

      // update the selection.
      selectionService.removeAllSelectedDimensions();

      // update the parameters in the URL
      urlParameterService.setLayoutDimensionGroups(currentLayoutDimensionGroups);
      $scope.layoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

      // recompute the sorting and layout.
      $scope.glyphLayoutChanged();

    };

    /**
     * Set the sorting of the current dimension group.
     */
    $scope.setSortingOfGroupTo = function(sortingId){

      let currentLayoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

      // Get the dimension group in which the method was executed.
      let groupName = d3.select("#dimensionGroupMenu")
        .attr("dimensionGroup");
      let dimGroup = _.findWhere(currentLayoutDimensionGroups._dimensionGroups, {_name:groupName});

      console.log("dimGroup", dimGroup);

      // set the sorting
      dimGroup._sortingStrategy = new Sorting({'_sortOrder' : sortingId});

      console.log("dimGroup", dimGroup);

      currentLayoutDimensionGroups.replaceGroupWithName(dimGroup._name, dimGroup);

      console.log("currentLayoutDimensionGroups", currentLayoutDimensionGroups);

      // update the parameters in the URL
      urlParameterService.setLayoutDimensionGroups(currentLayoutDimensionGroups);
      $scope.layoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

      // recompute the sorting and layout.
      $scope.glyphLayoutChanged();

      // modify the sorting in the GUI.
      checkSortingInGroups();


    };

    /**
     * Set the normalization of the current dimension group.
     */
    $scope.setNormalizationOfGroupTo = function(normalizationStrategy){

      let currentLayoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

      // Get the dimension group in which the method was executed.
      let groupName = d3.select("#dimensionGroupMenu")
        .attr("dimensionGroup");
      let dimGroup = _.findWhere(currentLayoutDimensionGroups._dimensionGroups, {_name:groupName});

      // set the sorting
      dimGroup._normalizingStrategy = new Normalizing({'_normalizing' : normalizationStrategy});

      currentLayoutDimensionGroups.replaceGroupWithName(dimGroup._name, dimGroup);

      // update the parameters in the URL
      urlParameterService.setLayoutDimensionGroups(currentLayoutDimensionGroups);
      $scope.layoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

      // recompute the sorting and layout.
      $scope.glyphLayoutChanged();

      // modify the sorting in the GUI.
      checkNormalizationInGroups();

    };

    /**
     * Set the distance measure of the current dimension group.
     */
    $scope.setDistanceMeasureOfGroupTo = function(distanceMeasure){

      let currentLayoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

      // Get the dimension group in which the method was executed.
      let groupName = d3.select("#dimensionGroupMenu")
        .attr("dimensionGroup");
      let dimGroup = _.findWhere(currentLayoutDimensionGroups._dimensionGroups, {_name:groupName});

      // set the sorting
      dimGroup._dimensionDistance = new DimensionDistance({'_distanceFunction' : distanceMeasure});

      currentLayoutDimensionGroups.replaceGroupWithName(dimGroup._name, dimGroup);

      // update the parameters in the URL
      urlParameterService.setLayoutDimensionGroups(currentLayoutDimensionGroups);
      $scope.layoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

      // recompute the sorting and layout.
      $scope.glyphLayoutChanged();

      // modify the sorting in the GUI.
      checkDistanceMeasuresInGroups();

    };

    /**
     * Delete the selected dimensions from the group which is currently hovered..
     * Recomputation of sorting is NOT necessary; however, we have to adapt the values for the ordering in the field ...
     */
    $scope.menuDeleteSelectedDimensions = function () {

      // get the current dimension groups
      let currentLayoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

      // get the selected dimensions
      let dimensions = selectionService.getSelectedDimensions();

      // get the dimension group in which the method was executed.
      let groupName = d3.select("#dimensionGroupMenu")
        .attr("dimensionGroup");
      let dimGroup = _.findWhere(currentLayoutDimensionGroups._dimensionGroups, {_name : groupName});

      // remove all dimensions from the group
      for(let i=0; i<dimensions.length; i++){
        dimGroup._dimensions = _.without(dimGroup._dimensions, dimensions[i]);
      }

      console.log("dimGroup", dimGroup);

      // update the similarity position
      let sortableArray = [];
      _.keys(dimGroup._dimensionToSimilarityPos).forEach(function(dimName){
        if(!_.contains(dimensions, dimName)){
          sortableArray.push({'dimName' : dimName, 'pos' : dimGroup._dimensionToSimilarityPos[dimName]});
        }
      });
      sortableArray = _.sortBy(sortableArray, function(d){return d.pos});
      let nextPos = 0;
      let newPos = {};
      sortableArray.forEach(function(d){
        newPos[d.dimName] = nextPos++;
      });

      dimGroup._dimensionToSimilarityPos = newPos;

      // Store the dimension group in the layout.
      // currentLayoutDimensionGroups.replaceGroupWithName(groupName, dimGroup);

      // update the parameters in the URL
      urlParameterService.setLayoutDimensionGroups(currentLayoutDimensionGroups);
      $scope.layoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

      // update the selection.
      selectionService.removeAllSelectedDimensions();

      // update the visualization, recomputation not necessary.
      $scope.updateVisualization();

    };

    /**
     * Delete the selected data record groups.
     */
    $scope.menuDeleteSelectedDataRecordGroup = function(){

      // get the current record groups
      let currentLayoutRecordGroups = urlParameterService.getLayoutDataRecordGroups();

      // get the selected record groups
      let recordGroups = selectionService.getSelectedDataRecordGroups();

      // remove all record groups
      recordGroups.forEach(function(g){
        let posInArray = _.findIndex(currentLayoutRecordGroups._dataRecordGroups, function(d){return d._name === g;})
        currentLayoutRecordGroups._dataRecordGroups.splice(posInArray, 1);
      });

      // update the parameters in the URL
      urlParameterService.setLayoutDataRecordGroups(currentLayoutRecordGroups);
      $scope.layoutDataRecordGroups = urlParameterService.getLayoutDataRecordGroups();

      // update the selection
      selectionService.removeAllSelectedDataRecordGroups();

      // update the visualization; normalization + sorting is affected and needs to be recomputed.
      $scope.glyphLayoutChanged();

    };

    /**
     * Merge the selected data record groups.
     */
    $scope.menuMergeSelectedDataRecordGroup = function(){

      // Show dialog to select the name of the merged data record group
      let confirm = $mdDialog.prompt()
        .title('Please enter the name of the new record group.')
        .placeholder('Name of Group')
        .initialValue(undefined)
        // .targetEvent(ev)
        .required(true)
        .ok('merge')
        .cancel('cancel');

      $mdDialog.show(confirm).then(function(result) {

        console.log("new label: ", result);

        // get the current record groups
        let currentLayoutRecordGroups = urlParameterService.getLayoutDataRecordGroups();

        // get the selected record groups
        let recordGroupNames = selectionService.getSelectedDataRecordGroups();
        let recordGroups = [];
        let lastIndex = 0;
        recordGroupNames.forEach(function(g){
          lastIndex = _.findIndex(currentLayoutRecordGroups._dataRecordGroups, function(d){return d._name === g;})
          recordGroups.push(currentLayoutRecordGroups._dataRecordGroups[lastIndex]);

          // remove the groups which will be merged.
          currentLayoutRecordGroups._dataRecordGroups.splice(lastIndex, 1);
        });

        let newRecordGroup = DataRecordGroup.mergeGroups(result, recordGroups);

        // add the new record group
        currentLayoutRecordGroups._dataRecordGroups.splice(lastIndex, 0, newRecordGroup);

        console.log("newRecordGroup", newRecordGroup);
        console.log("currentLayoutRecordGroups", currentLayoutRecordGroups);

        // update the parameters in the URL
        urlParameterService.setLayoutDataRecordGroups(currentLayoutRecordGroups);
        $scope.layoutDataRecordGroups = urlParameterService.getLayoutDataRecordGroups();

        // update the selection
        selectionService.removeAllSelectedDataRecordGroups();

        // update the visualization; normalization + sorting is affected and needs to be recomputed.
        $scope.glyphLayoutChanged();


      }, function() {
        // cancel, no action required.
      });


    };

    /**
     * Sort the dimensions in the current group by similarity to the currently selected dimension
     */
    $scope.menuSortBySelectedDimensionWithinGroup = function(){

      // get the current dimension groups
      let currentLayoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

      // The dimension group which is currently selected.
      let groupName = d3.select("#dimensionGroupMenu").attr("dimensionGroup");
      let dimGroup = _.findWhere(currentLayoutDimensionGroups._dimensionGroups, {_name : groupName});

      // the selected dimension; the GUI ensures that there is only one selected dim possible.
      let dimension = selectionService.getSelectedDimensions()[0];

      // change the sort order.
      dimGroup._sortingStrategy = new Sorting({'_sortOrder' : 'bySimilarDimension', '_dimension' : dimension});

      // Store the dimension group in the layout.
      currentLayoutDimensionGroups.replaceGroupWithName(groupName, dimGroup);

      // update the parameters in the URL
      urlParameterService.setLayoutDimensionGroups(currentLayoutDimensionGroups);
      $scope.layoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();

      // update the selection.
      selectionService.removeAllSelectedDimensions();

      // recompute the sorting and layout.
      $scope.glyphLayoutChanged();

      // Update the selected sorting strategy in the GUI.
      checkSortingInGroups();

    };

    /**
     * Highlight the most similar dimensions to the currently selected one (considering only dimensions of the current dimension group)
     */
    $scope.menuHighlightMostSimilarDimensionsWithinGroup = function(){

      // the selected dimension; the GUI ensures that there is only one selected dim possible.
      let dimension = selectionService.getSelectedDimensions()[0];

      // the current dimension group.
      let groupName = d3.select("#dimensionGroupMenu").attr("dimensionGroup");

      // configure the similarity based sorting.
      let config = new SimilarityBasedHighlightingConfigureDto({'_acrossGroups' : false, '_dimension' : dimension, '_nameOfDimensionGroup' : groupName});

      let tmp = dataService.getSimilarityOfDimensionsTo(config).$promise;
      tmp.then(function(similarityData){

        // this sets also the slider on the gui as visible.
        $scope.highlight =  new SimilarityBasedHighlightingDto(similarityData);

        // see $scope.$watch(function () {return $scope.highlight}

      })

    };

    // Belongs to $scope.menuHighlightMostSimilarDimensionsWithinGroup
    // Once the highlight changed (e.g., also when the slider changed, we need to update highlighted dimensions.
    $scope.$watch(function () {return $scope.highlight}, function () {
      if($scope.highlight){

        // now highlight the recommended number of dimensions
        selectionService.removeAllSelectedDimensions();
        for(let i=0; i<$scope.highlight._threshold; i++){
          selectionService.addSelectedDimension($scope.highlight._sortedDimensions[i]);
        }
      }
    }, true);

    /**
     * Highlight the most similar dimensions to the currently selected one (considering all dimensions in all dimension groups)
     */
    $scope.menuHighlightMostSimilarDimensionsAcrossGroups = function(){

      // TODO not implemented: problem: different normalization strategies across all groups.

      // Check if the distance measure is identical for all dimension groups.
      // If not, show an error message.
      // checkDistanceMeasuresInGroups();
      // if($scope.distanceFunctionsAllDimensionGroups == undefined){
      //
      //   $mdDialog.show(
      //
      //     $mdDialog.alert()
      //       // .parent(angular.element(document.querySelector('#popupContainer')))
      //       .clickOutsideToClose(true)
      //       .title('Error')
      //       .textContent('Different distance functions are selected for the dimension groups. Please use the menu and select the same function for all groups.')
      //       // .ariaLabel('Alert Dialog Demo')
      //       .ok("Ok, I'll fix it!")
      //       //.targetEvent(ev)
      //   );
      //
      //   return;
      // }

      // the selected dimension; the GUI ensures that there is only one selected dim possible.
      // let dimension = selectionService.getSelectedDimensions()[0];

      // configure the similarity based sorting.
      // let config = new SimilarityBasedHighlightingConfigureDto({'_acrossGroups' : true, '_dimension' : dimension, '_nameOfDimensionGroup' : undefined});


    };



  });

'use strict';

// TODO can be deleted once it is implemented on the server.
/** Example dataset, no missing values. */
// let datasetExample1 = new DatasetInfo({_id: "example-1", _datasetName : "Example Dataset (without missing values)",    _databaseName : "EXAMPLE-NO-MISSING-VALUES"});
/** Example dataset, with missing values. */
// let datasetExample2 = new DatasetInfo({_id: "example-2", _datasetName : "Example Dataset (including missing values)",  _databaseName : "EXAMPLE-WITH-MISSING-VALUES"});

/**
 * @ngdoc function
 * @name smartexploreApp.controller:DatasetCtrl
 * @description
 * # DatasetCtrl
 * Controller of the smartexploreApp
 */
angular.module('smartexploreApp')
  .controller('DatasetCtrl', function ($scope, dataService, urlParameterService, $q, $route, $location) {

    // console.log("Dataset Controller activated");


    /** All available datasets on the server. */
    $scope.availableDatasets = {};

    /** The dataset which is currently used. Type DatasetInfo; Bound to the view. */
    $scope.selectedDataset = null;

    // =======================================================================
    // Load all data that is required to show this page - independent of the current selection.
    let initDataLoad = function(){

      let allDatasetsTmp = dataService.getAllAvailableDatasets().$promise;

      // wait until all calls are done
      $q.all({allDatasetsTmp}).then(data => {

        _.each(data.allDatasetsTmp, function (json) {
          let info = new DatasetInfo(json);
          $scope.availableDatasets[info.id] = info;
        });

        // check if the URL parameters contain and information about the currently used dataset
        let idOfDataset = urlParameterService.getDatabaseId();

        if(idOfDataset === undefined){
          // no dataset selected yet - do not choose any specific dataset.
          $scope.selectedDataset = null;
        }else{
          // find the dataset with the id in the list of available datasets
          $scope.selectedDataset = $scope.availableDatasets[idOfDataset];
        }

        // console.log("idOfDataset",idOfDataset);
        // console.log("selectedDataset",$scope.selectedDataset);
        // console.log("availableDatasets", $scope.availableDatasets, $scope.availableDatasets[idOfDataset]);

      })

    };

    initDataLoad();



    /**
     * Method which will be called, once the dataset is changed.
     * The selected dataset is stored in $scope.selectedDataset.
     * All url parameters are deleted
     */
    $scope.switchDataset = function () {

      // don't know why we need to delete this field
      // delete $scope.selectedDataset['$$hashKey'];

      urlParameterService.resetAllParameters();
      urlParameterService.setDatabaseId($scope.selectedDataset._databaseName);

      // put dataset on the server side.
      // let t = dataService.putActiveDataset(JSON.stringify($scope.selectedDataset));
      // t.$promise.then(function(success){

        // let d = dataService.getActiveDataset();
        // d.$promise.then(function(datasetInfoJson){
        //   $scope.activeDataset = new DatasetInfo(datasetInfoJson);
        // });

        // change the dataset in the URL paramters
        // $location.search(UrlParameters.DB_NAME, $scope.selectedDataset._databaseName);


      // }, function(error){

        // console.log(error);
        // console.log("Error: " + error.message);

      // });
    };

    $scope.uploadDatasetFiles = function () {

      let datasetFile = document.getElementById('dataset-file').files[0];

      // check for a valid dataset name
      if ($scope.uploadDatasetName === undefined || $scope.uploadDatasetName === undefined || $scope.uploadDatasetName.length === 0) {
        alert("Please provide a name for the dataset.");
        return;
      }

      let putUploadDatasetResource = dataService.putUploadDataset(datasetFile, $scope.uploadDatasetName);
      putUploadDatasetResource.$promise.then(function () {

        // reload the page, do not select any datasets.
        $location.search(UrlParameters.DB_NAME, null);

        // hard reload.
        $route.reload();

      });
    };

    /** Init function. Retrieves the currently used dataset from the server and sets the correct properties. */
    // let init = function(){
    //
    //   // Retrieve the currently used dataset.
    //   let d = dataService.getActiveDataset();
    //   d.$promise.then(function(datasetInfoJson){
    //     $scope.activeDataset = new DatasetInfo(datasetInfoJson);
    //   });
    // };
    //
    // init();

  });

'use strict';

/**
 * Service to communicate between the server and the client.
 * Communication is done via REST.
 *
 * For some REST calls are done with POST although GET would be the right choice.
 * POST is done with $resource(...).save().
 * Reason: We store all information in the URL. The parameter can be very large. For GET there is a limit of about 4000 characters. Therefore, we decided for POST.
 *
 * @param $resource to make REST calls to the server.
 * @param $location access the URL parameters in the application.
 */
angular.module('smartexploreApp')
  .service('dataService', function ($resource, $location, urlParameterService) {

    // set the correct base url for the rest calls.
    this.baseUrl = urlParameterService.getBaseUrl();
    let self = this;

    /** Retrieve information about the app such as name and version. */
    this.getAppInfo = function () {
      return $resource(this.baseUrl + 'api/info').get();
    };

    /** Retrieves the currently used dataset from the server. Uses the current url parameters as parameters */
    this.getDataset = function () {
      let urlParam = $location.search();
      let param = {};
      param["dbName"] = urlParam[UrlParameters.DB_NAME];
      return $resource(this.baseUrl + 'api/dataset/selection').get(param);
    };

    /** Retrieves all available datasets from the server.
     * @return array. */
    this.getAllAvailableDatasets = function(){
      return $resource(this.baseUrl + 'api/dataset/getAll').query();
    };

    /** Upload a dataset file to the server. */
    this.putUploadDataset = function (datasetFile, datasetName) {
      let formData = new FormData();
      formData.append("dataset-file", datasetFile);
      formData.append("dataset-name", datasetName);
      let putResource = $resource(this.baseUrl + 'api/dataset/upload', null, {
        put: {
          method: 'PUT',
          params: formData,
          transformRequest: angular.identity,
          headers: { 'Content-Type': undefined }
        }
      });
      return putResource.put(formData);
    };

    /**
     * @returns all dimensions of the current dataset, except for the dimension with the id.
     */
    this.getAllDimensionsOfDataset = function(){
      let urlParam = $location.search();
      let param = {};
      param["dbName"] = urlParam[UrlParameters.DB_NAME];
      return $resource(this.baseUrl + 'api/dataset/dimensions').query(param);
    };

    /**
     * Returns a resource which returns the feature table for the visualization.
     * The parameters 'param' can be specified as follows:
     *  parameters: all url parameters (aggregation, etc.)
     */
    this.getDefaultFeatureTable = function(){
      return $resource(this.baseUrl + 'api/featuretable/aggregator/default').save($location.search());
    };

    /**
     * Returns a resource which returns the feature table for the visualization.
     * The parameters 'param' can be specified as follows:
     *  parameters: all url parameters (aggregation, etc.)
     */
    this.getFeatureTableByLayout = function(){
      return $resource(this.baseUrl + 'api/featuretable/custom').save($location.search());
    };

    /**
     * Returns a resource which returns the default layout for the data record groups (rows in the visualization) in the visualization.
     */
    this.getDefaultLayoutDataRecordGroups = function(){
      return $resource(this.baseUrl + 'api/featuretable/datarecordgroups/default').save($location.search());
    };

    /**
     * Returns a resource which returns the layout for the data record groups (rows in the visualization) for the defined aggregation level.
     */
    this.getLayoutDataRecordGroupsByAggregation = function(){
      return $resource(this.baseUrl + 'api/featuretable/datarecordgroups/aggregation').save($location.search());
    };

    /**
     * Returns a resource which returns the layout for the data record groups (rows in the visualization) for the defined aggregation level.
     */
    this.getLayoutDataRecordGroupsByAggregation2Level = function(){
      return $resource(this.baseUrl + 'api/featuretable/datarecordgroups/aggregation/level2').save($location.search());
    };

    /**
     * Returns a resource which returns the default layout for the dimension groups (multiple rows in the visualization)
     */
    this.getDefaultLayoutDimensionGroups = function(){
      return $resource(this.baseUrl + 'api/featuretable/dimensiongroups/default').save($location.search());
    };

    /**
     * Returns a resource which returns the layout for the dimension groups based on the Glyph Layout.
     * See FeatureTableResource#getDimensionGroupsForGlyphLayout.java.
     */
    this.getLayoutDimensionGroupsForGlyphLayout = function(){
      return $resource(this.baseUrl + 'api/featuretable/dimensiongroups').save($location.search());
    };

    /**
     * Returns the data type of the given dimension
     * @param dimName the name of the dimension.
     */
    this.getDataTypeOfDimension = function(dimName){
      let urlParam = $location.search();
      let param = {};
      param["dimensionName"] = dimName;
      param["dbName"] = urlParam[UrlParameters.DB_NAME];
      return $resource(this.baseUrl + 'api/dataset/dimension/type').get(param);
    };

    /**
     * Returns all values of the given dimension.
     * @param dimName the name of the dimension.
     * @param dimType the type of the dimension. {CATEGORICAL, NUMERICAL}
     */
    this.getValuesOfDimension = function(dimName, dimType){

      let urlParam = $location.search();
      let param = {};
      param["dimensionName"] = dimName;
      param["dbName"] = urlParam[UrlParameters.DB_NAME];

        if(dimType === 'CATEGORICAL'){
          return $resource(self.baseUrl + 'api/dataset/dimension/values/categorical').query(param);

        }else if(dimType === 'NUMERICAL'){
          return $resource(self.baseUrl + 'api/dataset/dimension/values/numerical').query(param);

        }else{
          return null;
        }

    };

    /**
     * Returns the similarity based highlighting for the given parameters.
     * @param config of type SimilarityBasedHighlightingConfigureDto.js
     * @return {Array|*|{method}}
     */
    this.getSimilarityOfDimensionsTo = function(config){

      let params = $location.search();
      params['config'] = angular.toJson(config);

      return $resource(this.baseUrl + 'api/featuretable/dimensions/similarity').save(params);
    }


  });

'use strict';

/**
 * @ngdoc service
 * @description
 * # labelService
 *
 * Service to compute statistics such as size of labels.
 *
 *
 */
angular.module('smartexploreApp')
  .service('labelService', function () {

    const DEFAULT_FONT = 'Lato';

    /**
     * Returns the width of the given string considering the following font size, and optionally the given font.
     * @param string the strign to be measured
     * @param fontSize the size of the font in px
     * @param font the font, optional
     */
    this.getWidthOfString = function(string, fontSize, font=DEFAULT_FONT){

      // convert the font size to the right format.
      if(!fontSize.endsWith("px")){
        fontSize = fontSize + "px";
      }

      let canvas = document.createElement('canvas');
      let ctx = canvas.getContext("2d");
      ctx.font = fontSize + font;
      return ctx.measureText(string).width;
    }

  });




'use strict';

/**
 * @ngdoc service
 * @description
 * # labelService
 *
 * Service to show and hide tooltips.
 *
 *
 */
angular.module('smartexploreApp')
  .service('tooltipService', function ($rootScope, urlParameterService, statisticsService) {

    /** Id of the tooltip of the numerical glyph (directive) */
    const ID_TOOLTIP_NUMERICAL_GLYPH = '#tooltipNumericalGlyph';
    /** Id of the tooltip of the categorical glyph (directive) */
    const ID_TOOLTIP_CATEGORICAL_GLYPH = '#tooltipCategoricalGlyph';

    /** Data of the tooltip for the categorical glyph. */
    this.dataTooltipCategoricalGlyph = {
      'dimName': undefined,
      'distanceToAverage' : undefined,
      'numMissingValues' : undefined,
      'percentageMissingValues' : undefined,
      'numRecords' : undefined,
      'measure' : undefined,
      'usedMeasure' : undefined,
      'uniqueValues' : undefined,
      'distribution' : undefined,
      'dimensionDistribution' : undefined
    };

    /** Data of the tooltip for the numerical glyph. */
    this.dataTooltipNumericalGlyph = {
      'dimName': undefined,

      'min': undefined,
      'max': undefined,
      'mean': undefined,
      'median': undefined,
      'standardDeviation': undefined,
      'variance': undefined,

      'minDimension': undefined,
      'maxDimension': undefined,
      'meanDimension': undefined,
      'medianDimension': undefined,
      'standardDeviationDimension': undefined,
      'varianceDimension': undefined,

      'numMissingValues' : undefined,
      'percentageMissingValues' : undefined,

      'numRecords' : undefined, /* missing values are ignored. */
      'measure' : undefined,
      'usedMeasure' : undefined,
      'distanceToMean' : undefined,
    };

    let self = this;

    /**
     * Show the tooltip for the categorical glyph.
     * @param pos array with [x,y].
     * @param dimension data about the whole dimension.
     * @param categoricalGlyphData the data of the glyph.
     * @param recordGroupName the name of the record group.
     */
    this.showCategoricalGlyphTooltip = function(pos, dimension, categoricalGlyphData, recordGroupName){

      // update the data
      $rootScope.$apply(function(){
        self.dataTooltipCategoricalGlyph['recordGroupName'] = recordGroupName;
        self.dataTooltipCategoricalGlyph['dimName'] = dimension.dimName;
        self.dataTooltipCategoricalGlyph['distanceToAverage'] = Number(categoricalGlyphData['distanceToAverage']).toFixed(2);
        self.dataTooltipCategoricalGlyph['numMissingValues'] = categoricalGlyphData['missingValues'];
        self.dataTooltipCategoricalGlyph['percentageMissingValues'] = categoricalGlyphData['percentageMissingValues'];
        self.dataTooltipCategoricalGlyph['numRecords'] = categoricalGlyphData['allValues'].length;
        self.dataTooltipCategoricalGlyph['measure'] = Number(categoricalGlyphData['measure']).toFixed(2);
        self.dataTooltipCategoricalGlyph['usedMeasure'] = dimension['measure'];

        self.dataTooltipCategoricalGlyph['uniqueValues'] = dimension['globalUniqueValues'];
        self.dataTooltipCategoricalGlyph['distribution'] = categoricalGlyphData['distribution'];
        self.dataTooltipCategoricalGlyph['dimensionDistribution'] = dimension['globalDistribution'];

        // console.log("categoricalGlyphData", categoricalGlyphData);

        let statInfo = statisticsService.categoricalValuePandEffect(categoricalGlyphData);
        self.dataTooltipCategoricalGlyph['pValue'] = Number(statInfo._pValue);
        self.dataTooltipCategoricalGlyph['effectSize'] = Number(statInfo._effectSize);

      });

      // show the tooltip
      d3.select(ID_TOOLTIP_CATEGORICAL_GLYPH)
        .style("left", (pos[0] + 10) + "px")
        .style("top", (pos[1] + 10) + "px")
        .style("display", "inline");
    };

    this.showNumericalGlyphTooltip = function(pos, dimension, numericalGlyphData, recordGroupName){

      // update the data
      $rootScope.$apply(function(){

        self.dataTooltipNumericalGlyph['recordGroupName'] = recordGroupName;
        self.dataTooltipNumericalGlyph['dimName'] = dimension.dimName;
        self.dataTooltipNumericalGlyph['numMissingValues'] = numericalGlyphData['missingValues'];
        self.dataTooltipNumericalGlyph['percentageMissingValues'] = numericalGlyphData['percentageMissingValues'];
        self.dataTooltipNumericalGlyph['numRecords'] = numericalGlyphData['allValues'].length;

        self.dataTooltipNumericalGlyph['measure'] = Number(numericalGlyphData['measure']).toFixed(2);
        self.dataTooltipNumericalGlyph['usedMeasure'] = dimension['measure'];
        self.dataTooltipNumericalGlyph['distanceToMean'] = Number(numericalGlyphData['distanceToMean']).toFixed(2);

        self.dataTooltipNumericalGlyph['min'] = Number(numericalGlyphData['min']).toFixed(2);
        self.dataTooltipNumericalGlyph['max'] = Number(numericalGlyphData['max']).toFixed(2);
        self.dataTooltipNumericalGlyph['mean'] = Number(numericalGlyphData['mean']).toFixed(2);
        self.dataTooltipNumericalGlyph['median'] = Number(numericalGlyphData['median']).toFixed(2);
        self.dataTooltipNumericalGlyph['standardDeviation'] = Number(numericalGlyphData['standardDeviation']).toFixed(2);
        self.dataTooltipNumericalGlyph['variance'] = Number(numericalGlyphData['variance']).toFixed(2);

        self.dataTooltipNumericalGlyph['minDimension'] = Number(dimension['globalMin']).toFixed(2);
        self.dataTooltipNumericalGlyph['maxDimension'] = Number(dimension['globalMax']).toFixed(2);
        self.dataTooltipNumericalGlyph['meanDimension'] = Number(dimension['globalMean']).toFixed(2);
        self.dataTooltipNumericalGlyph['medianDimension'] = Number(dimension['globalMedian']).toFixed(2);
        self.dataTooltipNumericalGlyph['standardDeviationDimension'] = Number(dimension['globalStandardDeviation']).toFixed(2);
        self.dataTooltipNumericalGlyph['varianceDimension'] = Number(dimension['globalVariance']).toFixed(2);

        self.dataTooltipNumericalGlyph['distribution'] = numericalGlyphData['allValues'];
        self.dataTooltipNumericalGlyph['distributionDimension'] = dimension['globalAllValues'];

        // console.log("numericalGlyphData", numericalGlyphData);

        let statInfo = statisticsService.numericalValuePandEffect(numericalGlyphData);
        self.dataTooltipNumericalGlyph['pValue'] = Number(statInfo._pValue);
        self.dataTooltipNumericalGlyph['effectSize'] = Number(statInfo._effectSize);

      });

      // show the tooltip
      d3.select(ID_TOOLTIP_NUMERICAL_GLYPH)
        .style("left", (pos[0] + 10) + "px")
        .style("top", (pos[1]  + 10) + "px")
        .style("display", "inline");
    };

    /**
     * Hide the tooltip for the categorical glyph.
     */
    this.hideCategoricalGlyphTooltip = function(){
      d3.select(ID_TOOLTIP_CATEGORICAL_GLYPH).style("display", "none");
    };

    /**
     * Hide the tooltip for the numerical glyph.
     */
    this.hideNumericalGlyphTooltip = function(){
      d3.select(ID_TOOLTIP_NUMERICAL_GLYPH).style("display", "none");
    };

    /**
     * Returns the width of the given string considering the following font size, and optionally the given font.
     * @param string the strign to be measured
     * @param fontSize the size of the font in px
     * @param font the font, optional
     */
    this.getWidthOfString = function(string, fontSize, font=DEFAULT_FONT){

      // convert the font size to the right format.
      if(!fontSize.endsWith("px")){
        fontSize = fontSize + "px";
      }

      let canvas = document.createElement('canvas');
      let ctx = canvas.getContext("2d");
      ctx.font = fontSize + font;
      return ctx.measureText(string).width;
    }

  });




'use strict';

/**
 * @ngdoc service
 * @description
 * # dataService
 *
 * Service to modify and retrieve properties from the url
 *
 */
angular.module('smartexploreApp')
  .service('urlParameterService', function ($location) {

    const APP_NAME = 'SMARTexplore';

    /**
     * Returns the base url of the application so that REST calls can be accessed from the tomcat server for an
     * arbitrary name of the application.
     */
    this.getBaseUrl = function(){

      function isTopLevelDomain() {
        return absUrl.indexOf(APP_NAME) === -1 && absUrl.indexOf(APP_NAME.toLowerCase()) === -1;
      }

      // Determine whether the app runs in a the top-level-domain or not and if not extract the name
      let absUrl = $location.absUrl();
      if (!isTopLevelDomain()) {
        return '/' + absUrl.split('/')[3] + '/';
      }else{
        return '/';
      }

    };

    /**
     * Resets all parameters in the url to undefined.
     */
    this.resetAllParameters = function(){
      $location.search(UrlParameters.DB_NAME, null);
      $location.search(UrlParameters.AGGREGATION_OF_FEATURE_TABLE, null);
      $location.search(UrlParameters.AGGREGATION_OF_FEATURE_TABLE_SECOND_LEVEL, null);
      $location.search(UrlParameters.GLYPH_LAYOUT, null);
      $location.search(UrlParameters.SORTING, null);
      $location.search(UrlParameters.LAYOUT_DATA_RECORD_GROUPS, null);
      $location.search(UrlParameters.LAYOUT_DIMENSION_GROUPS, null);
    };

    /**
     * Set the given id as the id of the currently used dataset.
     * @param id the id of the dataset.
     */
    this.setDatabaseId = function(id){
      $location.search(UrlParameters.DB_NAME, id);
    };

    /**
     * @return {String} the id of the currently used dataset.
     */
    this.getDatabaseId = function(){
      return $location.search()[UrlParameters.DB_NAME];
    };

    /**
     * Set the given aggregation as the current one.
     * @param aggregation the aggregation to set
     */
    this.setAggregation = function(aggregation){
      $location.search(UrlParameters.AGGREGATION_OF_FEATURE_TABLE, JSON.stringify(aggregation));
    };

    /**
     * Set the given aggregation as the current one.
     * @param aggregation the aggregation to set
     */
    this.setAggregation2ndLevel = function(aggregation){
      $location.search(UrlParameters.AGGREGATION_OF_FEATURE_TABLE_SECOND_LEVEL, JSON.stringify(aggregation));
    };

    /**
     * Return the currently used aggregation.
     */
    this.getAggregation = function(){
      let result = $location.search()[UrlParameters.AGGREGATION_OF_FEATURE_TABLE];
      if(result !== undefined && result !== null){
        result = JSON.parse(result);

        if(result._name === 'AggregatorByUniqueValue'){
          return new AggregatorByUniqueValue(result);

        }else if(result._name === 'AggregatorByEqualWidthBinning'){
          return new AggregatorByEqualWidthBinning(result);

        }else{
          console.log("ERROR: No such aggregator.");
          return null;
        }

      }else{
        return undefined;
      }
    };

    /**
     * Return the currently used aggregation.
     */
    this.getAggregation2ndLevel = function(){
      let result = $location.search()[UrlParameters.AGGREGATION_OF_FEATURE_TABLE_SECOND_LEVEL];
      if(result !== undefined && result !== null){
        result = JSON.parse(result);

        if(result._name === 'AggregatorByUniqueValue'){
          return new AggregatorByUniqueValue(result);

        }else if(result._name === 'AggregatorByEqualWidthBinning'){
          return new AggregatorByEqualWidthBinning(result);

        }else{
          console.log("ERROR: No such aggregator.");
          return null;
        }

      }else{
        return undefined;
      }
    };

    /**
     * Set the layout of the glyphs in the URL parameters.
     * @param glyphLayout the glyph layout. Type GlyphLayout.js.
     */
    this.setGlyphLayout = function(glyphLayout){
      $location.search(UrlParameters.GLYPH_LAYOUT, JSON.stringify(glyphLayout));
    };

    /**
     * Retrieves the current setting of the glyph layout from the url parameters.
     * @return {GlyphLayout}
     */
    this.getGlyphLayout = function(){
      let result = $location.search()[UrlParameters.GLYPH_LAYOUT];
      if(result !== undefined && result !== null) {
        result = JSON.parse(result);
        return new GlyphLayout(result);
      }else{
        return undefined;
      }
    };

    /**
     * Set the sorting of the dimensions within each dimension group.
     * @param sorting the selected sorting. Type Sorting.js.
     */
    this.setSorting = function(sorting){
      // $location.search(UrlParameters.SORTING, JSON.stringify(sorting));
      console.error("set global sorting is disabled! Check code.");
    };

    /**
     * Retrieves the current setting of the sorting of dimensions from the url parameters.
     * @return {Sorting}
     */
    this.getSorting = function(){
      let result = $location.search()[UrlParameters.SORTING];
      if(result !== undefined && result !== null) {
        result = JSON.parse(result);
        return new Sorting(result);
      }else{
        return undefined;
      }
    };

    /**
     * Set the layout of the data record groups in the URL parameters.
     * @param layout the layout of the rows in the visualization {LayoutDataRecordGroups}
     */
    this.setLayoutDataRecordGroups = function(layout){
      if(layout === null){
        $location.search(UrlParameters.LAYOUT_DATA_RECORD_GROUPS, null);
      }else{
        $location.search(UrlParameters.LAYOUT_DATA_RECORD_GROUPS, JSON.stringify(layout));
      }
    };

    /**
     * Retrieves the current setting of the layout of the record groups of the URL parameters.
     * @return {LayoutDataRecordGroups}
     */
    this.getLayoutDataRecordGroups = function(){
      let result = $location.search()[UrlParameters.LAYOUT_DATA_RECORD_GROUPS];
      if(result !== undefined && result !== null) {
        result = JSON.parse(result);
        return new LayoutDataRecordGroups(result);
      }else{
        return undefined;
      }
    };


    /**
     * Set the layout of the dimension groups in the URL parameters.
     * @param layout the layout of the columns
     */
    this.setLayoutDimensionGroups = function(layout){
      if(layout === null){
        $location.search(UrlParameters.LAYOUT_DIMENSION_GROUPS, null);
      }else{
        $location.search(UrlParameters.LAYOUT_DIMENSION_GROUPS, JSON.stringify(layout));
      }
    };

    /**
     * Retrieves the current setting of the layout of the dimension groups of the URL parameters.
     * @return {LayoutDimensionGroups}
     */
    this.getLayoutDimensionGroups = function(){
      let result = $location.search()[UrlParameters.LAYOUT_DIMENSION_GROUPS];
      if(result !== undefined && result !== null) {
        result = JSON.parse(result);
        return new LayoutDimensionGroups(result);
      }else{
        return undefined;
      }
    };


  });

'use strict';

/**
 * @ngdoc service
 * @name smartexploreApp.aggregationPreviewService
 * @description
 *
 * Service to show a preview of a possible aggregation.
 *
 */
angular.module('smartexploreApp')
  .service('aggregationPreviewService', function ($mdDialog) {

    /**
     * Open a dialog to show the preview of a categorical dimension. The user can apply the aggregation or cancel it.
     * @param dimName the name of the dimension
     * @param dimValues all values of the dimension
     * @param fctSuccess the function which is executed once the user applied the aggregation.
     * @param fctCancel the function whcih is executed once the user cancels the aggregation
     */
    this.previewAggregationCategoricalDimension = function(dimName, dimValues, fctSuccess, fctCancel){

      function DialogController($scope, $mdDialog) {

        $scope.dimensionName = dimName;

        // the unique values of the distribution.
        $scope.uniqueValues = _.uniq(dimValues);

        // compute the distribution, for each unique value we need to have the percentage
        $scope.distribution = _.countBy(dimValues, function(v){return v});

        for(let i=0; i<$scope.uniqueValues.length; i++){
          let val = $scope.uniqueValues[i];
          $scope.distribution[val] = $scope.distribution[val] /  dimValues.length;
        }


        $scope.cancel = function() {
          $mdDialog.cancel();
        };

        $scope.apply = function() {
          $mdDialog.hide();
        };
      }


      $mdDialog.show({

        controller: DialogController,
        templateUrl: 'templates/aggregation/dialog-aggregation-categorical.html',
        parent: angular.element(document.body),
        // targetEvent: ev,
        clickOutsideToClose: true

      })
        .then(function() {
          // called ones $mdDialog.hide() is called.
          fctSuccess();
        }, function() {
          fctCancel();
        });


    }



    /**
     * Open a dialog to show the preview of a numerical dimension. The user can apply the aggregation or cancel it.
     * @param dimName the name of the dimension
     * @param dimValues all values of the dimension
     * @param fctSuccess the function which is executed once the user applied the aggregation.
     * @param fctCancel the function whcih is executed once the user cancels the aggregation
     */
    this.previewAggregationNumericalDimension = function(dimName, dimValues, fctSuccess, fctCancel){

      function DialogController($scope, $mdDialog) {

        $scope.dimensionName = dimName;
        $scope.numOfBins = 10;
        $scope.dimValues = dimValues;

        // data binding to access the returned histogram data.
        $scope.resultHistogramData = undefined;


        $scope.cancel = function() {
          $mdDialog.cancel();
        };

        $scope.apply = function() {
          $mdDialog.hide($scope.resultHistogramData);
        };
      }


      $mdDialog.show({

        controller: DialogController,
        templateUrl: 'templates/aggregation/dialog-aggregation-numerical.html',
        parent: angular.element(document.body),
        clickOutsideToClose: true

      })
        .then(function(bins) {
          fctSuccess(bins);
        }, function() {
          fctCancel();
        });


    }

  });

'use strict';

/**
 * @ngdoc service
 * @name smartexploreApp.aggregationPreviewService
 * @description
 *
 * Service to select the correct statistic parameters.
 *
 */
angular.module('smartexploreApp')
  .service('statisticsService', function () {

    /** The alpha value for statistical significance. */
    this.alpha = 0.05;

    /**
     * Returns the p value of the given numerical dimension.
     */
    this.numericalDimensionGetP = function(dimension){
      return dimension.statistics.pValueAnova;
    };

    /**
     * Returns the effect size of the given numerical dimension.
     */
    this.numericalDimensionGetEffectsize = function(dimension){
      return dimension.statistics.effectSizeEta;
    };

    /**
     * Returns the p value of the given aggregated numerical value (compared to the whole dimension)
     */
    this.numericalAggregatedValueGetP = function(aggregatedNumericalValue){
      // TODO set based on user selection.
      try{
        return aggregatedNumericalValue.statistics.pValueTTestIndependentComparedToDimensionEqualVar;
      }catch (err){
        console.log("TODO @Steffi: Make sure the errors are handled properly.");
        return 2.0;
      }

    };

    /**
     * Returns the effect size of the given aggregated numerical value (compared to the whole dimension)
     */
    this.numericalAggregatedValueGetEffectSize = function(aggregatedNumericalValue){
      return aggregatedNumericalValue.statistics.effectiveSizeCohen;
    };


    /**
     * Returns the p value of the given categorical dimension.
     * TODO needs to be implemented on the server.
     */
    this.categoricalDimensionGetP = function(dimension){
      return dimension.statistics.pValue
    };

    /**
     * Returns the effect size of the given categorical dimension.
     * TODO needs to be implemented on the server.
     */
    this.categoricalDimensionGetEffectsize = function(dimension){
      return Math.random();
    };

    /**
     * Returns the p value of the given aggregated categorical value (compared to the whole dimension)
     */
    this.categoricalAggregatedValueGetP = function(aggregatedCategoricalValue){
      return aggregatedCategoricalValue.statistics.pValueChiSquaredTest;
    };

    /**
     * Returns the effect size of the given aggregated categorical value (compared to the whole dimension)
     */
    this.categoricalAggregatedValueGetEffectSize = function(aggregatedCategoricalValue){
      return aggregatedCategoricalValue.statistics.effectSizeChramersV;
    };


    /**
     * Returns true, if the values of the given dimension are significantly different; false otherwise.
     * @param dimension the dimension to be checked.
     */
    this.numericalDimensionIsSignificant = function(dimension){
      return this.numericalDimensionGetP(dimension) <= this.alpha;
    };


    /**
     * Returns true, if the value is significantly different from the whole dimension; false otherwise.
     * @param dimension the dimension
     * @param aggregatedNumericalValue the value
     */
    this.numericalAggregatedValueIsSignificant = function(dimension, aggregatedNumericalValue){
      return this.numericalAggregatedValueGetP(aggregatedNumericalValue) <= this.alpha;
    };


    /**
     * Returns the pVale and the effect size of the given dimension.
     * @param dimension the dimension to check.
     * @return {StatisticInformation}
     */
    this.numericalDimensionPandEffect = function(dimension){
      return new StatisticInformation(this.numericalDimensionGetP(dimension), this.numericalDimensionGetEffectsize(dimension));
    };

    /**
     * Returns the pValue and the effect size of the aggregated numerical value.
     * @return {StatisticInformation}
     */
    this.numericalValuePandEffect = function(aggregatedValue){
      return new StatisticInformation(this.numericalAggregatedValueGetP(aggregatedValue), this.numericalAggregatedValueGetEffectSize(aggregatedValue));
    };

    /**
     * Returns the pVale and the effect size of the given dimension.
     * @param dimension the dimension to check.
     * @return {StatisticInformation}
     */
    this.categoricalDimensionPandEffect = function(dimension){
      return new StatisticInformation(this.categoricalDimensionGetP(dimension), this.categoricalDimensionGetEffectsize(dimension));
    };

    /**
     * Returns the pValue and the effect size of the aggregated numerical value.
     * @return {StatisticInformation}
     */
    this.categoricalValuePandEffect = function(aggregatedValue){
      return new StatisticInformation(this.categoricalAggregatedValueGetP(aggregatedValue), this.categoricalAggregatedValueGetEffectSize(aggregatedValue));
    };

    /**
     * Returns true, if the value is significantly different from the whole dimension; false otherwise.
     * @param dimension the dimension
     * @param aggregatedNumericalValue the value
     */
    this.categoricalDimensionIsSignificant = function(dimension){
      return this.categoricalDimensionGetP(dimension) <= this.alpha;
    };

    /**
     * Returns true, if the value is significantly different from the whole dimension; false otherwise.
     * @param dimension the dimension
     * @param aggregatedValue the value
     */
    this.categoricalAggregatedValueIsSignificant = function(dimension, aggregatedValue){
      return this.categoricalAggregatedValueGetP(aggregatedValue) <= this.alpha;
    };

    /**
     * Returns the p-value and effect size of the aggregated value Automatically selects the right measure based on the type of the dimension.
     * @param dimension the dimension to check.
     * @return {StatisticInformation}
     */
    this.aggregatedValuePandEffect = function(dimension, aggregatedValue){

      if(dimension.dimType === 'NUMERICAL'){
        return this.numericalValuePandEffect(aggregatedValue);
      }else if(dimension.dimType === 'CATEGORICAL'){
        return this.categoricalValuePandEffect(aggregatedValue);
      }
    };

    /**
     * Returns the p-value and effect size of the whole dimension. Automatically selects the right measure based on the type of the dimension.
     * @param dimension the dimension to check.
     * @return {StatisticInformation}
     */
    this.dimensionPandEffect = function(dimension){

      if(dimension.dimType === 'NUMERICAL'){
        return this.numericalDimensionPandEffect(dimension);
      }else if(dimension.dimType === 'CATEGORICAL'){
        return this.categoricalDimensionPandEffect(dimension);
      }
    };

    /**
     * Returns true or false, whether the dimension is significantly different.
     * @param dimension the dimension to check.
     */
    this.dimensionIsSignificant = function(dimension){
      if(dimension.dimType === 'NUMERICAL'){
        return this.numericalDimensionIsSignificant(dimension);
      }else if(dimension.dimType === 'CATEGORICAL'){
        return this.categoricalDimensionIsSignificant(dimension);
      }
    };

    /**
     * Class holding a p-value / effect size combination.
     */
    class StatisticInformation {

      constructor(pValue, effectSize){
        this.effect = effectSize;
        this._pValue = pValue;
        this._effectSize = effectSize;
      }

      get pValue() {
        return this._pValue;
      }

      set pValue(value) {
        this._pValue = value;
      }

      get effectSize() {
        return this._effectSize;
      }

      set effectSize(value) {
        this._effectSize = value;
      }
    }

  });

'use strict';

/**
 * Service managing selected dimensions, data record groups etc.
 *
 */
angular.module('smartexploreApp')
  .service('selectionService', function () {

    let self = this;

    /** The names of all selected dimensions. */
    this.selectedDimensions = [];

    /** The names of all selected data record groups. */
    this.selectedDataRecordGroup = [];

    /**
     * Add the given dimension to the selected dimensions.
     * @param dimName the name of the dimension.
     */
    this.addSelectedDimension = function(dimName){

      self.removeAllSelectedDataRecordGroups();
      self.updateSelectedDataRecordGroups();

      self.selectedDimensions.push(dimName);
      self.updateSelectionDimensions();
    };

    /**
     * Remove the given dimension from the selected dimensions.
     * @param dimName the name of the dimension.
     */
    this.removeSelectedDimension = function(dimName){
      self.selectedDimensions = _.without( self.selectedDimensions, dimName);
      self.updateSelectionDimensions();
    };

    /**
     * Checks if the given dimension is currently selected or not.
     * @param dimName the name of the dimension
     */
    this.isSelectedDimension = function (dimName) {
      return _.indexOf(self.selectedDimensions, dimName) > -1;
    };

    /**
     * Returns true, if any dimension is currently selected; false otherwise.
     */
    this.isAnyDimensionSelected = function () {
      return self.selectedDimensions.length > 0;
    };

    /**
     * @return {Array} of all selected dimensions;
     */
    this.getSelectedDimensions = function () {
      return self.selectedDimensions;
    };

    /**
     * Add the given data record group to the selected groups.
     * @param groupName the name of the group
     */
    this.addSelectedDataRecordGroup = function (groupName) {

      self.removeAllSelectedDimensions();
      self.updateSelectionDimensions();

      self.selectedDataRecordGroup.push(groupName);
      self.updateSelectedDataRecordGroups();
    };

    /**
     * Remove the given data record group from the selected groups.
     * @param groupName the name of the group
     */
    this.removeSelectedDataRecordGroup = function (groupName) {
      self.selectedDataRecordGroup = _.without(self.selectedDataRecordGroup, groupName);
      self.updateSelectedDataRecordGroups();
    };

    /**
     * Checks if the given data record group is currently selected or not.
     * @param groupName the name of the group-
     */
    this.isSelectedDataRecordGroup = function (groupName) {
      return _.indexOf(self.selectedDataRecordGroup, groupName) > -1;
    };

    /**
     * Returns true, if any data record group is currently selected; false otherwise.
     */
    this.isAnyDataRecordGroupSelected = function () {
      return self.selectedDataRecordGroup.length > 0;
    };

    /**
     * Remove all dimensions from the selection.
     */
    this.removeAllSelectedDimensions = function () {
      self.selectedDimensions = [];
      self.updateSelectionDimensions();
    };

    /**
     * Remove all data record groups from the selection.
     */
    this.removeAllSelectedDataRecordGroups = function () {
      self.selectedDataRecordGroup = [];
    };

    this.removeAllSelections = function () {
      self.removeAllSelectedDimensions();
      self.removeAllSelectedDataRecordGroups();
    };

    /**
     * @return {Array} of all selected data record groups.
     */
    this.getSelectedDataRecordGroups = function () {
      return self.selectedDataRecordGroup;
    };

    /**
     * Update the selected dimensions in the GUI based on the parameter 'self.selectedDimensions'.
     */
    this.updateSelectionDimensions = function () {

      if(self.selectedDimensions.length > 0){
        // remove highlight from all dimensions
        self.removeHighlightAllDimension();

        // highlight all dimensions.
        self.selectedDimensions.forEach(function(dimName){
          self.addHighlightDimension(dimName);
        });

      }else{
        // highlight all dimensions
        self.highlightAllDimensions();

      }


    };

    /**
     * Update the selected data record groups in the GUI based on the parameter 'self.selectedDataRecordGroup'.
     */
    this.updateSelectedDataRecordGroups = function () {

      if(self.selectedDataRecordGroup.length > 0){

        // remove highlight from all data record groups.
        self.removeHighlightAllDataRecordGroups();

        // highlight all selected record groups
        self.selectedDataRecordGroup.forEach(function(id){
          self.addHighlightDataRecordGroup(id);
        });

      }else{
        // highlight all data record groups.
        self.highlightAllDataRecordGroups();
      }

    };

    /**
     * Remove the highlight of all dimensions.
     */
    this.removeHighlightAllDimension = function(){

      // the distributions
      d3.selectAll(".backgroundDimDistribution").classed("highlight", false);
      // the aggregated values
      d3.selectAll(".highlightOverlay").classed("highlight", false);
    };

    /**
     * Highlight all dimensions in the dataset.
     */
    this.highlightAllDimensions = function(){
      // the distributions
      d3.selectAll(".backgroundDimDistribution").classed("highlight", true);
      // the aggregated values
      d3.selectAll(".highlightOverlay").classed("highlight", true);
    };

    /**
     * Highlight the dimension with the given name.
     * @param dimName
     */
    this.addHighlightDimension = function(dimName){
      let dim = d3.selectAll("#dim-" + dimName);
      dim.selectAll(".highlightOverlay").classed("highlight", true);
      dim.select(".backgroundDimDistribution").classed("highlight", true);
    };

    /**
     * Remove the highlight of all data record groups.
     */
    this.removeHighlightAllDataRecordGroups = function(){
      d3.selectAll(".recordGroupRowHist").classed("highlight",true);
      d3.selectAll(".recordGroupRowHistSecondAgg").classed("highlight",true);
      d3.selectAll(".highlightOverlayRow").classed("highlight", false);
    };

    /**
     * Highlight all data record groups in the dataset.
     */
    this.highlightAllDataRecordGroups = function(){
      d3.selectAll(".recordGroupRowHist").classed("highlight",false);
      d3.selectAll(".recordGroupRowHistSecondAgg").classed("highlight",false);
      d3.selectAll(".highlightOverlayRow").classed("highlight", true);
    };

    /**
     * Highlight the data record group with the given name.
     * @param groupName
     */
    this.addHighlightDataRecordGroup = function(groupName){
      // all glyphs
      d3.selectAll(".glyphOfRow" + groupName).selectAll(".highlightOverlayRow").classed("highlight", true);

      // the histogram on the very left.
      let row = d3.select("#recordRowGroup" + groupName);
      row.selectAll(".recordGroupRowHist").classed("highlight",false);
      row.selectAll(".recordGroupRowHistSecondAgg").classed("highlight",false);
    };

  });

'use strict';

/**
 * Service managing selected dimensions, data record groups etc.
 *
 */
angular.module('smartexploreApp')
  .service('updateVisService', function () {

    let self = this;
    this.updateVisCalls = [];

    this.addCall = function(call){
      self.updateVisCalls.push(call);
    };

    this.notifyUpdateVis = function(){
      self.updateVisCalls.forEach(function(d){
        d.updateVisualization();
      })
    }


  });

'use strict';

class DatasetInfo {

  constructor({_id, _datasetName, _databaseName}){
    this._id = _id;
    this._datasetName = _datasetName;
    this._databaseName = _databaseName;
  }

  get id() {
    return this._id;
  }

  set id(value) {
    this._id = value;
  }

  get datasetName() {
    return this._datasetName;
  }

  set datasetName(value) {
    this._datasetName = value;
  }

  get databaseName() {
    return this._databaseName;
  }

  set databaseName(value) {
    this._databaseName = value;
  }



}

'use strict';

/**
 * Corresponds to the FeatureTableDto and storing all information regarding the feature table.
 * No layout information.
 */
class FeatureTable {

  constructor({datasetName, nameToDimension, nameToDataRecordGroup}){

    this._datasetName = datasetName;

    // map record group name -> record group.
    this._nameToDataRecordGroup = {};
    for (let groupName of _.keys(nameToDataRecordGroup)) {
      this._nameToDataRecordGroup[groupName] = new DataRecordGroup(nameToDataRecordGroup[groupName]);
    }

    // map dimension name -> dimension
    this._nameToDimension = nameToDimension;
    this._nameToDataRecordGroup = nameToDataRecordGroup;

  }

  get datasetName() {
    return this._datasetName;
  }

  set datasetName(value) {
    this._datasetName = value;
  }

  get nameToDimension() {
    return this._nameToDimension;
  }

  set nameToDimension(value) {
    this._nameToDimension = value;
  }

  get nameToDataRecordGroup() {
    return this._nameToDataRecordGroup;
  }

  set nameToDataRecordGroup(value) {
    this._nameToDataRecordGroup = value;
  }
}

'use strict';

/**
 * All sortings of the columns which are available.
 * @type {Array}
 */
const AVAILABLE_SORTINGS = [
  {'id': 'noSorting', 'name': 'No sorting (same order as in input dataset)', 'inGUI' : true},
  {'id': 'mds', 'name': '1D MDS', 'inGUI' : true},
  {'id': 'avgMeasure', 'name': 'Avg. Norm. Measure', 'inGUI' : true},
  // {'id': 'bySimilarDimension', 'name': 'By Similarity To Dimension', 'inGUI' : false},
  // {'id': 'optimalLeaf', 'name': '1D Optimal Leaf Ordering'},
  // {'id': 'hierarchicalClust', 'name': 'Hierarchical Clustering'},
];

/**
 * Class describing the sorting of each group of the feature table.
 * See SortingDto.java on server side.
 */
class Sorting{

  constructor({_sortOrder, _dimension}){

    /** The actual sort order which has been selected by the user - or default value. */
    this._sortOrder = _sortOrder !== undefined ? _sortOrder : AVAILABLE_SORTINGS[0].id;

    /** Optional, only when sorting based on similarity to another dimension. */
    this._dimension = _dimension;
  }

  /**
   * @return {Sorting} a default representation of the sorting.
   */
  static getDefaultSorting(){
    return new Sorting({});
  }

  get sortOrder() {
    return this._sortOrder;
  }

  set sortOrder(value) {
    this._sortOrder = value;
  }


  get dimension() {
    return this._dimension;
  }

  set dimension(value) {
    this._dimension = value;
  }

  static get AVAILABLE_SORTINGS(){
    return AVAILABLE_SORTINGS;
  }
}

'use strict';

/**
 * All distance functions for the columns of an aggregated feature table.
 * @type {Array}
 */
const AVAILABLE_DISTANCES_DIMENSION = [
  {'id': 'euclidean', 'name': 'Euclidean distance'},
  {'id': 'euclideanSliding', 'name': 'Weighted Euclidean distance (neighboring rows)'},
  {'id': 'euclideanWeightedSigValue', 'name': 'Weighted Euclidean distance (significance)'},
  {'id': 'euclideanWeightedSigValueSliding', 'name': 'Weighted Euclidean distance (significance + neighboring)'},
];

/**
 * Class describing the different distance functions for dimensions.
 * See DimensionDistanceDto.java on server side.
 */
class DimensionDistance{

  constructor({_distanceFunction}){

    /** The actual sort order which has been selected by the user. */
    this._distanceFunction = _distanceFunction !== undefined ? _distanceFunction : AVAILABLE_DISTANCES_DIMENSION[0].id;
  }

  /**
   * @return {DimensionDistance} a default representation of the distance.
   */
  static getDefaultDistance(){
    return new DimensionDistance({});
  }

  get distanceFunction() {
    return this._distanceFunction;
  }

  set distanceFunction(value) {
    this._distanceFunction = value;
  }

  static get AVAILABLE_DISTANCES_DIMENSION(){
    return AVAILABLE_DISTANCES_DIMENSION;
  }
}

'use strict';

/**
 * Designs for significant glyphs.
 * @type {[*]}
 */
const GLYPH_DESIGN_SIGNIFICANCE = [
  {'id': 'noSignificance', 'name': 'Significance NOT Shown'},
  {'id': 'dot', 'name': 'DOT = Significant Difference'},
  {'id': 'full', 'name': 'FULL = Significant Difference'},
];

/**
 * Describes the visual representation of a single glyph.
 * See server: GlyphLayoutDto.java
 */
class GlyphLayout{

  constructor({_showDistributionOverlay, _showMissingValues,
    _numericalMeasure, _categoricalMeasure,
    _numericalDesign, _categoricalDesign, _designSignificance,
    _numericalSignificanceAnalysisValues, _categoricalSignificanceAnalysisValues,
    _numericalSignificanceAnalysisDimensions, _categoricalSignificanceAnalysisDimensions,
    _showDimensionIsSignificant, _useSmallColormapForNonSignificantDimensions,
    _alphaValue}){

    // set default values
    _showDistributionOverlay = _showDistributionOverlay !== undefined ? _showDistributionOverlay : false;
    _showMissingValues = _showMissingValues !== undefined ? _showMissingValues : 'disabled';
    _numericalMeasure = _numericalMeasure !== undefined ? _numericalMeasure : 'distanceToMean';
    _categoricalMeasure = _categoricalMeasure !== undefined ? _categoricalMeasure : 'distanceToAverage';
    _numericalDesign = _numericalDesign !== undefined ? _numericalDesign : NumericalGlyph.GLYPH_DESIGN[0].id;
    _categoricalDesign = _categoricalDesign !== undefined ? _categoricalDesign : CategoricalGlyph.GLYPH_DESIGN_CATEGORICAL[0].id;
    _designSignificance = _designSignificance !== undefined ? _designSignificance : GlyphLayout.GLYPH_DESIGN_SIGNIFICANCE[0].id;
    _numericalSignificanceAnalysisValues = _numericalSignificanceAnalysisValues !== undefined ? _numericalSignificanceAnalysisValues : 'tTestIndependent';
    _categoricalSignificanceAnalysisValues = _categoricalSignificanceAnalysisValues !== undefined ? _categoricalSignificanceAnalysisValues : 'chiSquareTest';
    _numericalSignificanceAnalysisDimensions = _numericalSignificanceAnalysisDimensions !== undefined ? _numericalSignificanceAnalysisDimensions : 'anova';
    _categoricalSignificanceAnalysisDimensions = _categoricalSignificanceAnalysisDimensions !== undefined ? _categoricalSignificanceAnalysisDimensions : 'anova';
    _showDimensionIsSignificant = _showDimensionIsSignificant !== undefined ? _showDimensionIsSignificant : true;
    _useSmallColormapForNonSignificantDimensions = _useSmallColormapForNonSignificantDimensions !== undefined ? _useSmallColormapForNonSignificantDimensions : false;
    _alphaValue = _alphaValue !== undefined ? _alphaValue : 0.05;





    this._showDistributionOverlay = _showDistributionOverlay;
    this._showMissingValues = _showMissingValues;
    this._numericalMeasure = _numericalMeasure;
    this._categoricalMeasure = _categoricalMeasure;
    this._numericalDesign = _numericalDesign;
    this._categoricalDesign = _categoricalDesign;
    this._designSignificance = _designSignificance;

    this._numericalSignificanceAnalysisValues = _numericalSignificanceAnalysisValues;
    this._categoricalSignificanceAnalysisValues = _categoricalSignificanceAnalysisValues;
    this._numericalSignificanceAnalysisDimensions = _numericalSignificanceAnalysisDimensions;
    this._categoricalSignificanceAnalysisDimensions = _categoricalSignificanceAnalysisDimensions;

    /** If true, the visualization shows, whether a whole dimension is significantly interesting (e.g., using an ANOVA test.)*/
    this._showDimensionIsSignificant = _showDimensionIsSignificant;
    /** If true, the non-significant dimensions (see this.showDimensionIsSignificant) are visualized only with the inner part of the colormap. */
    this._useSmallColormapForNonSignificantDimensions = _useSmallColormapForNonSignificantDimensions;

    this._alphaValue = _alphaValue;
    this._categoricalDesign = _categoricalDesign;
  }


  /**
   * @return {GlyphLayout} a default representation of the glyph layout.
   */
  static getDefaultGlyphLayout(){
    return new GlyphLayout({});
  }

  static get GLYPH_DESIGN_SIGNIFICANCE(){
    return GLYPH_DESIGN_SIGNIFICANCE;
  }

  get showDistributionOverlay() {
    return this._showDistributionOverlay;
  }

  set showDistributionOverlay(value) {
    this._showDistributionOverlay = value;
  }


  get showMissingValues() {
    return this._showMissingValues;
  }

  set showMissingValues(value) {
    this._showMissingValues = value;
  }


  get numericalMeasure() {
    return this._numericalMeasure;
  }

  set numericalMeasure(value) {
    this._numericalMeasure = value;
  }


  get categoricalMeasure() {
    return this._categoricalMeasure;
  }

  set categoricalMeasure(value) {
    this._categoricalMeasure = value;
  }


  get numericalDesign() {
    return this._numericalDesign;
  }

  set numericalDesign(value) {
    this._numericalDesign = value;
  }


  get numericalSignificanceAnalysisValues() {
    return this._numericalSignificanceAnalysisValues;
  }

  set numericalSignificanceAnalysisValues(value) {
    this._numericalSignificanceAnalysisValues = value;
  }


  get categoricalSignificanceAnalysisValues() {
    return this._categoricalSignificanceAnalysisValues;
  }

  set categoricalSignificanceAnalysisValues(value) {
    this._categoricalSignificanceAnalysisValues = value;
  }

  get numericalSignificanceAnalysisDimensions() {
    return this._numericalSignificanceAnalysisDimensions;
  }

  set numericalSignificanceAnalysisDimensions(value) {
    this._numericalSignificanceAnalysisDimensions = value;
  }

  get categoricalSignificanceAnalysisDimensions() {
    return this._categoricalSignificanceAnalysisDimensions;
  }

  set categoricalSignificanceAnalysisDimensions(value) {
    this._categoricalSignificanceAnalysisDimensions = value;
  }


  get showDimensionIsSignificant() {
    return this._showDimensionIsSignificant;
  }

  set showDimensionIsSignificant(value) {
    this._showDimensionIsSignificant = value;
  }

  get useSmallColormapForNonSignificantDimensions() {
    return this._useSmallColormapForNonSignificantDimensions;
  }

  set useSmallColormapForNonSignificantDimensions(value) {
    this._useSmallColormapForNonSignificantDimensions = value;
  }

  get alphaValue() {
    return this._alphaValue;
  }

  set alphaValue(value) {
    this._alphaValue = value;
  }


  get categoricalDesign() {
    return this._categoricalDesign;
  }

  set categoricalDesign(value) {
    this._categoricalDesign = value;
  }

  get designSignificance() {
    return this._designSignificance;
  }

  set designSignificance(value) {
    this._designSignificance = value;
  }


}

'use strict';

/**
 * Corresponds to the DimensionGroupDto and storing all information regarding one group of dimensions.
 */
class DimensionGroup {

  constructor({_name, _dimensions, _dimensionToSimilarityPos, _dimensionDistance, _sortingStrategy, _normalizingStrategy, _sortSimilarToDimension}) {
    this._name = _name;
    this._dimensions = _dimensions;
    this._dimensionToSimilarityPos = _dimensionToSimilarityPos;

    this._sortSimilarToDimension = _sortSimilarToDimension;

    this._dimensionDistance = new DimensionDistance(_dimensionDistance);
    this._sortingStrategy = new Sorting(_sortingStrategy);
    this._normalizingStrategy = new Normalizing(_normalizingStrategy);
  }

  get name() {
    return this._name;
  }

  set name(value) {
    this._name = value;
  }

  get dimensions() {
    return this._dimensions;
  }

  set dimensions(value) {
    this._dimensions = value;
  }

  get dimensionToSimilarityPos() {
    return this._dimensionToSimilarityPos;
  }

  set dimensionToSimilarityPos(value) {
    this._dimensionToSimilarityPos = value;
  }

  get dimensionDistance() {
    return this._dimensionDistance;
  }

  set dimensionDistance(value) {
    this._dimensionDistance = value;
  }

  get sortingStrategy() {
    return this._sortingStrategy;
  }

  set sortingStrategy(value) {
    this._sortingStrategy = value;
  }

  get normalizingStrategy() {
    return this._normalizingStrategy;
  }

  set normalizingStrategy(value) {
    this._normalizingStrategy = value;
  }
}

'use strict';

/**
 * Corresponds to the DataRecordGroupDto and storing all information regarding one group of data records.
 */
class DataRecordGroup {

  constructor({_name, _label, _size, _nextLevelAggregation, _idsOfRecords}) {
    this._name = _name;
    this._label = _label;
    this._size = _size;
    this._idsOfRecords = _idsOfRecords;
    this._nextLevelAggregation = _nextLevelAggregation;
  }

  get name() {
    return this._name;
  }

  set name(value) {
    this._name = value;
  }


  get label() {
    return this._label;
  }

  set label(value) {
    this._label = value;
  }

  get size() {
    return this._size;
  }

  set size(value) {
    this._size = value;
  }

  get idsOfRecords() {
    return this._idsOfRecords;
  }

  set idsOfRecords(value) {
    this._idsOfRecords = value;
  }

  get nextLevelAggregation() {
    return this._nextLevelAggregation;
  }

  set nextLevelAggregation(value) {
    this._nextLevelAggregation = value;
  }

  /**
   * Merge the given record groups.
   * @param label the label/name of the new group
   * @param recordGroups []
   * @return new DataRecordGroup.
   */
  static mergeGroups(label, recordGroups){

    console.log("record groups", recordGroups);

    let _name = recordGroups[0]._name; // the id of the record groups - use an previous to make sure it is valid.
    let _label = label;
    let _size = 0;
    let _nextLevelAggregation = recordGroups[0]._nextLevelAggregation ? [] : null;
    let _idsOfRecords = [];

    // check if we have a second level aggregation
    if(recordGroups[0]._nextLevelAggregation){
      recordGroups[0]._nextLevelAggregation.forEach(function(a){
        // for the second level, we only need the label and the size.
        let agg = new DataRecordGroup({
          '_name' : null,
          '_label' : a._label,
          '_size' : 0,
          '_nextLevelAggregation' : null,
          '_idsOfRecords' : null
        });
        _nextLevelAggregation.push(agg);
      })
    }

    recordGroups.forEach(function(group){
      _size += group._size;
      _idsOfRecords = _idsOfRecords.concat(group._idsOfRecords);

      if(group._nextLevelAggregation){

        // iterate over the 2nd level and adapt the size.
        for(let i=0; i<group._nextLevelAggregation.length; i++){
          _nextLevelAggregation[i]._size += group._nextLevelAggregation[i]._size;
        }
      }

    });

    return new DataRecordGroup({
      '_name' : _name,
      '_label' : _label,
      '_size' : _size,
      '_nextLevelAggregation' : _nextLevelAggregation,
      '_idsOfRecords' : _idsOfRecords
    })
  }
}

'use strict';

/**
 * All distance functions for the columns of an aggregated feature table.
 * @type {Array}
 */
const AVAILABLE_NORMALIZING_STRATEGIES = [
  {'id': 'perDimension', 'name': '[min,max] per dimension'},
  {'id': 'perDimensionGroup', 'name': '[min,max] per dimension group'},
];

/**
 * Class describing the different distance functions for dimensions.
 * See DimensionDistanceDto.java on server side.
 */
class Normalizing{

  constructor({_normalizing}){

    /** The actual sort order which has been selected by the user - alternative: a default value. */
    this._normalizing = _normalizing !== undefined ? _normalizing : AVAILABLE_NORMALIZING_STRATEGIES[0].id;
  }

  get normalizing() {
    return this._normalizing;
  }

  set normalizing(value) {
    this._normalizing = value;
  }

  static get AVAILABLE_NORMALIZING_STRATEGIES(){
    return AVAILABLE_NORMALIZING_STRATEGIES;
  }
}

/**
 * Aggregator to aggregate the feature table by a unique value.
 * This is particularly interesting for categorical dimensions
 */
class AggregatorByUniqueValue{

  /**
   * @param _name the name of this aggregator
   * @param _dimensionName the name of the dimension to aggregate the data.
   * @param _dimensionType the type of the dimension. Possible values are {numerical, categorical}
   */
  constructor({_name, _dimensionName, _dimensionType}){

    // set default values
    _name = _name !== undefined ? _name : "AggregatorByUniqueValue";

    this._name = _name;
    this._dimensionName = _dimensionName;
    this._dimensionType = _dimensionType;
  }


  get name() {
    return this._name;
  }

  set name(value) {
    this._name = value;
  }

  get dimensionName() {
    return this._dimensionName;
  }

  set dimensionName(value) {
    this._dimensionName = value;
  }

  get dimensionType() {
    return this._dimensionType;
  }

  set dimensionType(value) {
    this._dimensionType = value;
  }
}

/**
 * Aggregator to aggregate the feature table by a unique value.
 * This is particularly interesting for categorical dimensions
 */
class AggregatorByEqualWidthBinning{

  /**
   * @param _name the name of this aggregator
   * @param _dimensionName the name of the dimension to aggregate the data.
   * @param _binning the binning of this aggregator, same as d3.histogram, without the actual data records containing only x0 and x1 per entry.
   */
  constructor({_name, _dimensionName, _binning}){

    // set default values
    _name = _name !== undefined ? _name : "AggregatorByEqualWidthBinning";

    this._name = _name;
    this._dimensionName = _dimensionName;
    this._binning = _binning;

  }


  get name() {
    return this._name;
  }

  set name(value) {
    this._name = value;
  }

  get dimensionName() {
    return this._dimensionName;
  }

  set dimensionName(value) {
    this._dimensionName = value;
  }


  get binning() {
    return this._binning;
  }

  set binning(value) {
    this._binning = value;
  }
}

/**
 * Aggregator to aggregate the feature table by a unique value.
 * This is particularly interesting for categorical dimensions
 */
class AggregatorCustom{

  /**
   * @param _name the name of this aggregator
   * @param _dimensionName the name of the dimension to aggregate the data.
   * @param _dimensionType the type of the dimension. Possible values are {numerical, categorical}
   */
  constructor({_name, _dimensionName, _dimensionType}){

    // set default values
    _name = _name !== undefined ? _name : "AggregatorCustom";

    this._name = _name;
    this._dimensionName = _dimensionName;
    this._dimensionType = _dimensionType;
  }


  get name() {
    return this._name;
  }

  set name(value) {
    this._name = value;
  }

  get dimensionName() {
    return this._dimensionName;
  }

  set dimensionName(value) {
    this._dimensionName = value;
  }

  get dimensionType() {
    return this._dimensionType;
  }

  set dimensionType(value) {
    this._dimensionType = value;
  }
}

'use strict';

/** The name of the used database. */
const DB_NAME = "dbName";

/** The type of aggregation of the feature table. */
const AGGREGATION_OF_FEATURE_TABLE = "aggregation";

/** The type of aggregation of the feature table. */
const AGGREGATION_OF_FEATURE_TABLE_SECOND_LEVEL = "aggregation2ndLevel";

/** The layout of the glyphs. The server is not interested in this property (most likely). */
const GLYPH_LAYOUT = "glyphLayout";

/** The sorting of the dimensions within each dimension group. */
const SORTING = "sorting";

/** The layout of the data record groups (rows of the visualization) */
const LAYOUT_DATA_RECORD_GROUPS = "dataRecordGroups";

/** The layout of the dimension groups (groups of columns) */
const LAYOUT_DIMENSION_GROUPS = "dimensionGroups";

/**
 * Parameters (key's) for the $location.search.
 */
class UrlParameters{

  static get DB_NAME(){
    return DB_NAME;
  }

  static get AGGREGATION_OF_FEATURE_TABLE(){
    return AGGREGATION_OF_FEATURE_TABLE;
  }

  static get AGGREGATION_OF_FEATURE_TABLE_SECOND_LEVEL(){
    return AGGREGATION_OF_FEATURE_TABLE_SECOND_LEVEL;
  }

  static get GLYPH_LAYOUT(){
    return GLYPH_LAYOUT;
  }

  static get SORTING(){
    return SORTING;
  }

  static get LAYOUT_DATA_RECORD_GROUPS(){
    return LAYOUT_DATA_RECORD_GROUPS;
  }

  static get LAYOUT_DIMENSION_GROUPS(){
    return LAYOUT_DIMENSION_GROUPS;
  }

}

'use strict';

/**
 * The layout information for the data record groups (= the rows of the feature table)
 * The user can interactively also remove rows etc.
 * This class will also be bound to the URL parameters.
 */
class LayoutDataRecordGroups {

  /**
   * Converts the data record groups of the server into an appropriate json format.
   * @param _dataRecordGroups array of data record groups.
   * @param _maxLevelOfAggregation the level of aggregation
   */
  constructor({_dataRecordGroups, _maxLevelOfAggregation}){

    this._dataRecordGroups = [];
    for (let group of _dataRecordGroups) {
      this._dataRecordGroups.push(new DataRecordGroup(group));
    }

    this._maxLevelOfAggregation = _maxLevelOfAggregation;

  }

  get dataRecordGroups() {
    return this._dataRecordGroups;
  }

  set dataRecordGroups(value) {
    this._dataRecordGroups = value;
  }


  get maxLevelOfAggregation() {
    return this._maxLevelOfAggregation;
  }

  set maxLevelOfAggregation(value) {
    this._maxLevelOfAggregation = value;
  }

  /**
   * @returns {Number} the number of record groups.
   */
  getNumberOfRecordGroups() {
    return this._dataRecordGroups.length;
  };

  /**
   * @returns {Number} the number of record groups in the second level aggregation.
   */
  getNumberOfRecordGroups2ndLevel(){
    if(this._dataRecordGroups[0]._nextLevelAggregation === null || this._dataRecordGroups[0]._nextLevelAggregation === undefined){
      return 0;
    }else{
      return this._dataRecordGroups[0]._nextLevelAggregation.length;
    }
  }

  /**
   * @returns an array containing the labels of the data record groups of the second aggregation level.
   */
  getRecordGroupLabels2ndLevel(){
    let numGroups = this.getNumberOfRecordGroups2ndLevel();
    if(numGroups == 0) {
      return [];
    }
    else{
      let a = [];
      for(let i=0; i<numGroups; i++){
        a.push(this._dataRecordGroups[0]._nextLevelAggregation[i]._label);
      }
      return a;
    }
  }

}

'use strict';

/**
 * Corresponds to the FeatureTableLayoutDto and storing all layout information regarding the feature table.
 */
class LayoutDimensionGroups {

  /**
   * Converts the data record groups and dimension group representation of the server into the appropriate json format.
   *
   * @param _dimensionGroups array of dimensionGroups.
   */
  constructor({_dimensionGroups}){

    this._dimensionGroups = [];
    if(_dimensionGroups !== undefined && _dimensionGroups.length > 0){
      for (let dimGroup of _dimensionGroups) {
        this._dimensionGroups.push(new DimensionGroup(dimGroup));
      }
    }
  }

  get dimensionGroups() {
    return this._dimensionGroups;
  }

  set dimensionGroups(value) {
    this._dimensionGroups = value;
  }

  /**
   * @returns {Number} the number of dimension groups.
   */
  getNumberOfDimensionGroups(){
    return this._dimensionGroups.length;
  };

  /**
   * Replace the dimension group with the following name.
   * @param name
   * @param group
   */
  replaceGroupWithName(name, group){
    let posInArray = _.findIndex(this._dimensionGroups, function(d){return d._name === name;});
    this._dimensionGroups[posInArray] = group;
  }

}

'use strict';

/**
 * Configuration of the similarity based highlighting.
 */
class SimilarityBasedHighlightingConfigureDto{

  constructor({_acrossGroups, _dimension, _nameOfDimensionGroup}){

    /** boolean, decide if to use all dimensions or only the onces within a group.*/
    this._acrossGroups = _acrossGroups;
    /** The name of the dimension to compute a similarity on. */
    this._dimension = _dimension;
    /** Optional, the name of the dimension group, if this._acrossGroups === false*/
    this._nameOfDimensionGroup = _nameOfDimensionGroup;

  }

  get acrossGroups() {
    return this._acrossGroups;
  }

  set acrossGroups(value) {
    this._acrossGroups = value;
  }

  get dimension() {
    return this._dimension;
  }

  set dimension(value) {
    this._dimension = value;
  }

  get nameOfDimensionGroup() {
    return this._nameOfDimensionGroup;
  }

  set nameOfDimensionGroup(value) {
    this._nameOfDimensionGroup = value;
  }
}

'use strict';

/**
 * Holds the information about the similarity-based highlighting.
 */
class SimilarityBasedHighlightingDto{

  constructor({_acrossGroups, _nameOfDimensionGroup, _dimension, _sortedDimensions, _threshold}){

    this._acrossGroups = _acrossGroups;
    this._nameOfDimensionGroup = _nameOfDimensionGroup;
    this._dimension = _dimension;
    this._sortedDimensions = _sortedDimensions;
    this._threshold = _threshold;

  }


  get acrossGroups() {
    return this._acrossGroups;
  }

  set acrossGroups(value) {
    this._acrossGroups = value;
  }

  get nameOfDimensionGroup() {
    return this._nameOfDimensionGroup;
  }

  set nameOfDimensionGroup(value) {
    this._nameOfDimensionGroup = value;
  }


  get dimension() {
    return this._dimension;
  }

  set dimension(value) {
    this._dimension = value;
  }

  get sortedDimensions() {
    return this._sortedDimensions;
  }

  set sortedDimensions(value) {
    this._sortedDimensions = value;
  }

  get threshold() {
    return this._threshold;
  }

  set threshold(value) {
    this._threshold = value;
  }
}

'use strict';

/**
 * Directive for the distribution of a numerical dimension, e.g., for using in a table.
 */
angular.module('smartexploreApp')
  .directive('numericalDistribution', function () {
    return {
      template: "<div></div>",
      restrict: 'E',
      scope: {
        w: '=',             // optional, width of the directive
        distribution: '='   // the distribution for the directive.
      },
      link: function (scope, element) {

        scope.$watch('[w, distribution]', function(){
          scope.createPlot();
        });

        /**
         * Creates the actual plot
         */
        scope.createPlot = function(){

          scope.w = scope.w === undefined ? 80 : scope.w;

          if(scope.w !== undefined && scope.distribution !== undefined){

            let selector = d3.select(element[0]);
            selector.selectAll("*").remove();
            let svg = selector.append("svg")
              .attr("class", "LIGHTTHEME")
              .attr("width", Beanplot.size[0])
              .attr("height", Beanplot.size[1]);

            let plot = new Beanplot(svg, scope.distribution, [d3.mean(scope.distribution)]);
            let scaleFactor = scope.w / Beanplot.size[0];
            svg.attr("transform", "scale(" + scaleFactor + ")");

          }

        };
      }
    };
  });

'use strict';

/**
 * Directive for a colormap
 */
angular.module('smartexploreApp')
  .directive('colormap', function (statisticsService) {
    return {
      template: "<div></div>",
      restrict: 'E',
      scope: {
        w: '=',     // width and height of the colormap including labels.
        h: '=',
        label: '=',
        left : '=',
        right : '=',
        glyphLayout : '='
      },
      link: function (scope, element) {

        scope.$watch(function(){return scope.glyphLayout}, function(){
          drawColormap();
        }, true);

        let drawColormap = function(){

          let colormap = undefined;
          if(scope.label.includes("Numerical")){
            let layout = _.findIndex(GLYPH_DESIGN, function(d){return d.id === scope.glyphLayout.numericalDesign});
            colormap = GLYPH_DESIGN[layout].colormap;
          }else{
            let layout = _.findIndex(GLYPH_DESIGN_CATEGORICAL, function(d){return d.id === scope.glyphLayout.categoricalDesign});
            colormap = GLYPH_DESIGN_CATEGORICAL[layout].colormap;
          }

          let values = [];
          for(let i=1; i<=100; i++){
            values.push(i);
          }

          let normalizing = d3.scaleLinear().domain(d3.extent(values)).range([0,1]);

          let selector = d3.select(element[0]);
          selector.selectAll("*").remove();

          let svg = selector.append("svg")
            .attr("class", "LIGHTTHEME")
            .attr("width", scope.w)
            .attr("height", scope.h);

          let sliceWidth = scope.w / values.length;
          let heightForLabel = 11;
          let heightForTitle = 20;
          let heightColormap = scope.h - heightForLabel - heightForTitle;
          svg.selectAll("rect")
            .data(values)
            .enter()
            .append("rect")
            .attr("x", 0)
            .attr("y", 0)
            .attr("width", sliceWidth)
            .attr("height", heightColormap)
            .attr("stroke",0)
            .attr("fill", function(d){return VisualizationProperties.getColorMapValue(normalizing(d), colormap)})
            .attr("transform",function(d,i){return "translate(" + (i * sliceWidth) + "," + heightForTitle + ")"});

          // add the title of the colormap
          svg.append("text")
            .attr("class","label")
            .attr("x", 0)
            .attr("y", 0)
            .attr("text-anchor","middle")
            .attr("alignment-baseline","central")
            .text(scope.label)
            .attr("transform", "translate(" + scope.w/2 + "," + (heightForTitle/2)+ ")");

          // add labels to the colormap
          svg.append("text")
            .attr("class","label")
            .attr("x", 0)
            .attr("y", 0)
            .text(scope.left)
            .attr("transform", "translate(0," + scope.h+ ")");

          svg.append("text")
            .attr("class","label")
            .attr("x", 0)
            .attr("y", 0)
            .attr("text-anchor","end")
            .text(scope.right)
            .attr("transform", "translate(" + scope.w + "," + scope.h+ ")");

          if(scope.glyphLayout._showDimensionIsSignificant && scope.glyphLayout._useSmallColormapForNonSignificantDimensions){

            svg.append("rect")
              .attr("x",0)
              .attr("y",0)
              .attr("height",heightColormap/2)
              .attr("width", VisualizationProperties.MAX_COLORMAP_VALUE_INNER * scope.w)
              .attr("fill","white")
              .attr("transform","translate(" + (scope.w - VisualizationProperties.MAX_COLORMAP_VALUE_INNER * scope.w) + "," + (heightForTitle + heightColormap / 2) + ")");
          }

        };

      }
    };
  });

'use strict';

/**
 * Directive for a colormap
 */
angular.module('smartexploreApp')
  .directive('colormapNumerical', function (statisticsService) {
    return {
      template: "<div></div>",
      restrict: 'E',
      scope: {
        w: '=',     // width and height of the colormap including labels.
        h: '=',
        label: '=',
        middle : '=',
        left : '=',
        right : '=',
        glyphLayout : '='
      },
      link: function (scope, element) {

        scope.$watch(function(){return scope.glyphLayout}, function(){
          drawColormap();
        }, true);

        let drawColormap = function(){

          let colormapDark = scope.glyphLayout.numericalDesign.includes("Dark");

          let values = [];
          for(let i=0; i<=100; i++){
            values.push(i);
          }

          let normalizing = d3.scaleLinear().domain(d3.extent(values)).range([0,1]);

          let selector = d3.select(element[0]);
          selector.selectAll("*").remove();

          let svg = selector.append("svg")
            .attr("class", "LIGHTTHEME")
            .attr("width", scope.w)
            .attr("height", scope.h);

          let sliceWidth = scope.w / (values.length * 2);
          let heightForLabel = 11;
          let heightForTitle = 20;
          let heightColormap = scope.h - heightForLabel - heightForTitle;

          // left side (blue -> white)
          svg
            .append("g")
            .selectAll("rect")
            .data(values.reverse())
            .enter()
            .append("rect")
            .attr("x", 0)
            .attr("y", 0)
            .attr("width", sliceWidth)
            .attr("height", scope.h - heightForLabel - heightForTitle)
            .attr("stroke",0)
            .attr("fill", function(d){
              if(colormapDark){
                return VisualizationProperties.getColorMapValue(normalizing(d), VisualizationProperties.COLORMAP_WHITE_DARKBLUE);
              }else{
                return VisualizationProperties.getColorMapValue(normalizing(d), VisualizationProperties.COLORMAP_WHITE_BLUE);
              }
            })
            .attr("transform",function(d,i){return "translate(" + (i * sliceWidth) + "," + heightForTitle + ")"});

          // right side (white -> red)
          svg
            .append("g")
            .selectAll("rect")
          .data(values.reverse())
          .enter()
          .append("rect")
          .attr("x", 0)
          .attr("y", 0)
          .attr("width", sliceWidth)
          .attr("height", scope.h - heightForLabel - heightForTitle)
          .attr("stroke",0)
          .attr("fill", function(d){
            if(colormapDark){
              return VisualizationProperties.getColorMapValue(normalizing(d), VisualizationProperties.COLORMAP_WHITE_DARKRED);
            }else{
              return VisualizationProperties.getColorMapValue(normalizing(d), VisualizationProperties.COLORMAP_WHITE_RED);
            }
          })
          .attr("transform",function(d,i){return "translate(" + (scope.w/2 + i * sliceWidth) + "," + heightForTitle + ")"});



          // add the title of the colormap
          svg.append("text")
            .attr("class","label")
            .attr("x", 0)
            .attr("y", 0)
            .attr("text-anchor","middle")
            .attr("alignment-baseline","central")
            .text(scope.label)
            .attr("transform", "translate(" + scope.w/2 + "," + (heightForTitle/2)+ ")");

          // add labels to the colormap
          svg.append("text")
            .attr("class","label")
            .attr("x", 0)
            .attr("y", 0)
            .attr("text-anchor","middle")
            .text(scope.middle)
            .attr("transform", "translate(" + scope.w/2 + "," + scope.h+ ")");

          svg.append("text")
            .attr("class","label")
            .attr("x", 0)
            .attr("y", 0)
            .text(scope.left)
            .attr("transform", "translate(0," + scope.h+ ")");

          svg.append("text")
            .attr("class","label")
            .attr("x", 0)
            .attr("y", 0)
            .attr("text-anchor","end")
            .text(scope.right)
            .attr("transform", "translate(" + scope.w + "," + scope.h+ ")");

          if(scope.glyphLayout._showDimensionIsSignificant && scope.glyphLayout._useSmallColormapForNonSignificantDimensions){

            svg.append("rect")
              .attr("x",0)
              .attr("y",0)
              .attr("height",heightColormap/2)
              .attr("width", VisualizationProperties.MAX_COLORMAP_VALUE_INNER * scope.w/2)
              .attr("fill","white")
              .attr("transform","translate(" + (scope.w - VisualizationProperties.MAX_COLORMAP_VALUE_INNER * scope.w/2) + "," + (heightForTitle + heightColormap / 2) + ")");

            svg.append("rect")
              .attr("x",0)
              .attr("y",0)
              .attr("height",heightColormap/2)
              .attr("width", VisualizationProperties.MAX_COLORMAP_VALUE_INNER * scope.w/2)
              .attr("fill","white")
              .attr("transform","translate(" + 0 + "," + (heightForTitle + heightColormap / 2) + ")");
          }

        }

      }
    };
  });

'use strict';

/**
 * Directive for a dynamic tool tip in the visualization.
 */
angular.module('smartexploreApp')
  .directive('numericalTooltipDirective', function () {
    return {
      restrict: 'E',
      scope: {
        data: '=',     // see visualization.js $scope.dataTooltipCategorical
      },
      templateUrl: 'templates/tooltips/templateTooltipNumericalGlyph.html',
      link: function (scope, element) {
        // See templateUrl for details on how to visualize this directive.
      }
    };
  });

'use strict';

/**
 * Directive for a dynamic tool tip in the visualization.
 */
angular.module('smartexploreApp')
  .directive('categoricalTooltipDirective', function () {
    return {
      restrict: 'E',
      scope: {
         data: '=',     // see visualization.js $scope.dataTooltipCategorical
      },
      templateUrl: 'templates/tooltips/templateTooltipCategoricalGlyph.html',
      link: function (scope, element) {
        // See templateUrl for details on how to visualize this directive.
      }
    };
  });

'use strict';

/**
 * Directive for a dynamic tool tip in the visualization.
 */
angular.module('smartexploreApp')
  .directive('dimensionGroupMenuDirective', function () {
    return {
      restrict: 'E',
      templateUrl: 'templates/menu/dimensionGroupMenu.html',
      scope: false, // use the parent scope.
      //controller: 'VisualizationCtrl',
      link: function(scope, element, attribute){

        let originatorEv;
        let self = scope;

        /** Open the menu for the dimension groups. */
        scope.openMenuDimensionGroup = function($mdMenu, ev){

          // we need to init the current set of dimension groups
          scope.initDimensionGroups();

          originatorEv = ev;
          $mdMenu.open(ev);
        };

        // delete dimension group
        scope.menuDimensionGroupRemoveGroup = function(){

          let groupName = d3.select("#dimensionGroupMenu")
            .attr("dimensionGroup");
          // see visualization.js
          self.removeDimensionGroupVis(groupName);
        };

      }
    };
  });


'use strict';

/**
 * Directive for a dynamic menu for the data record groups.
 */
angular.module('smartexploreApp')
  .directive('recordGroupMenuDirective', function () {
    return {
      restrict: 'E',
      templateUrl: 'templates/menu/dataRecordGroupMenu.html',
      scope: false, // use the parent scope.
      //controller: 'VisualizationCtrl',
      link: function(scope, element, attribute){

        let originatorEv;
        let self = scope;

        /** Open the menu for the dimension groups. */
        scope.openMenuRecordGroup = function($mdMenu, ev){

          // we need to init the current set of dimension groups
          // scope.initDimensionGroups();

          originatorEv = ev;
          $mdMenu.open(ev);
        };

        // delete dimension group
        // scope.menuDimensionGroupRemoveGroup = function(){
        //
        //   let groupName = d3.select("#dimensionGroupMenu")
        //     .attr("dimensionGroup");
        //   // see visualization.js
        //   self.removeDimensionGroupVis(groupName);
        // };

      }
    };
  });


'use strict';

/**
 * Directive for the distribution of a numerical dimension, e.g., for using in a table.
 */
angular.module('smartexploreApp')
  .directive('categoricalHistogramDirective', function () {
    return {
      template: "<div></div>",
      restrict: 'E',
      scope: {
        w: '=',              // width of the histogram
        uniqueValues: '=',   // the order of the unique (bins of the histogram, may also be 0)
        distribution: '=',   // the distribution itself, object {value:height in %, value2:height in %, ...)
        labels: '=',        // optional config of the directive
      },
      link: function (scope, element) {

        scope.$watch('[w,uniqueValues,distribution]', function(){
          scope.createHistogram();
        });

        /**
         * Creates the actual histogram.
         */
        scope.createHistogram = function() {

          if (scope.w !== undefined && scope.distribution !== undefined) {

            let uniqueValues = undefined;
            if(scope.uniqueValues === undefined){
              uniqueValues = _.uniq(scope.distribution);
            }else{
              uniqueValues = scope.uniqueValues;
            }

            // create the svg element
            let selector = d3.select(element[0]);
            selector.selectAll("*").remove();
            let svg = selector.append("svg")
              .attr("class", "LIGHTTHEME")
              .attr("width", CategoricalHistogram.getSizeOfHistogram()[0])
              .attr("height", CategoricalHistogram.getSizeOfHistogram()[1]);

            let content = svg.append("g");

            let plot = new CategoricalHistogram(content, scope.distribution, uniqueValues, {'background':true, 'labels':scope.labels !== undefined && scope.labels === true});

            // let plot = new Beanplot(svg, scope.distribution, [d3.mean(scope.distribution)]);
            let scaleFactor = scope.w / CategoricalHistogram.getSizeOfHistogram()[0]; // CategoricalHistogram.getSizeOfHistogram()[0]

            // TODO schau hier; muss überall geändert werden.

            content.attr("transform", "scale(" + scaleFactor + ")");
            svg.attr("width", (scaleFactor * CategoricalHistogram.getSizeOfHistogram()[0]));
            svg.attr("height", (scaleFactor * CategoricalHistogram.getSizeOfHistogram()[1]))

          }

        }

      }
    };
  });

'use strict';

/**
 * Directive for the distribution of a numerical dimension, e.g., for using in a table.
 */
angular.module('smartexploreApp')
  .directive('numericalHistogramDirective', function () {
    return {
      template: "<div></div>",
      restrict: 'E',
      scope: {
        w: '=',              // width of the histogram.
        allValues: '=',      // all values which need to be converted into a histogram.
        numBins: '=',        // the number of bins for the histogram.
        labels: '=',         // optional config of the directive.
        resultHistogramData: '='   // get the histogram data back from the directive
      },
      link: function (scope, element) {

        scope.$watch('[w,allValues,numBins]', function(){
          scope.createHistogram();
        });

        /**
         * Creates the actual histogram.
         */
        scope.createHistogram = function() {

          if (scope.w !== undefined && scope.allValues !== undefined && scope.numBins) {

            // create the svg element
            let selector = d3.select(element[0]);
            selector.selectAll("*").remove();
            let svg = selector.append("svg")
              .attr("class", "LIGHTTHEME")
              .attr("width", NumericalHistogram.getSizeOfHistogram()[0])
              .attr("height", NumericalHistogram.getSizeOfHistogram()[1]);

            let content = svg.append("g");

            let plot = new NumericalHistogram(content, scope.allValues, scope.numBins, {'background':true, 'labels':scope.labels !== undefined && scope.labels === true});

            // let plot = new Beanplot(svg, scope.distribution, [d3.mean(scope.distribution)]);
            let scaleFactor = scope.w / NumericalHistogram.getSizeOfHistogram()[0];

            content.attr("transform", "scale(" + scaleFactor + ")");
            svg.attr("width", (scaleFactor * NumericalHistogram.getSizeOfHistogram()[0]));
            svg.attr("height", (scaleFactor * NumericalHistogram.getSizeOfHistogram()[1]))

            scope.resultHistogramData = plot._storeHistogram;
          }

        }

      }
    };
  });

'use strict';

/**
 * Parent class of {Numerical,Categorical}DimensionVis. Summarizes common method for both dimension types.
 */
class DimensionVis{

  constructor(selector, tooltipService, dimension, glyphLayout, layoutDataRecordGroups, statisticsService, selectionService){

    let self = this;

    this._selector = selector;
    this.statisticsService = statisticsService;
    this._group = selector.append("svg:g").attr("id","dim-" + dimension.dimName);
    this._layoutDataRecordGroups = layoutDataRecordGroups;
    this.numGroups = layoutDataRecordGroups.getNumberOfRecordGroups();
    this.selectionService = selectionService;

    /** Group for the whole dimension. */
    this.glyphGroup = this._group.append("g");

    /** Group for the glyphs. */
    this.groupOfGlyphs = this.glyphGroup.append("g")
      .attr("transform","translate(" + 0 + "," + (VisualizationProperties.GLYPH_SPACE + VisualizationProperties.GLYPH_SIZE) + ")"); // space and +1 for the global distribution of the dimension.

    /** Group for the statistics. */
    this.statisticInfoGroup = this.glyphGroup.append("g")
      .attr("transform","translate(" + 0 + "," + ((VisualizationProperties.GLYPH_SPACE + VisualizationProperties.GLYPH_SIZE) + (this.numGroups) * DimensionGroupVis.computeHeightForGlyph(layoutDataRecordGroups)) + ")");
    this._dimension = dimension; // space and +1 for the global distribution of the dimension.

  }

  get selector() {
    return this._selector;
  }

  get group() {
    return this._group;
  }


  get dimension() {
    return this._dimension;
  }

  set dimension(value) {
    this._dimension = value;
  }

  /** Add the statistics information for the whole dimension. */
  addStatisticSignificanceOfDimension() {

    let statInfo = this.statisticsService.dimensionPandEffect(this._dimension);
    let isSignificant = this.statisticsService.dimensionIsSignificant(this._dimension);

    let green = '#229954';
    let red = '#C0392B';

    if(this._dimension.dimType === 'NUMERICAL') {

      this.statisticInfoGroup.append("rect")
        .attr("x", 0)
        .attr("y", 0)
        .attr("width", VisualizationProperties.GLYPH_SIZE)
        .attr("height", VisualizationProperties.GLYPH_SIZE - 15)
        .attr("fill", isSignificant ? green : red)
        .attr("opacity",0.5)
        .attr("stroke-width", "0.5px")
        .attr("stroke", "gray")
        .attr("transform", "translate(0,15)")
        .append("title")    // TODO implement in a nicer way with more details.
        .text("p: " + statInfo._pValue.toFixed(4) + "\n" + "effect size: " + statInfo._effectSize.toFixed(4));

    }else{

      this.statisticInfoGroup.append("rect")
        .attr("x", 0)
        .attr("y", 0)
        .attr("width", VisualizationProperties.GLYPH_SIZE)
        .attr("height", VisualizationProperties.GLYPH_SIZE - 15)
        .attr("fill", isSignificant ? green : red)
        .attr("opacity",0.5)
        .attr("stroke-width", "0.5px")
        .attr("stroke", "gray")
        .attr("transform", "translate(0,15)")
        .append("title");    // TODO implement in a nicer way with more details.
        //.text("p: " + statInfo._pValue.toFixed(4) + "\n" + "effect size: " + statInfo._effectSize.toFixed(4));

    }

    if(!isSignificant){
      this.addOverlayDimensionNotSignificant();
    }

  }

  /**
   * Return all values (measures) which should be considered to compute the normalization.
   * @param secondLevelEnabled true, if the second aggregation level is enabled; false otherwise.
   */
  getAllValuesForNormalization(secondLevelEnabled){

    if(!secondLevelEnabled){
      // only first aggregation level is enabled.
      return this._dimension.values;

    }else{
      // second aggregation level is enabled as well.
      let numOfSecondLevel = this._layoutDataRecordGroups.getNumberOfRecordGroups2ndLevel();
      let allMeasuresForColormap = this._dimension.values;
      for(let i=0; i<this._dimension.values.length; i++) {
        for (let j = 0; j < numOfSecondLevel; j++) {
          if(this._dimension.values[i].nextLevelAggregation !== undefined && this._dimension.values[i].nextLevelAggregation !== null){
            allMeasuresForColormap.push(this._dimension.values[i].nextLevelAggregation[j]);
          }
        }
      }
      return allMeasuresForColormap;
    }
  }

  /**
   * Add a transparant overlay of dimensions which are not significant.
   */
  addOverlayDimensionNotSignificant(){

    // this.groupOfGlyphs.append("g")
    //   .attr("class","overlay")
    //   .append("rect")
    //   .attr("x",0)
    //   .attr("y",0)
    //   .attr("width", VisualizationProperties.GLYPH_SIZE)
    //   .attr("height", (this.dimension.values.length * VisualizationProperties.GLYPH_SIZE))
    //   .attr("fill","#FFFFFF")
    //   .attr("opacity",0.7);

  }

  addListenerForHighlighting(dimName, secondLevelEnabled){

    let selectionService = this.selectionService;

    this.glyphGroup
      .data([dimName])
      .append("rect")
      .attr("class","dimDistributionHighlightOverlay")
      .attr("x",0)
      .attr("y",0)
      .attr("height", VisualizationProperties.GLYPH_SIZE)
      .attr("width", function(){return secondLevelEnabled ? 2*VisualizationProperties.GLYPH_SIZE : VisualizationProperties.GLYPH_SIZE;})
      .on("mouseover", function(dimName){

        // change mouse icon to finger (indicate, that a click is possible)
        d3.select(this).style("cursor", "pointer");

        // Remove any highlight of the rows
        if(selectionService.isAnyDataRecordGroupSelected()){
          selectionService.removeAllSelectedDataRecordGroups();
          selectionService.updateSelectedDataRecordGroups();
        }

        if(!selectionService.isAnyDimensionSelected()){

          // No dimension is currently selected.
          // Make all dimensions dark, then highlight this dimension.
          selectionService.removeHighlightAllDimension();
          selectionService.addHighlightDimension(dimName);

        }else{
          // There is at least one dimension already selected. So most dimensions are already dark.
          // Highlight only the current dimension.
          selectionService.addHighlightDimension(dimName);
        }

      })
      .on("mouseout", function(){

        // change mouse icon to the default
        d3.select(this).style("cursor", "default");

        selectionService.updateSelectionDimensions();
      })
      .on("click", function(dimName){

        // add or remove dimension from selection.
        if(selectionService.isSelectedDimension(dimName)){
          selectionService.removeSelectedDimension(dimName);
        }else{
          selectionService.addSelectedDimension(dimName);
        }
      });
  }

  /**
   * Returns the width per dimension considering the max level of aggregation.
   * @param maxLevelAggregation the max level of aggregation.
   * @return {Number} or undefined if wrong aggregation level is selected.
   */
  static getWidthPerDimension(maxLevelAggregation) {
    if (maxLevelAggregation === 1) {
      return VisualizationProperties.GLYPH_SIZE;

    } else if (maxLevelAggregation === 2) {
      return 2 * VisualizationProperties.GLYPH_SIZE + VisualizationProperties.GLYPH_SPACE;

    } else {
      console.log("No such aggregation level: " + maxLevelAggregation + ".");
      return undefined;

    }
  }

}

'use strict';

/**
 * Parent class of a {Numerical,Categorical}Glyph.
 */
class Glyph{


  constructor(selector, dimension, glyphLayout, heightOfGlyph){

    this._selector = selector;
    this._group = selector.append("g");
    this._glyphLayout = glyphLayout;
    this._dimension = dimension;
    this._heightOfGlyph = heightOfGlyph !== undefined ? heightOfGlyph : VisualizationProperties.GLYPH_SIZE;

  }

  /**
   *
   * @param color the color of the base glyph (background color.
   * @param significant true, if the value is significant, false otherwise.
   */
  checkOverlayForSignificance(color, significant){

    if(this._glyphLayout.designSignificance === 'dot' && significant){
      this.addOverlayDot(this._group, "#3e3e3e");
    }

    if(this._glyphLayout.designSignificance === 'full' && !significant){
      this.addOverlayInnerRect(color);
    }

  };

  /**
   * Add a glyph with a single color as a glyph (=rect in the given color)
   * @param color the color to choose.
   * @param opacity [optional]
   */
  addSingleColorGlyph(color, opacity){

    opacity = opacity === undefined ? 1.0 : opacity;
    this._group.append("rect")
      .attr("class", "glyph")
      .attr("x",0)
      .attr("y",0)
      .attr("width", VisualizationProperties.GLYPH_SIZE)
      .attr("height", this._heightOfGlyph)
      .attr("fill", color)
      .attr("opacity", opacity);
  }

  /**
   * Add a glyph in which a full rect = significant, small rect = not significant.
   * @param color the color to choose.
   */
  addOverlayInnerRect(color){

    let sizeInnerRect = 0.6;
    let opacityForOverlay = 0.85;

    // transparent figure on top, reduce color intensity
    this.addSingleColorGlyph("white", opacityForOverlay);

    // rect in middle of glyph, use once again the full color
    let w = sizeInnerRect * VisualizationProperties.GLYPH_SIZE;
    let h = sizeInnerRect * this._heightOfGlyph;
    this._group.append("rect")
      .attr("x", 0)
      .attr("y", 0)
      .attr("width", w)
      .attr("height", h)
      .attr("fill", color)
      .attr("transform", "translate(" + (VisualizationProperties.GLYPH_SIZE / 2 - w / 2) + "," + (this._heightOfGlyph / 2 - h / 2) + ")");

  }

  /**
   * Add an overlay (dot) to the glyph.
   * @param parent d3.selector to append the glyph to.
   * @param color the color of the dot.
   * @param opacity [optional]
   */
  addOverlayDot(color, opacity){

    opacity = opacity === undefined ? 1.0 : opacity;

    let w = 0.3 * VisualizationProperties.GLYPH_SIZE;
    let h = 0.3 * this._heightOfGlyph;

    this._group.append("rect")
      .attr("class", "typeOfDistanceRect")
      .attr("x", 0)
      .attr("y", 0)
      .attr("width", w)
      .attr("height", h)
      .attr("fill", color)
      .attr("opacity", opacity)
      .attr("transform", "translate(" + (VisualizationProperties.GLYPH_SIZE / 2 - w / 2) + "," + (this._heightOfGlyph / 2 - h / 2) + ")");
  }

  /**
   * Add an overlay (dot) to the glyph.
   * @param parent d3.selector to append the glyph to.
   * @param color the color of the dot.
   */
  addOverlayLine(color, opacity){

    opacity = opacity === undefined ? 1.0 : opacity;

    this._group.append("line")
      .attr("class","typeOfDistancePos")
      .attr("x1", 2)
      .attr("y1", this._heightOfGlyph - 2)
      .attr("x2", VisualizationProperties.GLYPH_SIZE - 2)
      .attr("y2",  2)
      .attr("stroke", color)
      .attr("opacity",opacity);
  }

  /**
   * Add a transparent rectangle, which can be used to highlight the glyphs.
   */
  addHighlightingOverlay(){

    // for the dimension highlighting
    this._group.append("rect")
      .attr("class", "highlightOverlay highlight")
      .attr("x",0)
      .attr("y",0)
      .attr("width", VisualizationProperties.GLYPH_SIZE)
      .attr("height", this._heightOfGlyph)

    // for the row highlighting
    this._group.append("rect")
      .attr("class", "highlightOverlayRow highlight")
      .attr("x",0)
      .attr("y",0)
      .attr("width", VisualizationProperties.GLYPH_SIZE)
      .attr("height", this._heightOfGlyph)

  }

  /**
   * Add the missing values as overlay to the visualization.
   *
   * @param dimension the dimension
   * @param glyphData the glyph data
   * @param missingValueStyle the style of the missing values, possible values are:
   *        - disabled
   *        - noise
   *        - triangle
   */
  addMissingValues(dimension, glyphData, missingValueStyle){

    let mv = glyphData.percentageMissingValues;
    let margin = 2;
    let size = VisualizationProperties.GLYPH_SIZE - 2*margin;

    // missing value using random noise.
    if(mv > 0 && missingValueStyle === 'noise') {


      let numPixelsForGlyph = size * size;
      let numPixelsMV = numPixelsForGlyph * glyphData.percentageMissingValues;
      let randomPoints = [];
      for(let i=0; i<numPixelsMV; i++){
        randomPoints.push({
          'x': Math.random() * size,
          'y': Math.random() * size
        })
      }

      this._group.append("g")
        .attr("class", "missingvalues")
        .selectAll(".missingValuePoints")
        .data(randomPoints)
        .enter()
        .append("rect")
        .attr("x",0)
        .attr("y",0)
        .attr("width",1)
        .attr("height",1)
        .attr("fill","#2E4053")
        .attr("opacity","0.65")
        .attr("transform", function(d){ return "translate(" + (margin + d.x) + "," + (margin + d.y) + ")"; });


    }else if(mv > 0 && missingValueStyle === 'triangle'){
      // missing value using a triangle as overlay.


      let path = undefined;

      if(mv <= 0.5){

        // only upper part needs to be considered.
        let length = mv * 2 * size;
        path = "M" + size + " 0 L" + size + " " + length + " L" + (size - length) + " 0 Z";

      }else{
        // upper part and part of the lower triangle.
        let length = (mv-0.5) * 2 * size;
        let pos = [];
        pos.push([size, 0]);
        pos.push([size, size]);
        pos.push([size-length, size]);
        pos.push([0, length]);
        pos.push([0, 0]);

        path = "M" + pos[0][0] + " " + pos[0][1] +
              " L" + pos[1][0] + " " + pos[1][1] +
              " L" + pos[2][0] + " " + pos[2][1] +
              " L" + pos[3][0] + " " + pos[3][1] +
              " L" + pos[4][0] + " " + pos[4][1] +
              " Z";

      }

      this._group.append("g")
        .attr("class", "missingvalues")
        .append("path")
        .attr("d",path)
        .attr("fill","#2E4053")
        .attr("opacity","0.65")
        .attr("transform", "translate(" + (margin) + "," + (margin) + ")");


    }

  };

  addTooltipDetails(tooltipService, dimension, glyphData, recordGroupName){

    this._group.append("rect")
      .attr("class", "glyph")
      .attr("x",0)
      .attr("y",0)
      .attr("width", VisualizationProperties.GLYPH_SIZE)
      .attr("height", this._heightOfGlyph)
      .attr("fill", "red")
      .attr("opacity",0.0)
      .on("mouseenter", function() {
        if(dimension.dimType === 'NUMERICAL'){
          tooltipService.showNumericalGlyphTooltip([d3.event.pageX-5,d3.event.pageY-5], dimension, glyphData, recordGroupName);
        }else if(dimension.dimType === 'CATEGORICAL'){
          tooltipService.showCategoricalGlyphTooltip([d3.event.pageX-5,d3.event.pageY-5], dimension, glyphData, recordGroupName);
        }
      })
      .on("mouseleave", function() {
        if(dimension.dimType === 'NUMERICAL'){
          tooltipService.hideNumericalGlyphTooltip();
        }else if(dimension.dimType === 'CATEGORICAL'){
          tooltipService.hideCategoricalGlyphTooltip();
        }
      });

  }

  get selector() {
    return this._selector;
  }

  get group() {
    return this._group;
  }


  get dimension() {
    return this._dimension;
  }

  get glyphLayout() {
    return this._glyphLayout;
  }
}

'use strict';

class DimensionGroupVis{

  /**
   * The visualization of one group of dimensions. The css class for a whole group is ".dimensionGroupVis".
   *
   * @param selector a d3.selector to attach the visualization to.
   * @param groupName the name of the dimension group, also a unique identifier of this group.
   * @param tooltipService
   * @param featureTable data of the feature table
   * @param glyphLayout
   * @param layoutDataRecordGroups
   * @param layoutDimensionGroups
   * @param statisticsService the service for the statistics, used to select the correct statistical test.
   */
  constructor(selector, groupName, tooltipService, featureTable, glyphLayout, layoutDataRecordGroups, layoutDimensionGroups, statisticsService, selectionService){

    let group = _.findWhere(layoutDimensionGroups.dimensionGroups, {"name": groupName});
    let dimensions = group.dimensions;
    let maxAggregationLevel = layoutDataRecordGroups._maxLevelOfAggregation;

    this._selector = selector;
    this._group = selector
      .data([{'groupName': groupName}])   // bind the group name to this element.
      .append("svg:g")
      .attr("id", "dimensionGroup-" + groupName)
      .attr("class","dimensionGroupVis");

    let self = this._group;

    // radius for rounded corners.
    const radius = 10;

    let widthForDimensions = dimensions.length * DimensionVis.getWidthPerDimension(maxAggregationLevel);
    let heightForGroup = layoutDataRecordGroups.getNumberOfRecordGroups() * DimensionGroupVis.computeHeightForGlyph(layoutDataRecordGroups) + 1 * (VisualizationProperties.GLYPH_SIZE + VisualizationProperties.GLYPH_SPACE); // +1 for the global distribution.

    // if the statistic of the dimension is enabled, we need more space to visualize it.
    if(glyphLayout._showDimensionIsSignificant){
      heightForGroup += (VisualizationProperties.GLYPH_SIZE + VisualizationProperties.GLYPH_SPACE);
    }

    let widthRect = widthForDimensions + 2*radius;
    let heightRect = heightForGroup + 2*radius;


    // only the border
    let border = this._group.append("rect")
      .attr("class", "groupElement groupElementBorder")
      .attr("x", 0)
      .attr("y", 0)
      .attr("rx", radius)
      .attr("ry", radius)
      .attr("width", widthRect)
      .attr("height", heightRect);

    // remove the inner / middle part of the border
    this._group.append("rect")
      .attr("class", "groupElement groupTmp")
      .attr("x", 0)
      .attr("y", 0)
      .attr("width", widthForDimensions)
      .attr("height", heightRect - 2) // do not remove the bottom part.
      .attr("transform","translate(" + radius + "," + 0 + ")");

    // overlay for capturing the mouse events do highlighting.
    this._group.append("rect")
      .attr("class", "groupElement groupElementOverlay")
      .attr("x", 0)
      .attr("y", 0)
      .attr("rx", radius)
      .attr("ry", radius)
      .attr("width", widthRect)
      .attr("height", heightRect)
      .on("mouseover", function(){highlight(self); showMenuForGroup();})
      .on("mouseout", function(){ removeHighlight(self);});

    // draw inner element to prevent mouse events in the gaps between the cells
    this._group.append("rect")
      .attr("class", "groupElement groupElementOverlay")
      .attr("x", radius)
      .attr("y", radius)
      .attr("width", widthRect - 2*radius)
      .attr("height", heightRect - 2*radius)
      .attr("fill", "white");


    // ======================================
    // draw the actual dimensions and the labels
    let g = this._group.append("g");
    let labelGroup = this._group.append("g");
    for(let i=0; i<dimensions.length; i++){

      let dimName = dimensions[i];
      let xPosition = group.dimensionToSimilarityPos[dimName];

      // center of the dimension (also for 2nd level aggregation)
      let xpos = radius + (xPosition * DimensionVis.getWidthPerDimension(maxAggregationLevel)) + (DimensionVis.getWidthPerDimension(maxAggregationLevel) / 2);


      // ---------------------------------------------------
      // add label
      new DimensionLabel(labelGroup, dimensions[i], [xpos, 0]);

      let dim = featureTable._nameToDimension[dimensions[i]];


      // ---------------------------------------------------
      // circle for the menu of this dimension.
      let circleSize = 5;
      labelGroup.append("circle")
        .attr("cx", 0)
        .attr("cy", 0)
        .attr("r", circleSize)
        .attr("fill", "#c0c6ca")
        .attr("transform", "translate(" + (xpos - circleSize/2) + "," + (VisualizationProperties.GLYPH_SIZE/6) + ")");


      let dimGroup;
      if(dim.dimType === "NUMERICAL"){
        dimGroup = new NumericalDimensionVis(g, tooltipService, dim, glyphLayout, layoutDataRecordGroups, statisticsService, selectionService)._group;

      }else if (dim.dimType === "CATEGORICAL") {
        dimGroup = new CategoricalDimensionVis(g, tooltipService, dim, glyphLayout, layoutDataRecordGroups, statisticsService, selectionService)._group;
      }

      if(dimGroup !== undefined) {
        dimGroup.attr("transform", "translate(" + (radius + xPosition * DimensionVis.getWidthPerDimension(maxAggregationLevel)) + "," + (radius) + ")");
      }

    }


    // ======================================
    // the name of the group (including background which removes part of the bottom border)

    // compute the expected width of the text
    let canvas = document.createElement('canvas');
    let ctx = canvas.getContext("2d");
    ctx.font = "12px Lato";                // c.f. featureTableLight.css .dimGroupName
    let widthText = 20 + ctx.measureText(groupName).width;
    // let widthText = 20 + ctx.measureText("Group: " + groupName).width;

    this._group.append("rect")
      .attr("class","dimGroupBackground")
      .attr("x", 0)
      .attr("y", 0)
      .attr("width", widthText)
      .attr("height",5)
      .attr("transform","translate(" + (widthRect / 2 - (widthText / 2)) + "," + (heightRect - 2) + ")");

    this._group.append("text")
      .attr("class", "label dimGroupName")
      .attr("x", 0)
      .attr("y", 0)
      .text(groupName)
      //.text("Group: " + groupName)
      .attr("transform","translate(" + (widthRect / 2) + "," + (heightRect + 4) + ")");


    /**
     * Highlight the given group.
     * @param object
     */
    function highlight(object){
      object.selectAll(".groupElement").classed("highlight", true);
      object.style("cursor", "move");
    }

    function removeHighlight(object){
      object.selectAll(".groupElement").classed("highlight", false);
      object.style("cursor", "default");

      // remove the menu for the group is done in the svg element itself.
    }

    /**
     * Show the menu for this group.
     */
    function showMenuForGroup(){
      // show the menu for this group
      let pos = border.node().getBoundingClientRect();
      let x = window.scrollX + pos.x + widthForDimensions;
      let y = window.scrollY + pos.y;
      DimensionGroupVis.showMenuForGroup(groupName, x, y);
      d3.select("#dimensionGroupMenu")
    }

  }


  get selector() {
    return this._selector;
  }

  get group() {
    return this._group;
  }

  /**
   * Compute the expected width of a group with the given number of dimensions.
   * @param numDimensions the number of dimensions.
   * @param maxLevelOfAggregation the max aggregation level which is selected. Possible values: 1, 2
   */
  static computeWidthForNumberOfDimensions(numDimensions, maxLevelOfAggregation){

    // default values.
    maxLevelOfAggregation = maxLevelOfAggregation !== undefined ? maxLevelOfAggregation : 1;

    // the radius of the group; see DimensionGroupVis.
    let radius = 10;

    if(maxLevelOfAggregation === 1){
      return 2 * radius + numDimensions * VisualizationProperties.GLYPH_SIZE;

    }else if(maxLevelOfAggregation === 2){
      let width = 2 * radius;
      width += numDimensions * 2 * VisualizationProperties.GLYPH_SIZE;
      width += numDimensions * VisualizationProperties.GLYPH_SPACE;
      return width;

    }else{
      console.log("No such aggregation level allowed yet.");
      return undefined;
    }

  }

  /**
   * Computes the height of a single glyph depending on the max aggregation level.
   * @param layoutDataRecorcGroups - the layout of the data record groups.
   * @return {Number} or undefined, if the aggregation level is unknown.
   */
  static computeHeightForGlyph(layoutDataRecorcGroups){

    if(layoutDataRecorcGroups._maxLevelOfAggregation === 1){
      return VisualizationProperties.GLYPH_SIZE;
    }

    if(layoutDataRecorcGroups._maxLevelOfAggregation === 2){
      // number of groups for second level
      return layoutDataRecorcGroups.getNumberOfRecordGroups2ndLevel() * (VisualizationProperties.GLYPH_SIZE + VisualizationProperties.GLYPH_SPACE);
    }

    console.log("No such aggregation level.");
    return undefined;

  }

  /**
   * Computes the location (x-position) of the dimension groups.
   * @param dimensionGroups the dimension groups from scope.layoutDimensionGroups._dimensionGroups in the correct order.
   * @param maxLevelAggregation the max aggregation level which is visualized, currently supported 1 or 2.
   * @return {} object {grouName: x-pos}
   */
  static computeLocationOfGroups(dimensionGroups, maxLevelAggregation){

    // offset of the table to draw, space for labels, one space and then all groups.
    let offset = 0;
    let positions = {};

    for(let i=0; i<dimensionGroups.length; i++){
      let group = dimensionGroups[i];
      let widthOfGroup = DimensionGroupVis.computeWidthForNumberOfDimensions(group.dimensions.length, maxLevelAggregation);
      positions[group.name] = offset;
      offset += VisualizationProperties.GROUP_SPACE;
      offset += widthOfGroup;
    }

    return positions;
  }

  /**
   * Computes the new location of the dimension groups and translates them to it.
   * No return value.
   * @param dimensionGroups the dimension groups from scope.layoutDimensionGroups._dimensionGroups in the correct order.
   * @param maxLevelAggregation the max aggregation level which is visualized, currently supported 1 or 2.
   */
  static updatePositionOfGroups(dimensionGroups, maxLevelAggregation){

    // Update the position.
    let positionsOfGroups = DimensionGroupVis.computeLocationOfGroups(dimensionGroups, maxLevelAggregation);
    d3.selectAll(".dimensionGroupVis")
      .transition()
      .duration(2000)
      .attr("transform",function(d){
        return "translate( " + positionsOfGroups[d.groupName] + ", " + (0) + ")";
      });
  }

  /**
   * Show the menu for the dimension group with the given name.
   * @param dimensionGroupName the name of the dimension group.
   * @param x x-pos
   * @param y y-pos
   */
  static showMenuForGroup(dimensionGroupName, x, y){
    d3.select("#dimensionGroupMenu")
      .style("left", x + "px")
      .style("top", y + "px")
      .style("display", "inline")
      .attr("dimensionGroup", dimensionGroupName);
  }

  /**
   * Remove the menu for all dimension groups.
   */
  static removeMenuForGroup(){
    d3.select("#dimensionGroupMenu")
      .style("display", "none");
  }



}

/**
 * Add Drag & Drop Bahavior to all dimension groups.
 */
function addDragAndDropToAllDimensionGroups(urlParameterService, rootscope){

  let dragDrop = d3.drag()
    .on("start", dragGroupStarted)
    .on("drag", draggingGroup)
    .on("end", dragGroupSEnded);

  d3.selectAll(".dimensionGroupVis").call(dragDrop);

  // the distance from the x-pos of the mouse to the left border of the object.
  let offsetMouseToLeftBorderGroup = 0;
  let offsetY = 0;

  /**
   * Start dragging a dimension group.
   */
  function dragGroupStarted() {
    let t = d3.select(this).attr("transform");
    let translate = t.substring(t.indexOf("(")+1, t.indexOf(")")).split(",");
    offsetMouseToLeftBorderGroup = d3.event.x - translate[0];
    offsetY = translate[1];
  }

  /**
   * Currently draging a dimension group. Update x position.
   */
  function draggingGroup() {
    d3.select(this)
      .attr("transform", "translate(" + (d3.event.x - offsetMouseToLeftBorderGroup) + "," + offsetY + ")");
  }

  /**
   * Dragging of a dimension group is complete.
   * Store the new order of the dimension groups.
   * Store that information in the URL.
   */
  function dragGroupSEnded() {

    // get all groups, sort them by the transformed x position
    let groupPos = {};
    let allGroups = Array.prototype.slice.call(d3.selectAll(".dimensionGroupVis")._groups[0]);
    for(let i=0; i<allGroups.length; i++) {
      let g = d3.select(allGroups[i]);
      let name = g.attr("groupName");
      let translate = g.attr("transform").substring(g.attr("transform").indexOf("(")+1, g.attr("transform").indexOf(")")).split(",");
      groupPos[name] = +translate[0];
    }

    let currentLayoutDimensionGroups = urlParameterService.getLayoutDimensionGroups();
    let currentLayoutDataRecordGroups = urlParameterService.getLayoutDataRecordGroups();

    // sort the groups correctly.
    currentLayoutDimensionGroups._dimensionGroups = _.sortBy(currentLayoutDimensionGroups._dimensionGroups, function(g){
      return groupPos[g._name];
    });

    // move the groups to the correct location. Use some transitions.
    DimensionGroupVis.updatePositionOfGroups(currentLayoutDimensionGroups._dimensionGroups, currentLayoutDataRecordGroups.maxLevelOfAggregation);

    // Update the parameters in the URL.
    rootscope.$apply(function(){
      urlParameterService.setLayoutDimensionGroups(currentLayoutDimensionGroups);
    });

  }

}

'use strict';

// The width and height of the vis.
const WIDTH = 100, HEIGHT = 80;

class Beanplot{

  /**
   * Creates a new BeanPlot. The min value is on the left, the max value on the right (=flipped BeanPlot)
   * The distribution is shown only on the top. The width and height of the plot is 100*40 and can scaled easily.
   *
   * Use the array with the marks to visualize specific marks in the visualization, e.g. multiple mean vlaues.
   * The first mark will be visualized in green, the last in red, inbetween in dark gray.
   *
   * @param selector The selector to bind the visualization on.
   * @param arrayWithData an array with numerical data to be visualized.
   * @param marks data which should be marked in the vis
   * @param config optional config parameter. Contains the following info:
   *        config.background = {true, false}   // draw background or not
   *        config.domain = [min, max]          // domain of the bean plot.
   *
   */
  constructor(selector, arrayWithData, marks, config){

    this._selector = selector;
    this._group = selector.append("g").attr("class","BEANPLOT");

    // number of bins for the histogram; used also to compute the kernel density estimation.
    let numBins = 5;
    let bandwidth = estimateBandwidth(arrayWithData);
    let extent = d3.extent(arrayWithData);

    let widthPerBin = WIDTH / numBins;
    let xScaleDensityForComputation = d3.scaleLinear()
      .domain([extent[0], extent[1]])
      .range([widthPerBin/2, WIDTH-widthPerBin/2]);   // move to center of the bin.

    let density = kernelDensityEstimator(kernelEpanechnikov(bandwidth), xScaleDensityForComputation.ticks(numBins - 1))(arrayWithData);
    let maxDensity = d3.max(density, function (d) { return d[1]; });

    let yScaleDensity = d3.scaleLinear().domain([0, maxDensity]).range([0, HEIGHT]);

    // smooth the beginning and end of the density plot
    // begin: x=start of plot, y=0.5*height of the first value (dito for last bin)
    let step = density[1]['0'] - density[0]['0'];
    let xMinDensity = density[0]['0'] - step/2;
    let xMaxDensity = density[density.length - 1]['0'] + step/2;
    density.unshift([xMinDensity, density[0]['1']/2]);
    density.push([xMaxDensity, density[density.length - 1]['1']/2]);

    // check if the given domain is smaller than the min and larger than the max value
    let domain = (config === undefined || config !== undefined && config.domain === undefined) ? extent : config.domain;
    if(domain[0] < xMinDensity){
      xMinDensity = domain[0];
      density.unshift([xMinDensity, 0]);
    }

    if(domain[1] > xMaxDensity){
      xMaxDensity = domain[1];
      density.push([xMaxDensity, 0]);
    }


    let x = d3.scaleLinear()
      .domain([xMinDensity,xMaxDensity])
      .range([0, WIDTH]);


    // optional background
    if(config === undefined || config !== undefined && config.background === true) {
      this._group.append("rect")
        .attr("class", "background")
        .attr("x", 0)
        .attr("y", 0)
        .attr("width", WIDTH)
        .attr("height", HEIGHT);
    }

    // this is the histogram; optional
    // this._group.insert("g", "*")
    //   .attr("fill", "black")
    //   .selectAll("rect")
    //   .data(bins)
    //   .enter().append("rect")
    //   .attr("x", function(d) { return x(d.x0); })
    //   .attr("y", function(d) { return yScaleHistogram(d.length); })
    //   .attr("width", function(d) { return x(d.x1) - x(d.x0); }) // was return x(d.x1) - x(d.x0) - 1
    //   .attr("height", function(d) { return yScaleHistogram(0) - yScaleHistogram(d.length); })
    //   .attr("stroke-width", "1px")
    //   .attr("stroke", "black")
    //   .attr("fill","darkgray");

    this._group.append("path")
      .datum(density)
      // .attr("fill", "none")
      // .attr("stroke", "#000")
      // .attr("stroke-width", 1.5)
      // .attr("stroke-linejoin", "round")
      .attr("d",  d3.area()
        .curve(d3.curveCatmullRom.alpha(0.5))
        .x(function (d) {
          return x(d[0]);
        })
        .y0(HEIGHT)
        .y1(function (d) {
          return HEIGHT-yScaleDensity(d[1]);
        }));


    // add marks in the distribution.
    // if(marks !== undefined && marks.length !== 0) {
    //
    //   if (marks.length >= 1) {
    //
    //     this._group.append("rect")
    //       .attr("class", "markLower")
    //       .attr("x", 0)
    //       .attr("y", 0)
    //       .attr("width", "6px")
    //       .attr("height", HEIGHT / 3)
    //       .attr("transform", "translate(" + (x(marks[0]) -3) + "," + (HEIGHT - HEIGHT / 3) + ")");
    //
    //   }
    //
    //   if (marks.length >= 2) {
    //
    //     this._group.append("rect")
    //       .attr("class", "markUpper")
    //       .attr("x", 0)
    //       .attr("y", 0)
    //       .attr("width", "6px")
    //       .attr("height", HEIGHT / 3)
    //       .attr("transform", "translate(" + (x(marks[marks.length - 1]) -3) + "," + (HEIGHT - HEIGHT / 3) + ")");
    //
    //     for (let i = 1; i < marks.length - 1; i++) {
    //       this._group.append("rect")
    //         .attr("class", "mark")
    //         .attr("x", 0)
    //         .attr("y", 0)
    //         .attr("width", "1px")
    //         .attr("height", HEIGHT / 5)
    //         .attr("transform", "translate(" + x(marks[i]) + "," + (HEIGHT - HEIGHT / 5) + ")");
    //     }
    //   }
    // }


    /**
     * Taken from: https://bl.ocks.org/mbostock/4341954
     * @param kernel the kernel to be used.
     * @param X
     * @returns {Function}
     */
    function kernelDensityEstimator(kernel, X) {
      return function(V) {
        return X.map(function(x) {
          return [x, d3.mean(V, function(v) { return kernel(x - v); })];
        });
      };
    }

    /**
     * Taken from: https://bl.ocks.org/mbostock/4341954
     * @param k
     * @returns {Function}
     */
    function kernelEpanechnikov(k) {
      return function(v) {
        return Math.abs(v /= k) <= 1 ? 0.75 * (1 - v * v) / k : 0;
      };
    }

    /**
     * Estimate / proposes the bandwith parameter for the kernel density estimation as suggested here
     * https://en.wikipedia.org/wiki/Kernel_density_estimation
     * @param samples all values which should be plotted.
     */
    function estimateBandwidth(samples){

      let std = d3.deviation(samples);
      let n = samples.length;
      return Math.pow((4 * Math.pow(std,5)) / (3 * n) ,(1/5));

    }

    /**
     * Estimate the number of ticks / bins for the kernel density estimation.
     * Based on d3.js histogram function.
     * @param samples all values which should be plotted.
     */
    function estimateNumberOfTicks(samples){

      let h = d3.histogram()(samples);
      return h.length;

    }

  }

  get selector() {
    return this._selector;
  }

  get group() {
    return this._group;
  }

  /**
   * Return the size [width, height] of this component.
   */
  static get size(){
    return [WIDTH, HEIGHT];
  }

}

'use strict';

/** SVG element for a dimension label in the visualization. */
class DimensionLabel{

  /**
   * @param selector the selector to attach this label to
   * @param dimName the name of the dimesion (text)
   * @param position the position, position=[x,y]. The label will automatically be rotated by -45°
   */
  constructor(selector, dimName, position){

    this._selector = selector;

    this._group = selector.append("text")
      .attr("class", "label dimName")
      .attr("x", 0)
      .attr("y", 0)
      .text(dimName)
      .attr("transform", "translate(" + position[0] + "," + position[1] + ") rotate(-45) translate(0,-2)"); // 5,2 only to move the label to the correct position.
  }


  get selector() {
    return this._selector;
  }

  get group() {
    return this._group;
  }
}

'use strict';

/** Width and height of a glyph. */
const GLYPH_SIZE = 30;
/** The space between two glyhs (horizontally and vertically). */
const GLYPH_SPACE = 3;

/** The space between the small distribution chart above each dimension and the dimension itself. */
const SPACE_DISTRIBUTION_GLYPH = 3;

/** If this.useSmallColormapForNonSignificantDimensions = true, use this value as max value for the colormap. */
const MAX_COLORMAP_VALUE_INNER = 0.5;

/** The space between two dimension groups */
const GROUP_SPACE = 50;

const COLORMAP_WHITE_RED = ['#f7f7f7','#ef8a62'];
const COLORMAP_WHITE_DARKRED = ['#f7f7f7','#b54112'];

const COLORMAP_WHITE_BLUE = ['#f7f7f7','#67a9cf'];
const COLORMAP_WHITE_DARKBLUE = ['#f7f7f7','#296283'];

const COLORMAP_BIPOLAR = ['#67a9cf', '#f7f7f7', '#ef8a62'];
const COLORMAP_BIPOLAR_DARK = ['#296283', '#f7f7f7', '#b54112'];

/**
 * Class is only for syntactic sugar.
 */
class VisualizationProperties{

  static get GLYPH_SIZE(){
    return GLYPH_SIZE;
  }

  static get GLYPH_SPACE(){
    return GLYPH_SPACE;
  }

  static get GROUP_SPACE(){
    return GROUP_SPACE;
  }

  static get SPACE_DISTRIBUTION_GLYPH(){
    return SPACE_DISTRIBUTION_GLYPH;
  }

  static get COLORMAP_WHITE_RED(){
    return COLORMAP_WHITE_RED;
  }

  static get COLORMAP_WHITE_BLUE(){
    return COLORMAP_WHITE_BLUE;
  }

  static get COLORMAP_WHITE_DARKRED(){
    return COLORMAP_WHITE_DARKRED;
  }

  static get COLORMAP_WHITE_DARKBLUE(){
    return COLORMAP_WHITE_DARKBLUE;
  }

  static get COLORMAP_BIPOLAR(){
    return COLORMAP_BIPOLAR;
  }

  static get COLORMAP_BIPOLAR_DARK(){
    return COLORMAP_BIPOLAR_DARK;
  }

  static get MAX_COLORMAP_VALUE_INNER(){
    return MAX_COLORMAP_VALUE_INNER;
  }

  /**
   * Assume the data is normalized in range [0,1]
   * @param value the value you want to get the color for.
   * @param colormap optional, value for the colormap.
   */
  static getColorMapValue(value, colormap=undefined){

    let heatmapColour = undefined;

    if(colormap !== undefined){

      if(colormap.length === 2){
        heatmapColour = d3.scaleLinear()
          .domain([0.0, 1.0])
          .range(colormap);
      }else if(colormap.length === 3){
        heatmapColour = d3.scaleLinear()
          .domain([0.0, 0.5, 1.0])
          .range(colormap);
      }else{
        console.error("VisualizationProperties.js wrong colormap.");
        return "black";
      }

    }else {
      heatmapColour = d3.scaleLinear()
        // .domain([0.0, 0.5, 1.0])   // strong colormap
        .domain([0.0, 1.0])
        .range(COLORMAP_WHITE_RED);
    }

    return heatmapColour(value);
  }

}

'use strict';

/**
 * A numerical dimension with the glyphs for all groups.
 */
class NumericalDimensionVis extends DimensionVis{

  /**
   * @param selector a d3.selector to attach the visualization to.
   * @param tooltipService the service used to manage the different tooltips. See services/tooltipService.js
   * @param dimension the dimension information; see featureTable#_nameToDimension[dimName]
   *        - dimName
   *        - dimType
   *        - globalAllValues
   *        - globalMax, Mean, Median, Min, MissingValues, StandardDeviation, Variance
   *        - values [g1, g2, ..., gn]
   *          g_i = {
   *            allValues,
   *            min,max,median,missingValues,standardDeviation,variance
   *          }
   * @param glyphLayout {GlyphLayout} the layout of the glyph.
   * @param layoutDataRecordGroups {LayoutDataRecordGroups} the layout of the data record groups.
   * @param statisticsService the service for the statistics, used to select the correct statistical test.
   */
  constructor(selector, tooltipService, dimension, glyphLayout, layoutDataRecordGroups, statisticsService, selectionService){

    super(selector, tooltipService, dimension, glyphLayout, layoutDataRecordGroups, statisticsService, selectionService);

    let dimName = dimension.dimName;

    // info whether to visualize the second level or not.
    let secondLevelEnabled = layoutDataRecordGroups._maxLevelOfAggregation > 1;
    let numOfSecondLevel = layoutDataRecordGroups.getNumberOfRecordGroups2ndLevel();

    // min and max values from all data records; used to normalize the distribution plots
    let globalExtent = d3.extent(dimension.globalAllValues);
    let globalDiff = globalExtent[1] - globalExtent[0];
    let globalDomain = [globalExtent[0] - 0.1*globalDiff, globalExtent[1] + 0.1*globalDiff];


    // ---------------------------------------------------------------------
    // background of the distribution glyph
    this.glyphGroup
      .append("rect")
      .attr("class","backgroundDimDistribution highlight")
      .attr("x",0)
      .attr("y",0)
      .attr("height", VisualizationProperties.GLYPH_SIZE)
      .attr("width", function(){return secondLevelEnabled ? 2*VisualizationProperties.GLYPH_SIZE : VisualizationProperties.GLYPH_SIZE;});

    // -----------------
    // glyph for the distribution of the whole dimension + scale to the correct size.
    let marks = d3.extent(dimension.values, function(d){return d.measure});
    let distanceBetweenHistograms = VisualizationProperties.SPACE_DISTRIBUTION_GLYPH;
    let w = secondLevelEnabled ? VisualizationProperties.GLYPH_SIZE/2 : 0;
    let dimDistribution = new Beanplot(this.glyphGroup, dimension.globalAllValues, marks, {"background": true, "domain": globalDomain})._group;
    let sizeOfDistribtuion = Beanplot.size;

    // scale to the correct size + move to the bottom.
    let scale = (VisualizationProperties.GLYPH_SIZE - distanceBetweenHistograms) / sizeOfDistribtuion[0]; // also translate - make sure there is a small gap between the histograms.
    dimDistribution.attr("transform","translate(" + w + ",0) scale(" + scale + ") translate(" + (distanceBetweenHistograms) + "," + (VisualizationProperties.GLYPH_SIZE - (VisualizationProperties.GLYPH_SIZE - (sizeOfDistribtuion[1] * scale))) + ")");

    // -----------------
    // listener for highlighting
    super.addListenerForHighlighting(dimName, secondLevelEnabled);

    let heightPerGlyph = secondLevelEnabled ? numOfSecondLevel * VisualizationProperties.GLYPH_SIZE : VisualizationProperties.GLYPH_SIZE;
    let heightPerGlyphWithSpace = DimensionGroupVis.computeHeightForGlyph(layoutDataRecordGroups);

    // iterate over all rows of the visualization
    for(let i=0; i<this.numGroups; i++){

      let glyph = undefined;
      let glyphGroup = this.groupOfGlyphs.append("g");

      // check if the current value is empty or not.
      if(dimension.values[i].aggregatedValueEmpty === true){

        glyph = new EmptyGlyph(glyphGroup, dimension, dimension.values[i], glyphLayout, layoutDataRecordGroups._dataRecordGroups[i]._label, heightPerGlyph, true)._group;

        // if second level aggregation enabled: show also empty glyphs there.
        if(secondLevelEnabled){
          for(let j=0; j<numOfSecondLevel; j++){
            let innerGlyph = new EmptyGlyph(glyphGroup, dimension, dimension.values[i], glyphLayout, layoutDataRecordGroups._dataRecordGroups[i]._label, VisualizationProperties.GLYPH_SIZE, true)._group;
            innerGlyph.attr("transform","translate(" + VisualizationProperties.GLYPH_SIZE + "," + (j * VisualizationProperties.GLYPH_SIZE) + ")");
          }
        }

      }else{

        // compute whether the value is significant
        let significant = statisticsService.numericalAggregatedValueIsSignificant(dimension, dimension.values[i]);
        glyph = new NumericalGlyph(glyphGroup, tooltipService, dimension, dimension.values[i], globalDomain, glyphLayout, significant,layoutDataRecordGroups._dataRecordGroups[i]._label, heightPerGlyph, statisticsService)._group;

        // show second level aggregation, if enabled.
        if(secondLevelEnabled){
          for(let j=0; j<numOfSecondLevel; j++){

            let innerGlyph = undefined;
            if(dimension.values[i].nextLevelAggregation !== null && dimension.values[i].nextLevelAggregation !== undefined && !dimension.values[i].nextLevelAggregation[j].aggregatedValueEmpty){
              let aggValue = dimension.values[i].nextLevelAggregation[j];
              let sign = statisticsService.numericalAggregatedValueIsSignificant(dimension, aggValue);
              innerGlyph = new NumericalGlyph(glyphGroup, tooltipService, dimension, aggValue, globalDomain, glyphLayout, sign, layoutDataRecordGroups._dataRecordGroups[i]._label, VisualizationProperties.GLYPH_SIZE, statisticsService)._group;
            }else{
              innerGlyph = new EmptyGlyph(glyphGroup, dimension, dimension.values[i], glyphLayout, layoutDataRecordGroups._dataRecordGroups[i]._label, VisualizationProperties.GLYPH_SIZE, true)._group;
            }
            innerGlyph.attr("transform","translate(" + VisualizationProperties.GLYPH_SIZE + "," + (j * VisualizationProperties.GLYPH_SIZE) + ")");

          }
        }

      }

      // move to correct position
      if(secondLevelEnabled){
        glyphGroup.attr("transform","translate(" + 0 + "," + (VisualizationProperties.GLYPH_SPACE + (i * heightPerGlyphWithSpace)) + ")");
      }else{
        glyphGroup.attr("transform","translate(" + 0 + "," + (i * heightPerGlyphWithSpace) + ")");
      }

      // add the name of the glyph to the object, so that we can select specific object.
      glyphGroup.classed("glyph", true);
      glyphGroup.classed("glyphOfRow" + layoutDataRecordGroups._dataRecordGroups[i]._name, true);

    }

    // if enabled, show the statistical information of the whole dimension.
    if(glyphLayout._showDimensionIsSignificant){
      super.addStatisticSignificanceOfDimension();
    }


  }



}

'use strict';

/**
 * A categorical dimension with the glyphs for all groups.
 */
class CategoricalDimensionVis extends DimensionVis{

  /**
   * @param selector a d3.selector to attach the visualization to.
   * @param tooltipService the service used to manage the different tooltips. See services/tooltipService.js
   * @param dimension the dimension information; see featureTable#_nameToDimension[dimName]
   *        - dimName
   *        - dimType
   *        - globalAllValues
   *        - globalMax, Mean, Median, Min, MissingValues, StandardDeviation, Variance
   *        - values [g1, g2, ..., gn]
   *          g_i = {
   *            allValues,
   *            min,max,median,missingValues,standardDeviation,variance
   *          }
   * @param glyphLayout {GlyphLayout} the layout of the glyph.
   * @param layoutDataRecordGroups {LayoutDataRecordGroups} the layout of the data record groups.
   * @param statisticsService the service for the statistics, used to select the correct statistical test.
   */
  constructor(selector, tooltipService, dimension, glyphLayout, layoutDataRecordGroups, statisticsService, selectionService){

    super(selector, tooltipService, dimension, glyphLayout, layoutDataRecordGroups, statisticsService, selectionService);

    let dimName = dimension.dimName;

    // info whether to visualize the second level or not.
    let secondLevelEnabled = layoutDataRecordGroups._maxLevelOfAggregation > 1;
    let numOfSecondLevel = layoutDataRecordGroups.getNumberOfRecordGroups2ndLevel();

    // ---------------------------------------------------------------------
    // background of the distribution glyph
    this.glyphGroup
      .append("rect")
      .attr("class","backgroundDimDistribution highlight")
      .attr("x",0)
      .attr("y",0)
      .attr("height", VisualizationProperties.GLYPH_SIZE)
      .attr("width", function(){return secondLevelEnabled ? 2*VisualizationProperties.GLYPH_SIZE : VisualizationProperties.GLYPH_SIZE;});

    // -----------------
    // glyph for the distribution of the whole dimension + scale to the correct size.
    let distanceBetweenHistograms = VisualizationProperties.SPACE_DISTRIBUTION_GLYPH;
    let w = secondLevelEnabled ? VisualizationProperties.GLYPH_SIZE/2 : 0;
    let dimDistribution = new CategoricalHistogram(this.glyphGroup, dimension.globalDistribution, dimension.globalUniqueValues)._group;
    let sizeOfDistribtuion = CategoricalHistogram.getSizeOfHistogram();

    // scale to the correct size + move to the bottom.
    let scale = (VisualizationProperties.GLYPH_SIZE - distanceBetweenHistograms) / sizeOfDistribtuion[0]; // also translate - make sure there is a small gap between the histograms.
    dimDistribution.attr("transform","translate(" + w + ",0) scale(" + scale + ") translate(" + (distanceBetweenHistograms) + "," + (VisualizationProperties.GLYPH_SIZE - (VisualizationProperties.GLYPH_SIZE - (sizeOfDistribtuion[1] * scale))) + ")");

    // -----------------
    // listener for highlighting
    super.addListenerForHighlighting(dimName, secondLevelEnabled);



    // ---------------------------------------------------------------------
    let heightPerGlyph = secondLevelEnabled ? numOfSecondLevel * VisualizationProperties.GLYPH_SIZE : VisualizationProperties.GLYPH_SIZE;
    let heightPerGlyphWithSpace = DimensionGroupVis.computeHeightForGlyph(layoutDataRecordGroups);

    // iterate over all rows of the visualization
    for(let i=0; i<this.numGroups; i++){

      let glyph = undefined;
      let glyphGroup = this.groupOfGlyphs
        .append("g");



      // check if the current value is empty or not.
      if(dimension.values[i].aggregatedValueEmpty === true){

        glyph = new EmptyGlyph(glyphGroup, dimension, dimension.values[i], glyphLayout, layoutDataRecordGroups._dataRecordGroups[i]._label, heightPerGlyph, true)._group;

        // if second level aggregation enabled: show also empty glyphs there.
        if(secondLevelEnabled){
          for(let j=0; j<numOfSecondLevel; j++){
            let innerGlyph = new EmptyGlyph(glyphGroup, dimension, dimension.values[i], glyphLayout, layoutDataRecordGroups._dataRecordGroups[i]._label, VisualizationProperties.GLYPH_SIZE, true)._group;
            innerGlyph.attr("transform","translate(" + VisualizationProperties.GLYPH_SIZE + "," + (j * VisualizationProperties.GLYPH_SIZE) + ")");
          }
        }

      }else{

      // compute whether the value is significant or not.
        let significant = statisticsService.categoricalAggregatedValueIsSignificant(dimension, dimension.values[i]);
        glyph = new CategoricalGlyph(glyphGroup, tooltipService, dimension, dimension.values[i], glyphLayout, significant, layoutDataRecordGroups._dataRecordGroups[i]._label, heightPerGlyph, statisticsService)._group;

        // show second level aggregation, if enabled.
        if(secondLevelEnabled){
          for(let j=0; j<numOfSecondLevel; j++){

            let innerGlyph = undefined;
            if(dimension.values[i].nextLevelAggregation !== null && dimension.values[i].nextLevelAggregation !== undefined && !dimension.values[i].nextLevelAggregation[j].aggregatedValueEmpty) {
              let aggValue = dimension.values[i].nextLevelAggregation[j];
              let sign = statisticsService.categoricalAggregatedValueIsSignificant(dimension, aggValue);
              innerGlyph = new CategoricalGlyph(glyphGroup, tooltipService, dimension, aggValue, glyphLayout, sign, layoutDataRecordGroups._dataRecordGroups[i]._label, VisualizationProperties.GLYPH_SIZE, statisticsService)._group;
            }else{
              innerGlyph = new EmptyGlyph(glyphGroup, dimension, dimension.values[i], glyphLayout, layoutDataRecordGroups._dataRecordGroups[i]._label, VisualizationProperties.GLYPH_SIZE, true)._group;
            }
            innerGlyph.attr("transform", "translate(" + VisualizationProperties.GLYPH_SIZE + "," + (j * VisualizationProperties.GLYPH_SIZE) + ")");

          }
        }

      }

      // move to correct position
      if(secondLevelEnabled){
        glyphGroup.attr("transform","translate(" + 0 + "," + (VisualizationProperties.GLYPH_SPACE + (i * heightPerGlyphWithSpace)) + ")");
      }else{
        glyphGroup.attr("transform","translate(" + 0 + "," + (i * heightPerGlyphWithSpace) + ")");
      }

      // add the name of the glyph to the object, so that we can select specific object.
      glyphGroup.classed("glyph", true);
      glyphGroup.classed("glyphOfRow" + layoutDataRecordGroups._dataRecordGroups[i]._name, true);

    }

    // if enabled, show the statistical information of the whole dimension.
    if(glyphLayout._showDimensionIsSignificant){
      super.addStatisticSignificanceOfDimension();
    }

  }

}

'use strict';

const GLYPH_DESIGN = [
  {'id': 'UnifiedSamePosNeg', 'name': 'WHITE-RED colormap', 'colormap': VisualizationProperties.COLORMAP_WHITE_RED},
  {'id': 'UnifiedSamePosNegBlue', 'name': 'WHITE-BLUE colormap', 'colormap': VisualizationProperties.COLORMAP_WHITE_BLUE},
  {'id': 'UnifiedSamePosNegDark', 'name': 'WHITE-DARK RED colormap', 'colormap': VisualizationProperties.COLORMAP_WHITE_DARKRED},
  {'id': 'UnifiedSamePosNegDarkBlue', 'name': 'WHITE-DARK BLUE colormap', 'colormap': VisualizationProperties.COLORMAP_WHITE_DARKBLUE},

  {'id': 'UnifiedDotDistPos', 'name': 'WHITE-RED colormap, DOT = positive distance', 'colormap': VisualizationProperties.COLORMAP_WHITE_RED},
  {'id': 'UnifiedDotDistPosBlue', 'name': 'WHITE-BLUE colormap, DOT = positive distance', 'colormap': VisualizationProperties.COLORMAP_WHITE_BLUE},
  {'id': 'UnifiedDotDistPosDark', 'name': 'WHITE-DARK RED colormap, DOT = positive distance', 'colormap': VisualizationProperties.COLORMAP_WHITE_DARKRED},
  {'id': 'UnifiedDotDistPosDarkBlue', 'name': 'WHITE-DARK BLUE colormap, DOT = positive distance', 'colormap': VisualizationProperties.COLORMAP_WHITE_DARKBLUE},

  //{'id': 'UnifiedLineDistPos', 'name': 'WHITE-RED colormap, same scheme for pos/neg distance // LINE = positive distance', 'colormap': VisualizationProperties.COLORMAP_WHITE_RED},
  {'id': 'BipolarBlueNegRedPos', 'name': 'Bi-polar Colormap, RED = positive / BLUE = negative distance', 'colormap': VisualizationProperties.COLORMAP_BIPOLAR},
  {'id': 'BipolarBlueNegRedPosDark', 'name': 'Bi-polar DARK Colormap, RED = positive / BLUE = negative distance', 'colormap': COLORMAP_BIPOLAR_DARK},
];

const COLORMAPS = [
  {'id': 'UnifiedSamePosNeg', 'name': 'WHITE-RED colormap', 'colormap': VisualizationProperties.COLORMAP_WHITE_RED},
  {'id': 'UnifiedSamePosNegBlue', 'name': 'WHITE-BLUE colormap', 'colormap': VisualizationProperties.COLORMAP_WHITE_BLUE},
  {'id': 'UnifiedSamePosNegDark', 'name': 'WHITE-DARK RED colormap', 'colormap': VisualizationProperties.COLORMAP_WHITE_DARKRED},
  {'id': 'UnifiedSamePosNegDarkBlue', 'name': 'WHITE-DARK BLUE colormap', 'colormap': VisualizationProperties.COLORMAP_WHITE_DARKBLUE},
  {'id': 'BipolarBlueNegRedPos', 'name': 'Bi-polar Colormap', 'colormap': VisualizationProperties.COLORMAP_BIPOLAR},
  {'id': 'BipolarBlueNegRedPosDark', 'name': 'Bi-polar DARK Colormap', 'colormap': VisualizationProperties.COLORMAP_BIPOLAR_DARK},
];

/**
 * The glyph of one numerical dimension/group combination
 */
class NumericalGlyph extends Glyph{

  /**
   * @param selector a d3.selector to attach the visualization to.
   * @param tooltipService the service used to manage the different tooltips. See services/tooltipService.js
   * @param dimension data about the whole dimension.
   * @param glyphData in the form
   *          {
   *            allValues -> array with actual data, not normalized,
   *            min,max,median,missingValues,standardDeviation,variance,
   *            measure <- used for the visualization
   *          }
   * @param dimensionDomain the domain of the whole dimension. Used for mouse over distributions.
   * @param glyphLayout {GlyphLayout} the layout of the glyph.
   * @param significant true/false. If true, the value is significantly interesting.
   * @param recordGroupName the name of the data record group.
   * @param heightOfGlyph the height of the glyph.
   */
  constructor(selector, tooltipService, dimension, glyphData, dimensionDomain, glyphLayout, significant, recordGroupName, heightOfGlyph, statisticsService){

    super(selector, dimension, glyphLayout, heightOfGlyph);

    this.dimensionDomain = dimensionDomain;
    this.glyphData = glyphData;

    // decide whether to use the full range of the colormap or only part.
    let measureNormalized = glyphData.measureNormalized;
    if(glyphLayout._showDimensionIsSignificant && glyphLayout._useSmallColormapForNonSignificantDimensions && !statisticsService.dimensionIsSignificant(dimension)){
      measureNormalized *= VisualizationProperties.MAX_COLORMAP_VALUE_INNER;
    }

    // the background color which is used for the glyph.
    this.color = undefined;

    // ============================== Distance to mean ========================================
    if(glyphLayout.numericalMeasure === 'distanceToMean'){

      let layout = _.findIndex(GLYPH_DESIGN, function(d){return d.id === glyphLayout.numericalDesign});
      layout = GLYPH_DESIGN[layout];

      // ---------------------------------------------------------------------------------
      // ----- use a unified colormap
      if(glyphLayout.numericalDesign.startsWith('Unified')){

        this.color = VisualizationProperties.getColorMapValue(measureNormalized, layout.colormap);

        super.addSingleColorGlyph(this.color);

        // rect in middle of glyph, if it is a positive distance
        if(glyphLayout.numericalDesign.includes('UnifiedDot') && glyphData.measure > 0) {
            super.addOverlayDot("#3e3e3e");
        }

        // make a line once the distance is positive, no not, if the distance is negative.
        if(glyphLayout.numericalDesign.includes('UnifiedLine') && glyphData.measure > 0){
            super.addOverlayLine("#3e3e3e");
        }

      }

      // ---------------------------------------------------------------------------------
      // ----- use a bipolar colormap
      if(glyphLayout.numericalDesign === 'BipolarBlueNegRedPos'){

        this.color = undefined;
        if(glyphData.measure > 0){
          this.color = VisualizationProperties.getColorMapValue(Math.abs(measureNormalized), VisualizationProperties.COLORMAP_WHITE_RED);
        }else{
          this.color = VisualizationProperties.getColorMapValue(Math.abs(measureNormalized), VisualizationProperties.COLORMAP_WHITE_BLUE);
        }

        super.addSingleColorGlyph(this.color);

      }

      if(glyphLayout.numericalDesign === 'BipolarBlueNegRedPosDark'){

        this.color = undefined;
        if(glyphData.measure > 0){
          this.color = VisualizationProperties.getColorMapValue(Math.abs(measureNormalized), VisualizationProperties.COLORMAP_WHITE_DARKRED);
        }else{
          this.color = VisualizationProperties.getColorMapValue(Math.abs(measureNormalized), VisualizationProperties.COLORMAP_WHITE_DARKBLUE);
        }

        super.addSingleColorGlyph(this.color);

      }


    // ============================== Regular Measure such as mean, min, etc. ========================================
    }else{

      // this.color = VisualizationProperties.getColorMapValue(normalization(glyphData.measure), VisualizationProperties.COLORMAP_BLUE_WHITE_RED);
      let layout = _.findIndex(COLORMAPS, function(d){return d.id === glyphLayout.numericalDesign});
      layout = COLORMAPS[layout];
      this.color = VisualizationProperties.getColorMapValue(measureNormalized, layout.colormap);

      super.addSingleColorGlyph(this.color);
    }

    // =======================================================================================================
    // check if the user enabled to highlight significant values
    this.checkOverlayForSignificance(this.color, significant);


    // =======================================================================================================
    // check if an overlay is selected by the user.
    if(glyphLayout.showDistributionOverlay === true){
      this.addDistributionAsOverlay()
    }

    // =======================================================================================================
    // check if the user enabled to visualize missing values
    if(glyphLayout._showMissingValues !== 'disabled'){
      super.addMissingValues(dimension, glyphData, glyphLayout._showMissingValues);
    }

    // =======================================================================================================
    // overlay to capture the mouse events
    super.addTooltipDetails(tooltipService, dimension, glyphData, recordGroupName);


    // =======================================================================================================
    // overlay for highlighting rows and columns
    super.addHighlightingOverlay();

  }

  /**
   * Overlay over the glyphs: show distribution.
   */
  addDistributionAsOverlay(){
    let dimDistribution = new Beanplot(this._group, this.glyphData.allValues, this.glyphData.measure, {"background": false, "domain": this.dimensionDomain})._group;
    let size = Beanplot.size;

    let margin = 3;
    let scale = (VisualizationProperties.GLYPH_SIZE - 2*margin) / size[0]; // - 2*margin

    // move to correct position
    dimDistribution.attr("transform","translate(" + margin  + "," + margin + ") scale(" + scale + ")"); // space and +1 for the global distribution of the dimension.
  }


  /**
   * Return all available glyph designs for numerical dimensions.
   * @return {[*]}
   * @constructor
   */
  static get GLYPH_DESIGN(){
    return GLYPH_DESIGN;
  }

  static get COLORMAPS(){
    return COLORMAPS;
  }
}

'use strict';

const GLYPH_DESIGN_CATEGORICAL = [
  {'id': 'UnifiedColormap', 'name': 'WHITE-RED', 'colormap': VisualizationProperties.COLORMAP_WHITE_RED},
  {'id': 'UnifiedColormapDark', 'name': 'WHITE-DARK RED', 'colormap': VisualizationProperties.COLORMAP_WHITE_DARKRED},
  {'id': 'UnifiedColormapBlue', 'name': 'WHITE-BLUE', 'colormap': VisualizationProperties.COLORMAP_WHITE_BLUE},
  {'id': 'UnifiedColormapDarkBlue', 'name': 'WHITE-DARK BLUE', 'colormap': VisualizationProperties.COLORMAP_WHITE_DARKBLUE},
];

/**
 * The glyph of one categorical dimension/group combination
 */
class CategoricalGlyph extends Glyph {

    /**
   * @param selector a d3.selector to attach the visualization to.
   * @param tooltipService the service used to manage the different tooltips. See services/tooltipService.js
   * @param dimension data about the whole dimension.
   * @param glyphData in the form
   *          {
   *            allValues -> array with actual data, not normalized,
   *            min,max,median,missingValues,standardDeviation,variance
   *          }
   * @param glyphLayout {GlyphLayout} the layout of the glyph.
   * @param significant true/false. If true, the value is significantly interesting.
   * @param recordGroupName the name of this record group.
   */
  constructor(selector, tooltipService, dimension, glyphData, glyphLayout, significant, recordGroupName, heightOfGlyph, statisticsService){

    super(selector, dimension, glyphLayout, heightOfGlyph);

    this.glyphData = glyphData;

      // decide whether to use the full range of the colormap or only part.
      let measureNormalized = glyphData.measureNormalized;
      if(glyphLayout._showDimensionIsSignificant && glyphLayout._useSmallColormapForNonSignificantDimensions){
        console.error("Adaptive colormap for categorical dimensions not enabled.");
      }
      // if(glyphLayout._showDimensionIsSignificant && glyphLayout._useSmallColormapForNonSignificantDimensions && !statisticsService.dimensionIsSignificant(dimension)){
      //   measureNormalized *= VisualizationProperties.MAX_COLORMAP_VALUE_INNER;
      // }

    // the background color which is used for the glyph.
    this.color = undefined;


    let layout = _.findIndex(GLYPH_DESIGN_CATEGORICAL, function(d){return d.id === glyphLayout.categoricalDesign});
    layout = GLYPH_DESIGN_CATEGORICAL[layout];

    this.color = VisualizationProperties.getColorMapValue(measureNormalized, layout.colormap);
    super.addSingleColorGlyph(this.color);


    // =======================================================================================================
    // check if the user enabled to highlight significant values
    this.checkOverlayForSignificance(this.color, significant);

    // =======================================================================================================
    // check if an overlay is selected by the user.
    if(glyphLayout.showDistributionOverlay === true){
      this.addDistributionAsOverlay();
    }

    // =======================================================================================================
    // check if the user enabled to visualize missing values
    if(glyphLayout._showMissingValues !== 'disabled'){
      super.addMissingValues(dimension, glyphData, glyphLayout._showMissingValues);
    }

    // =======================================================================================================
    // overlay to capture the mouse events
    super.addTooltipDetails(tooltipService, dimension, glyphData, recordGroupName);

    // =======================================================================================================
    // overlay for highlighting rows and columns
    super.addHighlightingOverlay();

  }

  static get GLYPH_DESIGN_CATEGORICAL(){
    return GLYPH_DESIGN_CATEGORICAL;
  }

  /**
   * Overlay over the glyphs: show distribution.
   */
  addDistributionAsOverlay(){
    let dimDistribution = new CategoricalHistogram(this._group, this.glyphData.distribution, this.glyphData.uniqueValues, {"background": false})._group;
    let size = CategoricalHistogram.getSizeOfHistogram();
    let margin = 3;
    let scale = (VisualizationProperties.GLYPH_SIZE - 2*margin) / size[0]; // - 2*margin

    // move to correct position
    dimDistribution.attr("transform","translate(" + margin  + "," + margin + ") scale(" + scale + ")"); // space and +1 for the global distribution of the dimension.
  }

}

'use strict';

/**
 * Histogram for categorical dimensions.
 */
class CategoricalHistogram{

  /**
   * Create a new histogram for categorical dimensions.
   * @param selector d3 selector to add the histogram to
   * @param distribution the distribution to be used. The distribution has the following order:
   *        - TODO
   * @param orderOfValues the order of the values -> order of bins in the histogram.
   * @param config optional config parameter. Contains the following info:
   *        config.background = {true, false}; default is false
   *        config.labels = {true, false}; default is false
   */
  constructor(selector, distribution, orderOfValues, config){

    // console.log("CategoricalHistogram - distribution", distribution);
    // console.log("CategoricalHistogram - orderOfValues", orderOfValues);

    this._selector = selector;
    this._group = selector
      .append("svg:g")
      .attr("class", "categoricalHistogram");

    // let self = this._group;

    const WIDTH = CategoricalHistogram.getSizeOfHistogram()[0];
    const HEIGHT = CategoricalHistogram.getSizeOfHistogram()[1];

    // optional background
    if(config === undefined || config !== undefined && config.background === true) {
      this._group.append("rect")
        .attr("class", "background")
        .attr("x", 0)
        .attr("y", 0)
        .attr("width", WIDTH)
        .attr("height", HEIGHT);
    }

    let marginLeftRight = 5;
    // the width of one bar
    let barWidth = (WIDTH - 2*marginLeftRight) / orderOfValues.length;

    // scale to determine the height of the bars. (assume domain is in [0,1], to make the plots comparable)
    let marginTopBottom = 5;
    let h = HEIGHT - (2 * marginTopBottom);
    let yScale = undefined;
    let offsetBottom = undefined;

    // Optional description / label for each bin
    // If enabled, use bottom 1/3 for labels, top 2/3 for the actual histogram
    if(config !== undefined && config.labels === true) {
      yScale = d3.scaleLinear().domain([0,1]).range([0, 2/3*h]);
      offsetBottom = marginTopBottom + h/3;

      // draw the necessary labels
      for(let i=0; i<orderOfValues.length; i++){
        this._group.append("text")
          .attr("class","label")
          .text(orderOfValues[i] + "\n (" + (100*distribution[orderOfValues[i]]).toFixed(1) + "%)")
          .attr("transform","translate(" + (marginLeftRight + i*barWidth + barWidth/2) + "," + (2/3*h + 20) + ") rotate(-90)");
      }

    }else{
      yScale = d3.scaleLinear().domain([0,1]).range([0, h]);
      offsetBottom = marginTopBottom;
    }

    // draw all rectangles
    for(let i=0; i<orderOfValues.length; i++){
      this._group.append("rect")
        .attr("class","bar")
        .attr("x", 0)
        .attr("y", 0)
        .attr("width", barWidth)
        .attr("height", yScale(distribution[orderOfValues[i]]))
        .attr("transform", "translate(" + (marginLeftRight + i*barWidth) + "," + (HEIGHT-offsetBottom - yScale(distribution[orderOfValues[i]])) + ")");
    }

  }


  get selector() {
    return this._selector;
  }

  get group() {
    return this._group;
  }

  /**
   * @returns {[number,number]} the size of the created histogram.
   */
  static getSizeOfHistogram(){
    return [1000,1000];
  }

}

'use strict';

/**
 * Histogram for numerical dimensions.
 */
class NumericalHistogram{


  /**
   * Create a new histogram for categorical dimensions.
   * @param selector d3 selector to add the histogram to
   * @param allValues the values which need to be converted into a histogram.
   * @param numBins the number of bins to be used.
   * @param config optional config parameter. Contains the following info:
   *        config.background = {true, false}; default is false
   *        config.labels = {true, false}; default is false
   */
  constructor(selector, allValues, numBins, config){

    this._selector = selector;
    this._group = selector
      .append("svg:g")
      .attr("class", "categoricalHistogram");   // same style as the categorical histogram.

    const WIDTH = NumericalHistogram.getSizeOfHistogram()[0];
    const HEIGHT = NumericalHistogram.getSizeOfHistogram()[1];

    // optional background TODO change back to grey background
    if(config === undefined || config !== undefined && config.background === true) {
      this._group.append("rect")
        .attr("class", "background")
        .attr("x", 0)
        .attr("y", 0)
        .attr("width", WIDTH)
        .attr("height", HEIGHT);
    }

    let margin = 5;
    let h = HEIGHT - (2 * margin);
    let w = WIDTH - (2 * margin);
    let minMax = d3.extent(allValues);

    let xScale = d3.scaleLinear().domain(minMax).range([0,w]);

    let stepSize = (minMax[1]-minMax[0]) / (numBins);
    let ticks = [];
    for(let i=0; i<numBins-1; i++){
      ticks.push(minMax[0] + (i+1)*stepSize);
    }

    // let histogram = d3.histogram().domain(xScale.domain()).thresholds(numBins)(allValues);
    let histogram = d3.histogram().domain(xScale.domain()).thresholds(ticks)(allValues);
    // let histogram = d3.histogram().domain(xScale.domain()).thresholds(xScale.ticks(numBins))(allValues);
    this._storeHistogram = histogram;
    // let histogram = d3.histogram().domain(xScale.domain()).thresholds(thresholds)(allValues);

    let maxHeightBin = d3.max(histogram, function(bin){return bin.length / allValues.length});
    maxHeightBin *= 1.1;

    let yScale = undefined;
    let offsetBottom = undefined;

    let binWidth = w / histogram.length;

    // Optional description / label for each bin
    // If enabled, use bottom 1/3 for labels, top 2/3 for the actual histogram
    if(config !== undefined && config.labels === true) {
      yScale = d3.scaleLinear().domain([0,maxHeightBin]).range([0, 2/3*h]);
      offsetBottom = margin + h/3;

      // draw the necessary labels
      // console.log("histogram",histogram);
      this._group.append("g")
        .selectAll("text")
        .data(histogram)
        .enter()
        .append("text")
        .attr("class","label")
        .text(function(d){return "[ " + d.x0.toFixed(1) + ", " + d.x1.toFixed(1) + " ] - " + (100 * d.length / allValues.length).toFixed(1) + "%";})
        .attr("transform",function(d,i){return "translate(" + (margin + i*binWidth + binWidth/2) + "," + (2/3*h + 20) + ") rotate(-90)"});


    }else{
      yScale = d3.scaleLinear().domain([0, maxHeightBin]).range([0, h]);
      offsetBottom = margin;
    }

    // draw the histogram

    this._group.append("g")
      .selectAll("rect")
      .data(histogram)
      .enter()
      .append("rect")
        .attr("class","bar")
        .attr("x", 0)
        .attr("y", 0)
        .attr("width",binWidth)
        // .attr("width", function (d, i) {
        //     return xScale(d.x1) - xScale(d.x0);
        // })
        .attr("height", function (d) {
          return yScale(d.length / allValues.length);
        })
        .attr("transform",function(d,i){return "translate(" + (margin + i*binWidth) + "," + (HEIGHT-offsetBottom - yScale(d.length / allValues.length)) + ")"});

  }


  get selector() {
    return this._selector;
  }

  get group() {
    return this._group;
  }

  get storeHistogram() {
    return this._storeHistogram;
  }

  /**
   * @returns {[number,number]} the size of the created histogram.
   */
  static getSizeOfHistogram(){
    return [1000,800];
  }

}

'use strict';

/**
 * The glyph cells which are empty. The glyph looks as follows: light grey with a darker grey x in the middle.
 */
class EmptyGlyph extends Glyph {

    /**
   * @param selector a d3.selector to attach the visualization to.
   * @param tooltipService the service used to manage the different tooltips. See services/tooltipService.js
   * @param dimension data about the whole dimension.
   * @param glyphData in the form
   *          {
   *            allValues -> array with actual data, not normalized,
   *            min,max,median,missingValues,standardDeviation,variance
   *          }
   * @param normalization d3.scale to normalize the data.
   * @param glyphLayout {GlyphLayout} the layout of the glyph.
   * @param significant true/false. If true, the value is significantly interesting.
   * @param recordGroupName the name of this record group.
   */
  constructor(selector, dimension, glyphData, glyphLayout, recordGroupName, heightOfGlyph, addHighlight){

      super(selector, dimension, glyphLayout, heightOfGlyph);

      this._selector = selector;
      this._group = selector.append("g");
      this.glyphData = glyphData;
      heightOfGlyph = heightOfGlyph !== undefined ? heightOfGlyph : VisualizationProperties.GLYPH_SIZE;
      addHighlight = addHighlight !== undefined ? addHighlight : false;

      // grey box with a cross in the center.
      this._group.append("rect")
        .attr("class", "glyph")
        .attr("x",0)
        .attr("y",0)
        .attr("width", VisualizationProperties.GLYPH_SIZE)
        .attr("height", heightOfGlyph)
        .attr("fill", "#eeeeee");

      let margin = 8;
      this._group.append("line")
        .attr("x1", 0)
        .attr("y1", 0)
        .attr("x2", VisualizationProperties.GLYPH_SIZE - 2*margin)
        .attr("y2", heightOfGlyph - 2*margin)
        .attr("stroke","#d7d7d7")
        .attr("stroke-width",2)
        .attr("stroke-linecap","round")
        .attr("transform","translate(" + margin + "," + margin + ")");

      this._group.append("line")
        .attr("x1", VisualizationProperties.GLYPH_SIZE - 2*margin)
        .attr("y1", 0)
        .attr("x2", 0)
        .attr("y2", heightOfGlyph - 2*margin)
        .attr("stroke","#d7d7d7")
        .attr("stroke-width",2)
        .attr("stroke-linecap","round")
        .attr("transform","translate(" + margin + "," + margin + ")");

      if(addHighlight){
        super.addHighlightingOverlay();
      }

  }


  get selector() {
    return this._selector;
  }

  set selector(value) {
    this._selector = value;
  }

  get group() {
    return this._group;
  }

  set group(value) {
    this._group = value;
  }
}

'use strict';

/**
 * Service which creates the feature table visualization.
 * Implemented as a service to use angular injection service.
 *
 */
angular.module('smartexploreApp')
  .service('featureTableVisualization', function ($rootScope, urlParameterService, labelService, tooltipService, statisticsService, selectionService) {


    this.createFeatureTable = function(d3Selector, featureTable, glyphLayout, layoutDataRecordGroups, layoutDimensionGroups){

      // check if all parameters are available.
      if(featureTable === null
        || featureTable === undefined
        || glyphLayout === null
        || glyphLayout === undefined
        || layoutDataRecordGroups === null
        || layoutDataRecordGroups === undefined
        || layoutDimensionGroups === null
        || layoutDimensionGroups === undefined
      ){
        return;
      }

      const SPACE_FOR_LABELS = 200;
      const MARGIN = {top:0, left: 50, bottom:100, right: 50};

      let secondAggregationEnaled = layoutDataRecordGroups.maxLevelOfAggregation > 1;
      let numRecordGroupsSecondLevel = undefined;
      let labels2ndAggregation = undefined;
      if(secondAggregationEnaled){
        numRecordGroupsSecondLevel = layoutDataRecordGroups.getNumberOfRecordGroups2ndLevel();
        labels2ndAggregation = layoutDataRecordGroups.getRecordGroupLabels2ndLevel();
      }

      // compute the expected size of the feature table to create the svg element.
      const NUM_DIMENSION_GROUPS = layoutDimensionGroups.getNumberOfDimensionGroups();
      let numOfGroups = layoutDataRecordGroups.getNumberOfRecordGroups();
      let dimensionGroups = layoutDimensionGroups._dimensionGroups;

      // compute the required space for the record group labels (=labels for the rows in the feature table)
      const WIDTH_RECORD_GROUP_DISTR_HISTOGRAM = 80;
      const SPACE_RECORD_GROUP_LABELS = 30 + d3.max(layoutDataRecordGroups._dataRecordGroups, function(d){  // 30 for percentage.
          return labelService.getWidthOfString(d._label + " (100%)", "12px");
        });
      const SPACE_RECORD_GROUP_LABELS_2_AGG = !secondAggregationEnaled ? 0 : 30 + d3.max(labels2ndAggregation, function(d){  // 30 for percentage.
          return labelService.getWidthOfString(d + " (100%)", "12px");
        });
      // the width of the record group vis (=histogram + labels, also for second level aggregation).
      const WIDTH_RECORD_GROUP_VIS = WIDTH_RECORD_GROUP_DISTR_HISTOGRAM + SPACE_RECORD_GROUP_LABELS + SPACE_RECORD_GROUP_LABELS_2_AGG;

      let svgHeight = MARGIN.top + MARGIN.bottom;
      svgHeight += SPACE_FOR_LABELS; // for the dimension labels
      svgHeight += numOfGroups * DimensionGroupVis.computeHeightForGlyph(layoutDataRecordGroups);
      svgHeight += 50;     // overhead for group visualization

      let svgWidth = MARGIN.left + WIDTH_RECORD_GROUP_VIS + VisualizationProperties.GROUP_SPACE + MARGIN.right;  // space -> space between row labels and dimension groups
      for(let i=0; i<dimensionGroups.length; i++){
        svgWidth += DimensionGroupVis.computeWidthForNumberOfDimensions(dimensionGroups[i].dimensions.length, layoutDataRecordGroups._maxLevelOfAggregation); // space for each group
        svgWidth += VisualizationProperties.GROUP_SPACE;
      }


      let chart = d3Selector
        .append('svg')
        .attr("id","smartexplore")
        .attr("class", "LIGHTTHEME")
        .attr('width', svgWidth)
        .attr('height', svgHeight);


      /* Background of the vis. */
      let background = chart.append("rect")
        .attr("class", "background")
        .attr("x", 0)
        .attr("y", 0)
        .attr("width", svgWidth)
        .attr("height", svgHeight)
        .on("mouseover", function(){

          // remove the menu of the dimension and record groups.
          removeMenuForRecordGroup();
          DimensionGroupVis.removeMenuForGroup();
        });


      /* Include now the margins etc. */
      chart = chart.append("g")
        .attr("x",0)
        .attr("y",0)
        .attr("transform","translate(" + MARGIN.left + "," + MARGIN.top + ")");

      /* Offset for the dimension groups = space for the bar chart of the record group vis. */
      let allDimGroups = chart.append("g")
        .attr("id","svgGroupAllDimensionGroups")
        .attr("transform","translate(" + (WIDTH_RECORD_GROUP_VIS + VisualizationProperties.GROUP_SPACE) + "," + SPACE_FOR_LABELS + ")");


      // ===========================================================================================================
      // create all groups and translate them to the correct position.
      // compute the positions of the different groups
      let positionsOfGroups = DimensionGroupVis.computeLocationOfGroups(dimensionGroups, layoutDataRecordGroups.maxLevelOfAggregation);
      for(let i=0; i<NUM_DIMENSION_GROUPS; i++){
        let group = dimensionGroups[i];
        let g = new DimensionGroupVis(allDimGroups, group.name, tooltipService, featureTable, glyphLayout, layoutDataRecordGroups, layoutDimensionGroups, statisticsService, selectionService).group;
        g.attr("groupName", group.name);
        g.attr("transform","translate(" + positionsOfGroups[group.name] + ", " + 0 + ")");
      }


      // ===========================================================================================================
      // add the labels of the record groups.

      // total number of records, used to compute the scale.
      let numAllRecords = d3.sum(layoutDataRecordGroups._dataRecordGroups, function(d){
        return d._size;
      });
      let maxFrequencyInGroup = d3.max(layoutDataRecordGroups._dataRecordGroups, function(d){
        return d._size/numAllRecords;
      });

      // scale for the histograms.
      let xScaleHistogramRows = d3.scaleLinear().domain([0,maxFrequencyInGroup]).range([0, WIDTH_RECORD_GROUP_DISTR_HISTOGRAM]);


      let labelGroup = chart.append("g")
        .attr("transform","translate(" + 0 + "," + (SPACE_FOR_LABELS + 10 + VisualizationProperties.GLYPH_SIZE + VisualizationProperties.SPACE_DISTRIBUTION_GLYPH) + ")");   // 10 for the radius and the distribution on top.


      let dragDropRow = d3.drag()
        .on("start", dragRowStarted)
        .on("drag", draggingRow)
        .on("end", dragRowEnded);

      // add the histogram below the labels.
      let heightGlyph = DimensionGroupVis.computeHeightForGlyph(layoutDataRecordGroups);
      let spaceAroundHist = 2;
      let heightInner = (heightGlyph-2*spaceAroundHist) / (2 * numRecordGroupsSecondLevel + 1);

      let group = labelGroup.selectAll(".recordGroupRowHist")
        .data(layoutDataRecordGroups._dataRecordGroups)
        .enter()
        .append("g")
        .attr("id", function(d){return "recordRowGroup" + d._name})
        .attr("class","recordRowGroup")
        .attr("recordGroupName", function(d){return d._name})
        .attr("transform",function(d,i){
          return "translate(" + 0 + "," + (i * heightGlyph) + ")";   // 1 for the border
        });

      // Add the histogram
      group.append("rect")
        .attr("class","recordGroupRowHist")
        .attr("x",0)
        .attr("y",0)
        .attr("width",function(d){return xScaleHistogramRows(d._size / numAllRecords)})
        .attr("height", heightGlyph - 2*spaceAroundHist)                           // -2 for the border
        .attr("transform","translate(" + SPACE_RECORD_GROUP_LABELS + "," + spaceAroundHist + ")") // border

      // add the labels
      group
        .append("text")
        .attr("class", "label recordGroupRowLabel")
        .attr("x",0)
        .attr("y",0)
        .text(function(d){return d._label + " (" + ((d._size / numAllRecords)*100).toFixed(1) + "%)"; })
        .attr("dominant-baseline","middle")
        .attr("text-anchor","end")
        .attr("transform","translate(" + (SPACE_RECORD_GROUP_LABELS - 10)+ "," + (heightGlyph/2) + ")");  // middle of the glyph

      if(secondAggregationEnaled){
        // add inner histogram for second level aggregation.
        group.selectAll(".recordGroupRowHistSecondAgg")
          .data(function(d){return secondAggregationEnaled ? d._nextLevelAggregation : [];})
          .enter()
          .append("g")
          .attr("transform", function(d,i){return "translate(" + SPACE_RECORD_GROUP_LABELS + "," + (spaceAroundHist + heightInner + i * (2*heightInner)) + ")"})
          .append("rect")
          .attr("class","recordGroupRowHistSecondAgg")
          .attr("x",0)
          .attr("y",0)
          .attr("width",function(d){return xScaleHistogramRows(d._size / numAllRecords);})
          .attr("height",heightInner)
          .select(function() { return this.parentNode; })
          .append("text")
          .attr("class", "label recordGroupRowLabel")
          .attr("x",0)
          .attr("y",0)
          .text(function(d){return d._label + " (" + ((d._size / numAllRecords)*100).toFixed(1) + "%)"; })
          .attr("dominant-baseline","middle")
          .attr("text-anchor","start")
          .attr("transform","translate(" + (WIDTH_RECORD_GROUP_DISTR_HISTOGRAM + 10)+ "," + (heightInner/2) + ")") // middle of the glyph
      }

      // Add drag/drop behavior
      group.call(dragDropRow);

      // Add listener for the drag & drop on top of a transparent rectangle.
      group
        .append("rect")
        .attr("recordGroupName", function(d){return d._name}) // attach e.g., dataRecordGroup200
        .attr("x",0)
        .attr("y",0)
        .attr("width", SPACE_RECORD_GROUP_LABELS + xScaleHistogramRows.range()[1] + 20)
        .attr("height",heightGlyph)                           // -2 for the border
        .attr("fill","red")
        .attr("opacity",0.0)  // transparent
        .attr("transform","translate(-10," + 0 + ")") // border
        .on("mouseover", function(d){

          // Change the style of the cursor.
          d3.select(this).style("cursor", "move");

          // Remove any highlight of the dimensions.
          if(selectionService.isAnyDimensionSelected()){
            selectionService.removeAllSelectedDimensions();
            selectionService.updateSelectionDimensions();
          }

          if(!selectionService.isAnyDataRecordGroupSelected()){

            // No row is currently selected.
            // Make all rows dark, then highlight this row.
            selectionService.removeHighlightAllDataRecordGroups();
            selectionService.addHighlightDataRecordGroup(d._name);

          }else{
            // There is at least one row already highlighted. So most rows are already dark.
            // Highlight only the current row.
            selectionService.addHighlightDataRecordGroup(d._name);
          }

          // show the menu for the data record groups
          let pos = d3.select(this).node().getBoundingClientRect();
          let x = window.scrollX + pos.x - 48; // icon is 48px.
          let y = window.scrollY + pos.y - 48/2 + heightGlyph/2;
          showMenuForRecordGroup(d._name, x, y);


        })
        .on("mouseout", function(){

          // Change the style of the cursor.
          d3.select(this).style("cursor", "default");

          selectionService.updateSelectedDataRecordGroups();

        })
        .on("click", function(d){

          if(selectionService.isSelectedDataRecordGroup(d._name)){
            selectionService.removeSelectedDataRecordGroup(d._name);
          }else{
            selectionService.addSelectedDataRecordGroup(d._name);
          }
        });


      let offsetMouseToTopBorder = undefined;
      let offsetDragRowX = undefined;

      // array of row + y-position (transformation)
      let yPositions = undefined;

      // the index of the element which is currently dragged.
      let indexOfDraggingElement = undefined;
      let nameOfDragged = undefined;

      /**
       * Start dragging a dimension group.
       */
      function dragRowStarted() {

        let row = d3.select(this);
        nameOfDragged = row.attr("recordGroupName");

        // extract the current transformation
        let t = row.attr("transform");
        let translate = t.substring(t.indexOf("(")+1, t.indexOf(")")).split(",");
        offsetMouseToTopBorder = d3.event.y - translate[1];
        offsetDragRowX = +translate[0];

        // get the index of the element which is currently dragged
        indexOfDraggingElement = _.findIndex(layoutDataRecordGroups._dataRecordGroups, function(d){
          return d._name === nameOfDragged;
        });

        // store the why position of all rows
        yPositions = [];
        layoutDataRecordGroups._dataRecordGroups.map(function(d){
          let o = d3.select("#recordRowGroup" + d._name).attr("transform");
          let trans = o.substring(o.indexOf("(")+1, o.indexOf(")")).split(",");
          yPositions.push({'groupName':d._name, 'y': +trans[1]});
        });

      }

      /**
       * Currently draging a dimension group. Update x position.
       */
      function draggingRow() {

        // move the label and the entire row of data
        translateRowToPosition(nameOfDragged,[0,d3.event.y - offsetMouseToTopBorder],false);

        // check whether a change with the upper or lower row needs to be done.
        if(indexOfDraggingElement > 0 && d3.event.y < yPositions[indexOfDraggingElement - 1].y){

          // move upper element one down
          let aboveName = yPositions[indexOfDraggingElement - 1].groupName;
          let newYPos = yPositions[indexOfDraggingElement].y;

          // move the label/histogram in the beginning
          translateRowToPosition(aboveName, [0,newYPos], true);

          // swap the two indices
          swap(layoutDataRecordGroups._dataRecordGroups, indexOfDraggingElement, indexOfDraggingElement-1);

          // swap the only the names of the two indices
          yPositions[indexOfDraggingElement-1].groupName = yPositions[indexOfDraggingElement].groupName;
          yPositions[indexOfDraggingElement].groupName = aboveName;

          // change the index of the currently dragged element.
          indexOfDraggingElement -= 1;

        }else if(indexOfDraggingElement < yPositions.length - 1 && d3.event.y > (VisualizationProperties.GLYPH_SIZE + yPositions[indexOfDraggingElement + 1].y)){
          // swap the two indices, move the lower row to the correct location.

          // move upper element one down
          let belowName = yPositions[indexOfDraggingElement + 1].groupName;
          let newYPos = yPositions[indexOfDraggingElement].y;

          // move the label/histogram in the beginning
          translateRowToPosition(belowName, [0,newYPos], true);

          // swap the two indices
          swap(layoutDataRecordGroups._dataRecordGroups, indexOfDraggingElement, indexOfDraggingElement+1);

          // swap the only the names of the two indices
          yPositions[indexOfDraggingElement+1].groupName = yPositions[indexOfDraggingElement].groupName;
          yPositions[indexOfDraggingElement].groupName = belowName;

          // change the index of the currently dragged element.
          indexOfDraggingElement += 1;

        }

      }

      /**
       * Dragging of a dimension group is complete.
       * Store the new order of the dimension groups.
       * Store that information in the URL.
       */
      function dragRowEnded() {

        // move the currently dragging element back to the grid
        translateRowToPosition(nameOfDragged,[0,yPositions[indexOfDraggingElement].y],true,800);

        offsetMouseToTopBorder = undefined;
        offsetDragRowX = undefined;
        yPositions = undefined;
        indexOfDraggingElement = undefined;
        nameOfDragged = undefined;

        // Update the parameters in the URL.
        $rootScope.$apply(function(){
          urlParameterService.setLayoutDataRecordGroups(layoutDataRecordGroups);
        });

      }

      /**
       * Swap the positions i and j of the given array
       * @param array
       * @param i
       * @param j
       * @return the swapped array.
       */
      function swap(array, i, j){
        let tmp = array[i];
        array[i] = array[j];
        array[j] = tmp;
        return array;
      }

      /**
       * Translate the row with the given recordGroupName to the given position.
       * Use a transition if enabled.
       * @param recordGroupName
       * @param position array with x and y of the new location.
       * @param transition
       */
      function translateRowToPosition(recordGroupName, position, transition=false, duration=2000){

        // move the label/histogram in the beginning
        if(transition) {
          d3.select("#recordRowGroup" + recordGroupName)
            .transition()
            .duration(duration)
            .attr("transform", "translate(" + position[0] + "," + position[1] + ")");

          // move all rows
          d3.selectAll(".glyphOfRow" + recordGroupName)
            .transition()
            .duration(duration)
            .attr("transform", "translate(" + position[0] + "," + position[1] + ")");

        }else{

          d3.select("#recordRowGroup" + recordGroupName)
            .attr("transform", "translate(" + position[0] + "," + position[1] + ")");

          // move all rows
          d3.selectAll(".glyphOfRow" + recordGroupName)
            .attr("transform", "translate(" + position[0] + "," + position[1] + ")");
        }

      }

      /**
       * Show the menu for the record group with the given name.
       * @param recordGroupName the name of the record group.
       * @param x x-pos
       * @param y y-pos
       */
      function showMenuForRecordGroup(recordGroupName, x, y){
        d3.select("#recordGroupMenu")
          .style("left", x + "px")
          .style("top", y + "px")
          .style("display", "inline")
          .attr("recordGroup", recordGroupName);
      }

      /**
       * Remove the menu for all record groups.
       */
      function removeMenuForRecordGroup(){
        d3.select("#recordGroupMenu")
          .style("display", "none");
      }



      // ====================================================================================
      // Add Drag & Drop to the dimension groups.
      addDragAndDropToAllDimensionGroups(urlParameterService, $rootScope);

    };

  });
